#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>
#include <time.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/times.h>

#include "tv_image_lib.h"

static int hist[6554];             /*6554 ~= 2^16 / 10*/

void mydelay(double dt)
{
  int nfd;
  struct timeval tv;
   tv.tv_sec = (int)dt;
   tv.tv_usec = (int)((dt - tv.tv_sec)*1000000.);
   if(tv.tv_sec==0 && tv.tv_usec<10000) tv.tv_usec=10000;
slipping:
   nfd = select(0, (fd_set *)NULL,(fd_set *)NULL,(fd_set *)NULL, &tv);
   if(nfd < 0) {
      if(errno == EINTR)
	/*On  Linux,  timeout  is  modified to reflect the amount of
	 time not slept; most other implementations DO NOT do this!*/
	 goto slipping;
      fprintf(stderr,"Error in mydelay(){ select() }. %s\n",strerror(errno));
   }
}

void main(int argc, char *argv[]) {
   short *img;                     /* image memory */
   int imgw, imgh;                 /* image size   */
   int nchan = 0;                  /* TV-chan. number */
   int col0, lin0, width, height;  /* frame: area of interest */
   int once = 1;
   int nseq;
   int i,j,k,x,y,pix;
   int n,nbot,nmid,ntop;
   int hist0,hbot,htop,avr,med,lev1,lev2,lev3;
   double dsp,spx,spy,s,xc,yc;


   if (argc>1) nchan = atoi(argv[1]);
   if (argc>2) once = atoi(argv[2]);
   get_vs56_sh_mem();
   img = tv_chan_image (nchan);
   imgw = tv_chan_img_width();
   imgh = tv_chan_img_height();
next_frame:
   nseq = tv_chan_seq_num();
   tv_chan_get_frame(&col0, &lin0, &width, &height);

   if(col0==0 && lin0==0 && width==imgw && height==imgh ) {
      fprintf(stderr,"No Frame defined for chan.%d\n",nchan);
      exit(1);
   }
   n = width*height;
   nbot = (int)(n*0.03 + 0.5);
   nmid = n/2;
   ntop = (int)(n*0.97 + 0.5);

   for (j=0; j<6554; j++) hist[j]=0;
   hist0 = -32768;
   for (j=0; j<height; j++) {
      y=lin0+j;
      for (i=0; i<width; i++) {
	 x=col0+i;
	 pix = img[imgw*y+x];
	 k = (pix-hist0)/10;
	 hist[k]++;
      }
   }
   for (i=j=0; j<6553 && i<nmid; j++, i += hist[j]);
   med = j*10+hist0;
   hist0 = med-3277;

   for (j=0; j<6554; j++) hist[j]=0;
   for (j=0; j<height; j++) {
      y=lin0+j;
      for (i=0; i<width; i++) {
	 x=col0+i;
	 pix = img[imgw*y+x];
	 k = pix-hist0;
	 k = (k<0)? 0 : ((k>6553)? 6553 : k);
	 hist[k]++;
      }
   }
   hbot = 0;
   htop = 6553;
   for (i=j=0; j<6553; j++) {
      i += hist[j];
      if(i<=nbot) hbot = j;
      if(i<=nmid) med  = j;
   /**if(i<=ntop) htop = j;**/
   }
   htop = med + (med-hbot);
/*
   printf("hist[%d:%d]\n",hist0,hist0+6553);
   for (j=hbot-10; j<=med; j++)
      printf("%d ",hist[j]);
   printf("\n");
   for (j=htop+10; j>=med; j--)
      printf("%d ",hist[j]);
   printf("\n");
*/
   med += hist0;
   for(k=0;k<7;k++) {
    /*printf("Range= %d - %d\n",hbot+hist0,htop+hist0);*/
      for (avr=i=0,j=hbot; j<htop; j++) {
	 avr += hist[j]*j;
	 i += hist[j];
      }
      avr /= i;
      for (dsp=0.,j=hbot; j<htop; j++) {
	 double d = j-avr;
	 dsp += d*d*hist[j];
      }
      avr += hist0;
      dsp = sqrt(dsp/i);
      /*printf("Med=%d Avr=%d Dsp=%.2f\n",med,avr,dsp);*/
      htop = avr+2.0*(dsp<1.?1:dsp)+0.5 - hist0;
   }
   lev1 = avr+dsp+0.5;
   lev2 = avr+2.*dsp+0.5;
   lev3 = avr+3.*dsp+0.5;

   spx=spy=s=0.;
   for (j=0; j<height; j++) {
      y=lin0+j;
      /*printf("%3d[ ",y);*/
      for (i=0; i<width; i++) {
	 x=col0+i;
	 pix = img[imgw*y+x];
	/*printf("%c",(pix<lev1)?' ':((pix<lev2)?'.':((pix<lev3)?'x':'X')));*/
	 if(pix>lev3){
	   spx += x*pix;
	   spy += y*pix;
	   s += pix;
	   /*printf("%c",'X');*/
	 } else {
	    /* printf("%c",' ');*/
	 }
      }
      /*printf("]\n");*/
   }
   if(s<1.) s=1.;
   xc=spx/s;
   yc=spy/s;
   printf("Center: Xc=%.1f Yc=%.1f\n",xc,yc);
   if(once)
      exit(0);
   while (nseq==(k=tv_chan_seq_num())) mydelay(0.2);
   while (k!=tv_chan_seq_num()) mydelay(0.2);
   goto next_frame;
}
