/* "line" & "column" background subtruction */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/times.h>

#include "tv_image_lib.h"

typedef int bool;
#define TRUE    ((bool)1)
#define FALSE   ((bool)0)

bool get_fits(char *fitsname, int col0, int lin0, int frmw, int frmh, short *array, int *tv_lev0, int *tv_elref, int *nadd);
static void stat_image(short *image, int width, int height);
static int vmn,med,vmx;
static double avr,dsp;

static double lin[1024],lin1[1024],col[1024],col1[1024];

int main(int argc, char *argv[]) {
   short *img;                     /* image memory */
   int imgw, imgh;                 /* image size   */
   int nchan = 0;                  /* TV-chan. number */
   short *frm1=NULL;               /* bgd-frame memory */
   int col0, lin0, frmw, frmh;     /* frame1: area of interest */
   int lev0,nadd,bg_lev0,bg_nadd;
   int width, height;
   int nseq;
   int i,j,k,x,y,pix,pix0,pix1,pix2;
   int n,mx,mn;
   double xmx=0.0,ymx=0.0;
   char tmp[100],fname[100];
   char tvdrv[20] = "";
   char *bgdfile=NULL;
   short *sp0,*sp;
   double av,av1,s,ds3;

   if (argc>1) {
      char *p = strchr(argv[1],':');
      if(p!=NULL || argv[1][0] == 'v' || argv[1][0] == 'V' || argv[1][0] == 'b') {
	 strncpy(tvdrv,argv[1],19);
	 if(p!=NULL) nchan = atoi(p+1);
      } else
	 nchan = atoi(argv[1]);
   }
   if (argc>2) bgdfile = argv[2];

   if(strncmp(tvdrv,"vs56",4)==0 || strncmp(tvdrv,"VS56",4)==0) get_vs56_sh_mem();
   else if(strncmp(tvdrv,"vs60",4)==0 || strncmp(tvdrv,"VS60",4)==0) get_vs60_sh_mem();
   else if(strncmp(tvdrv,"vsbt",4)==0 || strncmp(tvdrv,"bttv",4)==0) get_vsbt_sh_mem();
   else get_vs60_sh_mem();

   img = tv_chan_image (nchan);
   imgw = tv_chan_img_width();
   imgh = tv_chan_img_height();
   nseq = tv_chan_seq_num();

reload_bgd:
   if(isatty(0)) {
      fputs("Define the 1-st frame ...",stdout);
      fgets(tmp,10,stdin);
   }
   tv_chan_get_frame(&col0, &lin0, &frmw, &frmh);

   width=imgw; height=imgh;

   if(bgdfile==NULL &&(col0!=0 || lin0!=0 || frmw!=imgw || frmh!=imgh) ) {
      fprintf(stderr,"WARNING! No Frame but whole Image needs!\n");
   }

   if(frm1==NULL) {
      frm1 = (short *)malloc(width*height*sizeof(short));
   }
   if(bgdfile!=NULL) {
      if(!get_fits(bgdfile, 0,0, width,height, frm1, &bg_lev0, &i, &bg_nadd)) {
	 fputs("quit\n",stdout);
	 fflush(stdout);
	 exit(1);
      }
   } else {
      for (j=0; j<height; j++) {
	 for (i=0; i<width; i++) {
	    frm1[width*j+i] = img[imgw*j+i];
	 }
      }
      fprintf(stderr, "Load BGD from TV-image \n");
      bg_lev0 = tv_chan_addlev();
      bg_nadd = tv_chan_added();
   }
   stat_image(frm1, width, height);

   sp0 = frm1;
   for(j=0; j<height; j++, sp0 += width) {
      mx = (int)(avr+5.*dsp-1.5);;
      mn = (int)(avr-5.*dsp-1.5);
   col_av_ds3:
      sp = sp0;
      ds3 = 0.0;
      pix0 = pix1 = pix2 = *sp++;
      av = pix0;
      for(n=i=1; i<width; i++) {
	 pix = *sp++;
	 if(mn<=pix && pix<=mx) {
	    pix2 = pix1;
	    pix1 = pix0;
	    pix0 = pix;
	    av += pix;
	    s = pix1 - (pix0+pix2)/2;
	    ds3 +=  s*s;
	    n++;
	 }
      }
      if(n < width/2) {
	 mn = vmn; mx = vmx;
	 goto col_av_ds3;
      }
      av /= n;
      ds3 = sqrt(ds3/(n-1));
      for(k=1;k<8;k++) {
	 sp = sp0;
	 mx = (int)(av+7.*ds3/k+1.5);
	 mn= (int)(av-14.*ds3/k-1.5);
	 s=0.;
	 for(n=i=0; i<width; i++) {
	    pix = *sp++;
	    if(mn<=pix && pix<=mx) {
	       s += pix; n++;
	    }
	 }
	 if(n==0) n=1;
	 av = s/n;
      }
      col1[j] = av/bg_nadd;
   }
   sp0 = frm1;
   for(i=0; i<width; i++, sp0++) {
      mx = (int)(avr+5.*dsp-1.5);;
      mn = (int)(avr-5.*dsp-1.5);
   lin_av_ds3:
      sp = sp0;
      ds3 = 0.0;
      pix0 = pix1 = pix2  = *sp; sp += width;
      av = pix0;
      for(n=j=1; j<height; j++, sp += width) {
	 pix = *sp;
	 if(mn<=pix && pix<=mx) {
	    pix2 = pix1;
	    pix1 = pix0;
	    pix0 = pix;
	    av += pix;
	    s = pix1 - (pix0+pix2)/2;
	    ds3 +=  s*s;
	    n++;
	 }
      }
      if(n < width/2) {
	 mn = vmn; mx = vmx;
	 goto lin_av_ds3;
      }
      av /= n;
      ds3 = sqrt(ds3/(n-1));
      for(k=1;k<8;k++) {
	 sp = sp0;
	 mx = (int)(av+7.*ds3/k+1.5);
	 mn= (int)(av-14.*ds3/k-1.5);
	 s=0.;
	 for(n=j=0; j<height; j++, sp += width) {
	    pix = *sp;
	    if(mn<=pix && pix<=mx) {
	       s += pix; n++;
	    }
	 }
	 if(n==0) n=1;
	 av = s/n;
      }
      lin1[i] = av/bg_nadd;;
   }
   s = 0.0;
   for(i=0; i<width; i++) s += lin1[i];
   av1 = s/width;
   s = 0.0;
   for(j=0; j<height; j++) s += col1[j];
   av1 += s/height;

   free(frm1);
   frm1 = NULL;

get_command:
   if(isatty(0))
      fputs("Point to the 2-nd one ... (or type 'q' to quit) ",stdout);
   fgets(tmp,99,stdin);
   if(tmp[0]=='q')      exit(0);           /* quit */
   else if(tmp[0]=='r') goto reload_bgd;   /* reload */
   else if(tmp[0]=='f') {                  /* file .... */
       char *p;
       p=strtok(tmp," =\t\n");
       p=strtok(NULL," =\t\n");
       if(p!=NULL) {
	  if(strcmp(p,"no")) {
	     strncpy(fname,p,90);          /* file fname.mt */
	     bgdfile=fname;
	  } else
	     bgdfile=NULL;                 /* file no */
       } else
	  bgdfile=NULL;                    /* file */
       goto get_command;
   } else if(tmp[0]=='n' || tmp[0]=='\n')  /* next */
	goto next_frame;
   else if(tmp[0]=='s')                    /* subtract */
	goto subtruct_frm;
   else if(tmp[0]=='z') {                  /* zero ...*/
	if(strchr(tmp,'x'))
	   xmx = 0.;                       /* zero x */
	if(strchr(tmp,'y'))
	   ymx = 0.;                       /* zero y */
	if(strchr(tmp,'x')==NULL && strchr(tmp,'y')==NULL)
	   xmx = ymx = 0.;                 /* zero */
	goto get_command;
   } else goto get_command;

next_frame:
   printf("dX=0.0\ndY=0.0\n");
   fflush(stdout);
   goto get_command;

subtruct_frm:

   tv_chan_get_frame(&col0, &lin0, &frmw, &frmh);
   lev0 = tv_chan_addlev();
   nadd = tv_chan_added();

   for(i=0; i<width; i++) lin[i] = lin1[i]*nadd;
   for(j=0; j<height; j++) col[j] = col1[j]*nadd;
   av = av1*nadd;

   for (j=0; j<frmh; j++) {
     y=lin0+j;
     for (i=0; i<frmw; i++) {
	x=col0+i;
	img[imgw*y+x] -= (short)(lin[x]+col[y]-av+0.5);
     }
   }
   tv_chan_set_seq_num(tv_chan_seq_num()+1);
   goto get_command;
}



static void stat_image(short *image, int width, int height)
	   // do some statistics on image: calculate vmn,vmx,avr,med,dsp
{
   int i,j,pix;
   int n,nbot,nmid,ntop;
   int hist0,hbot,htop;
   short *sp;
   int *hist = (int *)calloc(65536, sizeof(int));
   if(hist==NULL) return;

   n = width*height;
   nbot = (int)(n*0.1 + 0.5);     //  ~20%  ""  ~10% ""
   ntop = (int)(n*0.8 + 0.5);     //   
   nmid = n/2;

   for (j=0; j<65536; j++) hist[j]=0;
   hist0 = -32768;
   vmn=32767;
   vmx=-32768;
   sp = image;
   for (j=0; j<height; j++) {
      for (i=0; i<width; i++) {
	 pix = *sp++;
	 hist[pix-hist0]++;
	 if(vmn>pix) vmn = pix;
	 if(vmx<pix) vmx = pix;
      }
   }
   hbot = 0;
   med = 32767;
   htop = 65535;                            //  "" 
   for (i=0,j=vmn-hist0; j<vmx-hist0; j++) {//    ~10% 
      i += hist[j];                         //  ""  avr  dsp
      if(i<=nbot) hbot = j;
      if(i<=nmid) med  = j+1;
      if(i<=ntop) htop = j+1;
   }
   med += hist0;
//printf("mbot=%d  nmid=%d  ntop=%d  hbot=%d  med=%d  htop=%d\n",nbot,nmid,ntop,hbot+hist0,med,htop+hist0);
   for (avr=0.,i=0,j=hbot; j<=htop; j++) {
      avr += (double)hist[j]*j;
      i += hist[j];
   }
   if(i<1) i=1;
   avr /= i;
   for (dsp=0.,j=hbot; j<=htop; j++) {
      double d = j-avr;
      dsp += d*d*hist[j];
   }
   avr += hist0;
   dsp = sqrt(dsp/i);
/*fprintf(stderr,"Min=%d Max=%d Med=%d Avr=%.1f Sigma=%.2f \n",
 *                  vmn,   vmx,    med,   avr,     dsp);
 */
   free(hist);
}
