#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>
#include <time.h>

#include "tv_image_lib.h"

void main(int argc, char *argv[]) {
   short *img;                     /* image memory */
   int imgw, imgh;                 /* image size   */
   int nchan = 0;                  /* TV-chan. number */
   int col0, lin0, width, height;  /* frame: area of interest */
   int i,j,k,n,x,y,pix;
   int xc,yc;                      /* object center */
   double hw;
   double avr,spx,spy,s,dx,dy,fx,fy,f,lev,kf,r,sgm;
   double fw1 = 2.*sqrt(2.*log(2.));           /* FW for Gigma=1 */
   double r0 = 1.;
   double A = 1000.;
   double err = 0.;

   if (argc>1) nchan = atoi(argv[1]);
   srandom(time(0));
   get_vs56_sh_mem();
   img = tv_chan_image (nchan);
   imgw = tv_chan_img_width();
   imgh = tv_chan_img_height();
   tv_chan_get_frame(&col0, &lin0, &width, &height);

   if (argc>2) xc = atoi(argv[2]);
   else        xc = width/2;
   if (argc>3) yc = atoi(argv[3]);
   else        yc = height/2;
   xc += col0;
   yc += lin0;
   if (argc>4) hw = atoi(argv[4]);
   else        hw = (width+height)/8;
   if (argc>5) r0 = atof(argv[5]);
   if (argc>6) err = atof(argv[6]);
   lev = A / exp(r0*r0/2.);

   printf("Model: Xc=%d Yc=%d FW=%.1f\n",xc,yc,hw*fw1);

   for (k=-20; k<20; k += 1) {
      for (j=0; j<height; j++) {
	 y=lin0+j;
	 for (i=0; i<width; i++) {
	    double rx,ry,r2;
	    x=col0+i;
	    rx = (double)(x-xc-k);
	    ry = (double)(y-yc-k);
	    r2 = rx*rx + ry*ry;
	    pix = (int)(A / exp(r2/(2.*hw*hw)) + (double)random()/RAND_MAX*10. );
//          img[imgw*y+x] = (j<yc-lin0-hw/2||j>yc-lin0+hw/2)? pix : 0;
	    img[imgw*y+x] = pix;
	 }
      }
      printf("Model:  Xc=%d Yc=%d\n",xc+k,yc+k);
      tv_chan_set_seq_num(tv_chan_seq_num()+1);

      r0 += 0.1;
      lev = 1. / exp(r0*r0/2.);
      s=0.;
      n=0;
      for (j=0; j<height; j++) {
	 y=lin0+j;
	 dy = y-yc-k;
	 for (i=0; i<width; i++) {
	    double r;
	    x=col0+i;
	    dx = x-xc-k;
	    r = sqrt(dx*dx + dy*dy);
	    if (r > r0*hw) {
	       s += img[imgw*y+x];
	       n++;
	    }
	 }
      }
      avr = s/n;
      printf("Model: BGD=%.2f+%.2f Lev=%.2f+%.2f\n",avr,err,lev,avr);
      avr += err;
      spx=spy=s=0.;
      for (j=0; j<height; j++) {
	 y=lin0+j;
	 dy = y-yc-k;
	 for (i=0; i<width; i++) {
	    double r;
	    x=col0+i;
	    dx = x-xc-k;
	    r = sqrt(dx*dx + dy*dy);
	    if (r<r0*hw) {
	       pix = img[imgw*y+x] - avr;
	       spx += x*pix;
	       spy += y*pix;
	       s += pix;
	    }
	 }
      }
      printf("Model: XC=%.2f YC=%.2f\n",spx/s,spy/s);
      spx=spy=s=0.;
      for (j=0; j<height; j++) {
	 y=lin0+j;
	 dy = y-yc-k;
	 for (i=0; i<width; i++) {
	    double r;
	    x=col0+i;
	    dx = x-xc-k;
	    r = sqrt(dx*dx + dy*dy);
	    if (r<r0*hw) {
	       pix = img[imgw*y+x] - avr;
	       spx += dx*dx*pix;
	       spy += dy*dy*pix;
	       s += pix;
	    }
	 }
      }
      if (lev>0.7)
	 kf =  0.744*sqrt(1.-lev);
      else if (lev>0.3)
	 kf = 0.9144-0.72727*lev;
      else
	 kf = 1.1515-0.78945*sqrt(lev+0.03781);
      fx = sqrt(fabs(spx/s)) * fw1 / kf;
      fy = sqrt(fabs(spy/s)) * fw1 / kf;
      f = sqrt((fx*fx+fy*fy)/2.);
      printf("Model: FWX=%.2f FWY=%.2f FW=%.2f(%.1f)\n",fx, fy, f, 100*f/(hw*fw1));
/*fprintf(stderr,"%.3f %.3f %.3f\n", lev, f/(hw*fw1), kf);*/
      sgm = f/fw1;
      for (f=r=0;r<r0; r+=0.01) {
	 f += r*r*r/exp(r*r/2.)*0.01;
      }
      f += r0*r0*lev;
      f *= M_PI;
      printf("Model: Int = %.1f %.5f %.5f  Sigma=%.1f  A=%.1f\n",
		      s, s/A/(hw*hw), f,  sgm, s/f/(sgm*sgm));
      sleep(1);
      fflush(stdout);
   }
   exit(0);
}
