
/*
 *    FITS-
 *   TV- vs56/vs60/vsbt
 *     
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>

#include "tv_image_lib.h"  /* use VS56/VS60-tools to get images */
#include "bta_shdata.h"    /* use New ASC BTA connection  */

static char *time_asc0(double t, char *lin);
static char *time_asc2(double t, char *lin);
static char *angle_asc(double a, char *lin);

int main(int argc, char *argv[])
{
    short *img;                     /* VS image memory */
    int imgw, imgh;                 /* VS image size   */
    int nchan = 0;                  /* TV-chan. number */
    int nseq, nseq_o;
    char tvdrv[20] = "";
    char fline[100];
    int ppid = getppid();

    if(argc>1) {
       char *p = strchr(argv[1],':');
       if(p!=NULL || argv[1][0] == 'v' || argv[1][0] == 'V' || argv[1][0] == 'b') {
	  strncpy(tvdrv,argv[1],19);
	  if(p!=NULL) nchan = atoi(p+1);
       } else
	  nchan = atoi(argv[1]);
    }
    if(strncmp(tvdrv,"vs60",4)==0 || strncmp(tvdrv,"VS60",4)==0) get_vs60_sh_mem();
    else if(strncmp(tvdrv,"vsbt",4)==0 || strncmp(tvdrv,"bttv",4)==0) get_vsbt_sh_mem();
    else get_vs56_sh_mem();

    img = tv_chan_image (nchan);
    imgw = tv_chan_img_width();
    imgh = tv_chan_img_height();
    nseq = nseq_o = tv_chan_seq_num();

 /* First - common (unchanged) FITS-header lines */
    {int i;
     char *lines[] = {
     "OBSERVAT= 'SAO RAS '           / Observatory (Organization)                     ",
     "OBSERVER= 'SAO Observers'      / MIDAS desc.: OBSERVER(1) / observers team      ",
     "AUTHOR  = 'Unknown '           / MIDAS desc.: AUTHOR(1) / author of program     ",
     "PROG-ID = 'Unknown '           / observational program identifier               ",
     "OBJECT  = 'Unknown '           / MIDAS desc.: IDENT(1)  / object name           ",
     "MAG-VIS = '0.0     '           / object magnitude                               ",
     "SEEING  = '0.0     '           / seeing (\")                                        ",
     "CLOUDS  = '00      '           / clouds (%)                                     ",
     "TRANSP  = '00      '           / transparency (%)                               ",
     "WIND    = '00      '           / wind (m/s)                                     ",
     "TELESCOP= '6m      '           / MIDAS desc.: TELESCOP(1) / telescope name      ",
     NULL
     };
       for(i=0; i<36 && lines[i]!=NULL; i++) {
	   if(tv_chan_fits_line(lines[i]) == NULL)
	      tv_chan_set_fits_line(lines[i]);
       }
    }

 /* Next - variable (telescope) FITS-header lines */

    get_shm_block( &sdat, ClientSide);
  /*get_cmd_queue( &ucmd, ClientSide);*/

    while(ppid>1 && kill(ppid, 0) >= 0) {
       nseq = tv_chan_seq_num();
       if(nseq != nseq_o) {
	  int i;
	  char tmp[80];
	  nseq_o = nseq;
	  for(i=0;i<80;i++) fline[i]=' ';
	  sprintf(fline,"RA      =          %11.7f / MIDAS desc.: O_POS(1)                          ",CurAlpha*15./3600.);
	  tv_chan_set_fits_line(fline);
	  sprintf(fline,"DEC     =          %11.7f / MIDAS desc.: O_POS(2)                          ",CurDelta/3600.);
	  tv_chan_set_fits_line(fline);
	  sprintf(fline,"RA-OBS  = '%-11s'        / current right ascession                         ", time_asc2(CurAlpha,tmp));
	  tv_chan_set_fits_line(fline);
	  sprintf(fline,"DEC-OBS = '%-11s'        / current declination                             ", angle_asc(CurDelta,tmp));
	  tv_chan_set_fits_line(fline);
	  sprintf(fline,"AZIMUTH =            %9.4f / current telescope azimuth (degr)               ",val_A/3600.);
	  tv_chan_set_fits_line(fline);
	  sprintf(fline,"ZENDIST =              %7.4f / current telescope zenith distance (degr)       ",val_Z/3600.);
	  tv_chan_set_fits_line(fline);
	  sprintf(fline,"FOCUS   =               %6.2f / focus of telescope (mm)                        ",val_F);
	  tv_chan_set_fits_line(fline);
	  sprintf(fline,"PARANGLE=               %6.2f / parallactic angle  (degr)                      ",tag_P/3600.);
	  tv_chan_set_fits_line(fline);
	  sprintf(fline,"ROTANGLE=               %6.2f / field rotation angle / P2 position (degr)      ",val_P/3600.);
	  tv_chan_set_fits_line(fline);
	  sprintf(fline,"T-MIRROR=                %5.1f / mirror temperature (C)                         ",val_T3);
	  tv_chan_set_fits_line(fline);
	  sprintf(fline,"T-INSIDE=                %5.1f / indome temperature (C)                         ",val_T2);
	  tv_chan_set_fits_line(fline);
	  sprintf(fline,"T-OUTSID=                %5.1f / outside temperature (C)                        ",val_T1);
	  tv_chan_set_fits_line(fline);
	  sprintf(fline,"PRESSURE=                %5.1f / atmospheric pressure (mm)                      ",val_B);
	  tv_chan_set_fits_line(fline);
	  sprintf(fline,"WIND    = '%4.1f    '           / wind (m/s)                                     ",val_Wnd);
	  tv_chan_set_fits_line(fline);
	  sprintf(fline,"LMST    = '%-8s'           / local mean sidereal time                        ", time_asc0(S_time,tmp));
	  tv_chan_set_fits_line(fline);
       }
       my_sleep(0.2);
    }
}

static char *time_asc0(double t, char *lin)
{
    int h, min;
    int sec;
    h   = (int)(t/3600.);
    min = (int)((t - (double)h*3600.)/60.);
    sec = (int)(t - (double)h*3600. - (double)min*60.);
    h %= 24;
    sprintf(lin, "%02d:%02d:%02d", h,min,sec);
    return lin;
}
static char *time_asc2(double t, char *lin)
{
    int h, min;
    double sec;
    h   = (int)(t/3600.);
    min = (int)((t - (double)h*3600.)/60.);
    sec = t - (double)h*3600. - (double)min*60.;
    h %= 24;
    if(sec>59.99) sec=59.99;
    sprintf(lin, "%02d %02d %05.2f", h,min,sec);
    return lin;
}

static char *angle_asc(double a, char *lin)
{
    char s;
    int d, min;
    double sec;
    if (a >= 0.)
       s = '+';
    else {
       s = '-'; a = -a;
    }
    d   = (int)(a/3600.);
    min = (int)((a - (double)d*3600.)/60.);
    sec = a - (double)d*3600. - (double)min*60.;
    d %= 360;
    if(sec>59.9) sec=59.9;
    sprintf (lin, "%c%02d %02d %04.1f", s,d,min,sec);
    return lin;
}
