
/*     
 *     TV-
 *    VS56  VS60
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>
#include <time.h>
#include <errno.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>

#include <qobject.h>
#include <qwidget.h>
#include <qpoint.h>
#include <qmenubar.h>
#include <qpopmenu.h>
#include <qpushbt.h>
#include <qbttngrp.h>
#include <qradiobt.h>
#include <qlayout.h>
#include <qframe.h>
#include <qtabdlg.h>
#include <qpen.h>
#include <qpainter.h>
#include <qstring.h>
#include <qlabel.h>
#include <qlined.h>
#include <qchkbox.h>
#include <qpixmap.h>
#include <qkeycode.h>
#include <qimage.h>
#include <qwmatrix.h>
#include <qdatetm.h>
#include <qtimer.h>
#include <qslider.h>
#include <qcombo.h>
#include <qfiledlg.h>
#include <qtooltip.h>
#include <qapp.h>
#include <qmlined.h>

#include "tv_image_lib.h"
#define SHM_OLD_SIZE
#include "bta_shdata.h"    /* use New ASC connection  */

static bool mydebug = FALSE;
#define IFDBG(S) if(mydebug){S}

struct ParamList {
   char name[16];
   char val[100];
};

class BTAserv : public QObject
{
   Q_OBJECT
public:
   BTAserv ( QObject * parent=0, const char * name=0 );
   ~BTAserv ();
   bool connected();
   int write_str(char *);
   int read_str(char *, int);
   char *Get(const char *name);
   bool Put(const char *name, const char *value);
   bool ACSconnected();
   char *time_asc(double t, char *lin);
   char *angle_asc(double a, char *lin);
   double asc_time(char *str);
   double asc_angle(char *str);
   double mtime;
   bool new_acs, old_acs;

private slots:
   bool GetList();

private:
   QTimer *timer;
   int write_blk(char *, int);
   int read_blk(char *, int);
   const char *serv_name;
   int ns;
   struct ParamList  par_list[20];
   void IniList();
   char *Find(const char *name);
};

struct GDbuff_val {
   int time;
   double da;
   double dz;
};

class GDgraph : public QFrame
{
    Q_OBJECT
public:
    GDgraph ( QWidget *parent=0, const char *name=0 );
    ~GDgraph();
    void next_center(double da, double dz);
    void next_corrtn(double ca, double cz);
    void next_bgdxy(double bx, double by);
    void next_fwhmxy(double fwx, double fwx);
    void next_fwbr(double fw,   double br);
    char savedir[100];
    void flush_save();
    void set_head(char *name, char *val);
    void set_head(char *name, int ival);
    void set_head(char *name, double fval);
    char *get_head(const char *name, char val[]);
    BTAserv *bta; // connection to bta_serv

public slots:
    void update();

protected:
    //void  resizeEvent( QResizeEvent * );
    void  paintEvent( QPaintEvent * );

private slots:
    void cyclic_update();
    void set_t_scale(int scr_val);
    void set_v_scale(int scr_val);
    void set_type(int new_type);
    void clear_buf();
    void open_save();
    void open_stdout();
    void close_save();

private:
#define GD_BUFF_SIZE 3600
    int graph_time_elapsed();   // fix 24-hour bag in QTime::elapsed()
    int daz_pnt;
    struct GDbuff_val bdaz[GD_BUFF_SIZE];
    double da_cur,dz_cur;
    int caz_pnt;
    struct GDbuff_val bbxy[GD_BUFF_SIZE];
    double bx_cur,by_cur;
    int bxy_pnt;
    struct GDbuff_val bcaz[GD_BUFF_SIZE];
    double ca_cur,cz_cur;
    int fwxy_pnt;
    struct GDbuff_val bfwxy[GD_BUFF_SIZE];
    double fx_cur,fy_cur;
    int fwbr_pnt;
    struct GDbuff_val bfwbr[GD_BUFF_SIZE];
    double fw_cur,br_cur;
    double time_cur;
    double time_open;
    double time_head;
    struct ParamList  head[30];
    FILE *save_fd;
    void open_file(char *savename);
    void init_head();
    void print_head();
    void print_acs();

    int type;
#define DAZ   0
#define CORAZ 1
#define BGDXY 2
#define FWXY  3
#define FWBR  4
    QComboBox * cbtype;
    void draw_graph( int y0,  bool az);

    QLabel *sfile;
    QPainter  lin;
    QTime grtime;
    QTimer *gr_timer;
    bool need_update;
    void update(double sdt);

    double t_scale;
    double v_scale;
};

#define TVguideWidthMin 470
#define TVguideWidthMax 750
#define TVguideHeight 370
class TVguide : public QWidget
{
    Q_OBJECT
public:
    TVguide( QWidget *parent=0, const char *name=0 );
    ~TVguide();
    int n_chan;                            // TV-channel number
    char tv_drv[10];                       // TV-grabber driver name
    void set_corr(double dx, double dy);   //set coor.values: dx/dy->da/dz
    void get_corr(double *da, double *dz); //get current coor.values: da/dz
    bool tvguide_busy();                   // return TRUE when TVguide working
    short *tv_image;
    BTAserv *bta; // connection to bta_serv
    GDgraph *graph;
    double ha;    // "hardware" angle (degr.)
		  //   (P2-angle when hardware was mounted at "zero"-position)
    double sx;    // TV X-scale ("/pixel)
    double sy;    // TV Y-scale ("/pixel)
    int flipx;    // horizontal flip
    int flipy;    // vertical flip
    int flipa;    // revers A-axis
    int flipz;    // revers Z-axis

    double aza;   // screen A/Z axis angle (rad)
    double ada;   // screen Alp/Del axis angle (rad)

    void show_msg_win();
    void add_msg(char *str);

public slots:
    void clear_msg();

signals:
    void tvframe_ok();   //stop images accumulation, needs centering ...
    void new_image();    //new data for GDimage, repaint needs ...

private slots:
    void set_max_size();
    void set_min_size();
    void start_tv_add(); // start tv-image accumulation process
    void start_corr();
    void destr_opt();    // fix that "Options" dialog window was destroyed
    void apply_opt();    // Use new (edited) options
    void toggle_send();  // Toggle "Really Sending Correction" mode
    void toggle_get();   // Toggle "Getting ACS BTA parameters" mode
    void toggle_flipx(); // Toggle horizontal flipping
    void toggle_flipy(); // Toggle vertical flipping
    void setup_opt();    // Show "Options" dialog panel
    void get_acs_opt();  // Get values from ACS BTA for "Options" dialog
    void tvguide_start();// start auto-guiding process
    void tvguide_stop(); // stop guiding process
    void tvguide_once(); // start one guiding action
    void acs_io_tmout();
    void tvimage_ok();
    void acs_corr_ok();
    void set_passwd();   // get & match password
    void get_tv_conf();  // get TV-config. file name and load it
    void toggle_bgd_shift(); // Toggle "Determine backgr.shift" mode
    void toggle_bgd_ushift(); // Toggle "Use backgr.shift for coord." mode
    void toggle_bgd_sub();   // Toggle "Do background subtraction" mode
    void new_bgd_frame();    // reload new backgr.frame
    void zero_bgd_shift();   // set zero bgd.shift (bgd_x=bgd_y=0)
    void use_bgd_frame();  // dummy | get backgr.shift | subtruct backgr.frame
    void find_bgd_prog();  // Choose programm for background operations
    void no_bgd_prog();    // set NO background operations at all
    void find_bgd_file();  // Choose background file name
    void no_bgd_file();    // no file, use TV-frame as a background


private:
    int status;        // current TVguide state
#define GD_IDLE   0
#define ACS_IO    1
#define WAIT_CORR 2
#define ADD_IMAGE 3
    int set_status(int new_stat);    // change current TVguide state
    QLabel *stat_lab;

    int modes;  // ORed flags: auto-cycle or once; send corr-n to ACS or not
#define GD_STOP    0  // No guiding
#define GD_RECYCLE 1  // Auto-cyclling guiding process
#define SEND_CORR  2  // Really Sending Correction
#define ACS_GET    4  // Getting ACS BTA parameters

    void get_from_acs();  // Get following values from ACS BTA:
    int    nfoc;   // BTA curr.focus number: 0-Prime 1-Nasmyth1 2-Nasmyth2
    double pa;    // if nfoc==0 - curr. calculated P-angle (degr.)
    double p2;    // if nfoc==0 - curr. P2 position        (degr.)
    double zd;    // curr. Zenith distance        (degr.)

    int col0,lin0;    // frame position
    int fr_w,fr_h;    // frame size

    void turn_axes();                      // calculate A/Z & Alp/Del angles
    void xy_to_az(double dx, double dy, double *da, double *dz); //convert coor.values from sceen to telescope
    void put_corr(double da, double dz);   // Send A/Z correction codes to ACS BTA
    void put_azshift(double da, double dz);  // Send A/Z-coord-s shift to ACS BTA

    double cor_x;
    double cor_y;
    double cor_a;
    double cor_z;

    int nseq;
    int mseq;
    int add_min;         // minimum number of tv-images to add
    double corr_fact;    // Multiplier of A/Z corr.values before sending
    double corr_delay;   // additional delay time ( now ~10-12sec )

    double *obj_lev;   // -> GDimage::obj.detec. thresh-hold level (in dispersions)
    int *obj_pnt;      // -> GDimage::obj.detec. "neighbor points" number (0-8)
    double *obj_sqr;   // -> GDimage::obj. minimum whole square (in SquareArcSec)

    QBoxLayout *tvl;
    QFrame     *td_misc;
    QTabDialog *td_opt;
    QLineEdit *ed_ha;
    QLineEdit *ed_sx;
    QLineEdit *ed_sy;
    QCheckBox *cb_fx;
    QCheckBox *cb_fy;
    QLineEdit *ed_pa;
    QLineEdit *ed_p2;
    QLineEdit *ed_zd;
    QLineEdit *ed_nf;
    QLineEdit *ed_mn;
    QLineEdit *ed_tm;
    QLineEdit *ed_dt;
    QCheckBox *cb_snd;
    QCheckBox *cb_acs;
    QLineEdit *ed_lv;
    QLineEdit *ed_np;
    QLineEdit *ed_sq;
    QLineEdit *ed_bp;
    QLineEdit *ed_bf;
    QCheckBox *cb_bc;
    QCheckBox *cb_su;
    QCheckBox *cb_bs;

    QTimer *acs_timer;
    QTimer *add_timer;
    QTimer *cor_timer;

    void start_bgd_proc(); // (re)start bgd-maint. process
    void check_bgd_proc();
    void stop_bgd_proc();  // stop bgd-maint. process
    int bgd_pid;           // bgd-proces PID
    int  pipe0[2], pipe1[2];
    FILE *pipe_in, *pipe_out; // i/o pipe to bgd-process
    char *bgd_proc;        // programm for background operations
    char *bgd_file;        // background FITS-file
    char *bgd_dir;         // curr. background files directory
    bool bgd_shift;        // determine backgr.shift
    bool bgd_ushift;       // use backgr.shift for A/Z calculation
    bool bgd_sub;          // do background subtraction
    double bgd_x, bgd_y;   // last backgr.shift values
    int bgd_col0,bgd_lin0,bgd_frmw,bgd_frmh;// last background frame

    void show_psw_win(); // setup or show password window
    char *passwd;
    //QWidget *wpsw;
    QFrame *wpsw;
    QLineEdit *ed_psw;

    QMultiLineEdit *wmesg;

    bool load_tv_conf(char *confname);  // load parameters from TV-config. file
    bool save_tv_conf(char *confname);  // save parameters to TV-config. file
    char confdir[100];
    char dfltfile[80];
};

#define TV_FRAME_MAX_WIDTH  700
#define TV_FRAME_MAX_HEIGHT 500
#define GDimageMaxWidth  300
#define GDimageMaxHeight 300

class GDimage : public QFrame
{
    Q_OBJECT
public:
    GDimage( QWidget *parent=0, const char *name=0 );
    ~GDimage();
    short *tv_image;             // image memory
    TVguide *tvg;
    GDgraph *graph;
    double obj_lev;   // object detection thresh-hold level (in dispersions)
    int obj_pnt;      // object detection "neighbor points" number (0-8)
    double obj_sqr;   // object minimum whole square (in SquareArcSec)

public slots:
    void update();

private slots:
    void cyclic_update();
    void set_x_pos(int ix);
    void set_y_pos(int iy);
    void set_xy_zoom(int i);
    void set_image_source(int nfrm);
    void get_tv_image();
    void fix_center(); // fix current coordinates as a new reference center coordinates

protected:
    void   resizeEvent( QResizeEvent * );
    void  paintEvent( QPaintEvent * );
    void  mousePressEvent( QMouseEvent * );
    //void  mouseReleaseEvent( QMouseEvent * );
    //void  mouseMoveEvent( QMouseEvent * );

private:
    QPixmap tv_frame_pm;          // pixmap for frame image
    QPixmap tv_pm;                // pixmap after zoom
    QPainter  lin;
    QWMatrix  mtx,rmtx;           // direct & revers image transformations
    int im_w;                     // "on screen" guide image size
    int im_h;
    int im_src;                   // get image from: 0-Frame 1-SubFrame1 2-SubFrame2

    int tvim_w;                   // tv_image size
    int tvim_h;
    int col0;                     // frame position
    int lin0;
    int fr_w;                     // frame size
    int fr_h;
    int x_c;                      // "telescope" center
    int y_c;
    int mouse_x;                  // "clicked" point
    int mouse_y;
    char *mask;                   // object detection "neighborhood" mask
    QButtonGroup *gbt;
    QRadioButton *rbt[3];
    bool need_update;
    void update(double sdt);
    QTime gdtime;
    QTimer *gd_timer;

    void stat_tv_image();     // do some statistics on tv_image frame
    int hist[6554];           // histogram array 6554 ~= 2^16 / 10
    int lev1,lev2,lev3;
    int vmin,vmax,med;        // minimum, maximum, median
    double avr, dsp;          // average level and dispersion

    bool cent_tv_image();     // calc. center coordinates at tv_image frame
    double xc,yc;
    double fwhm_x,fwhm_y;
    double *itab;             // table for integrals of gaussian (r=0->10 step 0.01)
};

TVguide::TVguide( QWidget *parent, const char *name )
    : QWidget( parent, name)
{
    char cap[40],tvdrv[20];
    int nc=0;
    int ac=qApp->argc();
    char **av=qApp->argv();

    if(ac>1) {
       char *p = strchr(av[1],':');
       if(p!=NULL || av[1][0] == 'v' || av[1][0] == 'V' || av[1][0] == 'b') {
	  strncpy(tvdrv,av[1],19);
	  if(p!=NULL) nc = atoi(p+1);
       } else
	  nc = atoi(av[1]);
    }
    if(ac>2) {
       if(av[2][0] == 'd')
	  mydebug = TRUE;
       else
	  strncpy(tvdrv,av[2],19);
    }
    if(strncmp(tvdrv,"vs56",4)==0 || strncmp(tvdrv,"VS56",4)==0) get_vs56_sh_mem();
    else if(strncmp(tvdrv,"vs60",4)==0 || strncmp(tvdrv,"VS60",4)==0) get_vs60_sh_mem();
    else if(strncmp(tvdrv,"vsbt",4)==0 || strncmp(tvdrv,"bttv",4)==0) get_vsbt_sh_mem();
    else get_vs60_sh_mem();

    if(vs_type()<800)
       sprintf(tv_drv,"vs%d:%d",vs_type(),nc);
    else
       sprintf(tv_drv,"vsbt:%d",nc);
IFDBG( printf("TVguide for %s\n",tv_drv); )

    n_chan = nc;
    tv_image = tv_chan_image (nc);

    sprintf(cap,"TVguide Chan.%d", nc);
    setCaption(cap);

    bta = NULL;
    td_opt = NULL;
    wmesg = NULL;
    wpsw = NULL;
    passwd = NULL;
    nfoc = 0;
    pa = 0.0;
    p2 = 0.0;
    zd = 80.0;
    aza = (p2 - ha) * M_PI / 180.;
    ada = ((p2 - ha) - pa) * M_PI / 180.;
    flipx = 0;
    flipy = 0;
    flipa = 0;
    flipz = 0;
    cor_x = 0.0;
    cor_y = 0.0;
    cor_a = 0.0;
    cor_z = 0.0;
    status = GD_IDLE;
    nseq = tv_chan_seq_num();
    mseq = tv_chan_seq_msg();
    add_min = 1;
    corr_fact = 0.5;
    corr_delay = 10.0;
    modes =  GD_STOP;
    bgd_pid = -1;
    pipe_in = pipe_out = NULL;
    bgd_proc = NULL;
    bgd_file = NULL;
    bgd_dir = ".";
    bgd_shift = FALSE;
    bgd_ushift = FALSE;
    bgd_sub = FALSE;
    bgd_x = 0.;
    bgd_y = 0.;

    if ( access("./Conf",R_OK|X_OK) == 0 )
       strcpy(dfltfile,"./Conf");
    else if ( access("./TVconf",R_OK|X_OK) == 0 )
       strcpy(dfltfile,"./TVconf");
    else
       strcpy(dfltfile,".");
    strcpy(confdir,dfltfile);
    strcat(dfltfile,"/tv_default.conf");

    tvl  =  new QHBoxLayout(this, 5);
    QBoxLayout *ll =  new QVBoxLayout(2);
    tvl->addLayout(ll);

    QFrame *far = new QFrame (this, "image_frame");
    far->setFrameStyle(QFrame::Box|QFrame::Raised);
    far->setFixedSize( 240, 240 );
    far->setLineWidth( 2 );
    ll->addWidget(far, 1, AlignLeft);
    GDimage *ar = new GDimage (far, "gd_image");
    ar->setGeometry( 5, 5, 230, 230 );
    ar->tvg=this;
    ar->tv_image=tv_image;
    obj_lev = &(ar->obj_lev);
    obj_pnt = &(ar->obj_pnt);
    obj_sqr = &(ar->obj_sqr);

    QBoxLayout *lll =  new QHBoxLayout(2);
    ll->addLayout(lll);
    QLabel * lb = new QLabel( this );
    lll->addWidget( lb );
    lb->setText( "Status:" );
    lb->setMinimumSize( 60, 15 );
    QToolTip::add(lb, "What TVguide doing now" );

    stat_lab =  new QLabel( this );
    lll->addWidget( stat_lab, 1, AlignLeft);
    stat_lab->setFrameStyle( QFrame::Box | QFrame::Raised );
    stat_lab->setText( "  Idle" );
    stat_lab->setMinimumSize( 100, 20 );
    stat_lab->setMaximumSize( 180, 30 );
    QToolTip::add(stat_lab, "Nothing\n\nAccumulate TV-images\n\nIn/Output with ASC BTA\n\nWait for Correction completion\n" );

    QBoxLayout *llb =  new QHBoxLayout(2);
    ll->addLayout(llb);

    lb = new QLabel( this );
    llb->addWidget( lb );
    lb->setText( "Guiding:" );
    lb->setMinimumSize( 60, 15 );
    QToolTip::add(lb, "Accumulate images,  Calc.Center and (optionally) Send Correction to ACS " );

    QPushButton* bst = new QPushButton( "Start", this );
    bst->setMinimumSize( bst->sizeHint() );
    bst->setMaximumWidth(60);
    QToolTip::add(bst, "Start Guiding cycling process" );
    connect( bst, SIGNAL(clicked()), SLOT(tvguide_start()) );
    llb->addWidget(bst, 1, AlignLeft);
    lb->setBuddy( bst );
    bst = new QPushButton( "Stop", this );
    bst->setMinimumSize( bst->sizeHint() );
    bst->setMaximumWidth(60);
    connect( bst, SIGNAL(clicked()), SLOT(tvguide_stop()) );
    QToolTip::add(bst, "Stop Guiding process" );
    llb->addWidget(bst, 1, AlignLeft);
    bst = new QPushButton( "Once", this );
    bst->setMinimumSize( bst->sizeHint() );
    bst->setMaximumWidth(60);
    connect( bst, SIGNAL(clicked()), SLOT(tvguide_once()) );
    QToolTip::add(bst, "Do single Guiding action" );
    llb->addWidget(bst, 1, AlignLeft);

    llb =  new QHBoxLayout(2);
    ll->addLayout(llb);
    lb = new QLabel( this );
    llb->addWidget( lb );
    lb->setText( "Manual:" );
    lb->setMinimumSize( 60, 15 );
    QToolTip::add(lb, "Calculate TV-image center or Send correction manually" );

    QPushButton* btv = new QPushButton( "Center", this );
    btv->setMinimumSize( btv->sizeHint() );
    btv->setMaximumWidth(70);
    connect( btv, SIGNAL(clicked()), SLOT(clear_msg()) );
    connect( btv, SIGNAL(clicked()), ar, SLOT(get_tv_image()) );
    QToolTip::add(btv, "Calculate TV-image frame center" );
    QPushButton* bfix = new QPushButton( "Fix", this );
    llb->addWidget(btv, 1, AlignLeft);
    bfix->setMinimumSize( bfix->sizeHint() );
    bfix->setMaximumWidth(40);
    connect( bfix, SIGNAL(clicked()), ar, SLOT(fix_center()) );
    QToolTip::add(bfix, "Fix current coordinates as a new reference center coordinates" );
    llb->addWidget(bfix, 1, AlignLeft);
    QPushButton* bcor = new QPushButton( "Correct.", this );
    bcor->setMinimumSize( bcor->sizeHint() );
    bcor->setMaximumWidth(70);
    connect( bcor, SIGNAL(clicked()), SLOT(clear_msg()) );
    connect( bcor, SIGNAL(clicked()), SLOT(start_corr()) );
    QToolTip::add(bcor, "Send A/Z-correction to ACS BTA" );
    llb->addWidget(bcor, 1, AlignLeft);

    llb =  new QHBoxLayout(2);
    ll->addLayout(llb);
    lb = new QLabel( this );
    llb->addWidget( lb );
    lb->setText( "Misc:" );
    lb->setMinimumSize( 60, 15 );
    QToolTip::add(lb, "Other (nonGuiding) actions" );

    QPushButton* bq = new QPushButton( "Quit", this );
    bq->setMinimumSize( bq->sizeHint() );
    bq->setMaximumWidth(90);
    connect( bq, SIGNAL(clicked()), qApp, SLOT(quit()) );
    QToolTip::add(bq, "Stop TVguide utility; GoodBye!" );
    llb->addWidget(bq, 1, AlignLeft);
    QPushButton* bopt = new QPushButton( "Options ==>", this );
    bopt->setMinimumSize( bopt->sizeHint() );
    bopt->setMaximumWidth(90);
    connect( bopt, SIGNAL(clicked()), SLOT(setup_opt()) );
    QToolTip::add(bopt, "Show 'TV-Guide Options' window" );
    llb->addWidget(bopt, 1, AlignLeft);

    graph = new GDgraph(this,"gd_graph");
    graph->setGeometry( 0, 0, 220, TVguideHeight );
    graph->setMinimumWidth(210);
    QToolTip::add(graph, "A/Z center position graphics" );
    tvl->addWidget(graph, 1, AlignRight);
    ar->graph = graph;

    acs_timer = new QTimer(this);
    connect( acs_timer, SIGNAL(timeout()), this, SLOT(acs_io_tmout()) );
    add_timer = new QTimer(this);
    connect( add_timer, SIGNAL(timeout()), this, SLOT(tvimage_ok()) );
    cor_timer = new QTimer(this);
    connect( cor_timer, SIGNAL(timeout()), this, SLOT(acs_corr_ok()) );

    connect( this, SIGNAL(tvframe_ok()), ar, SLOT(get_tv_image()) );
    connect( this, SIGNAL(new_image()), ar, SLOT(update()) );

    td_misc = new QFrame (this, "opt+msg");
    td_misc->setGeometry( 0, 0, 280, TVguideHeight );
    td_misc->setMinimumWidth(280);
    tvl->addWidget(td_misc);
    QPushButton* bmin = new QPushButton( "<== Minimize", td_misc );
    bmin->setGeometry( 0, TVguideHeight-40, 280, 30 );
    connect( bmin, SIGNAL(clicked()), SLOT(set_min_size()) );

    set_min_size();

    show();

    setup_opt();
    load_tv_conf(dfltfile);
    //td_opt->hide();
    //start_bgd_proc();  // try to start bgd-maint. process

}

TVguide::~TVguide()
{
    save_tv_conf(dfltfile);
fputs("TVguide delete\n",stderr);
    tv_image = NULL;
    if(pipe_in!=NULL)
       stop_bgd_proc();  // kill bgd-maint. process
    if(bgd_pid>0) {
       kill(bgd_pid,15);
       my_sleep(0.5);
       if(waitpid(-1, &status, WNOHANG) == 0)
	  fprintf(stderr,"Can't kill BGD-process!?");
    }
}

void TVguide::set_max_size() {
    if(width() < TVguideWidthMax)
       resize( TVguideWidthMax, TVguideHeight);
}
void TVguide::set_min_size() {
       resize( TVguideWidthMin, TVguideHeight);
}

void TVguide::setup_opt()  // Show "Options" dialog panel
{
    if (td_opt != NULL) {
       set_max_size();
       if(td_opt->isVisible())
	  td_opt->raise();
       else {
	  //td_opt->setGeometry( x() + width() - 200, y(), 200, 200 );
	  td_opt->show();
       }
       return;
    }
    td_opt = new QTabDialog( td_misc, "opt_dialog" );
    connect( td_opt, SIGNAL(destroyed()), SLOT(destr_opt()) );
    //QWidget * bopt = (QWidget *)sender();
    //td_opt->setGeometry( x() + bopt->x(), y() + bopt->y()+10, 200, 200 );
    //td_opt->setGeometry( x() + width() - 200, y(), 200, 200 );
    //td_opt->setCaption( "TVguide Options" );
    td_opt->setGeometry( 0, 190, 280, 180 );

    // set up 1-st page  of the tab dialog
    QWidget * w = new QWidget( td_opt, "screen" );
    QGridLayout * g = new QGridLayout( w, 4, 2, 5 );

    ed_ha = new QLineEdit( w );
    g->addWidget( ed_ha, 0, 1 );
    char str[100];
    sprintf(str,"%.2f",ha);
    ed_ha->setText( str );
    ed_ha->setMinimumSize( 30, 10 );
    ed_ha->setMaximumSize( 100, 30 );
    QToolTip::add(ed_ha, " \"Hardware\" position angle (degr.)\n E.g. P2-angle when hardware was mounted at \"zero\"-position " );
    QLabel * lb = new QLabel( w );
    g->addWidget( lb, 0, 0 );
    lb->setText( "Pos.angle:" );
    lb->setBuddy( ed_ha );
    lb->setMinimumSize( 30, 10 );
    lb->setMaximumSize( 100, 20 );
    QToolTip::add(lb, " \"Hardware\" position angle (degr.)\n E.g. P2-angle when hardware was mounted at \"zero\"-position " );

    ed_sx = new QLineEdit( w );
    g->addWidget( ed_sx, 1, 1 );
    sprintf(str,"%.2f",sx);
    ed_sx->setText( str );
    ed_sx->setMinimumSize( 30, 10 );
    ed_sx->setMaximumSize( 100, 30 );
    QToolTip::add(ed_sx,"TV X-scale (ArcSec/pixel)");
    lb = new QLabel( w );
    g->addWidget( lb, 1, 0 );
    lb->setText( "X-scale:" );
    lb->setBuddy( ed_sx );
    lb->setMinimumSize( 30, 10 );
    lb->setMaximumSize( 100, 20 );
    QToolTip::add(lb,"TV X-scale (ArcSec/pixel)");

    ed_sy = new QLineEdit( w );
    g->addWidget( ed_sy, 2, 1 );
    sprintf(str,"%.2f",sy);
    ed_sy->setText( str );
    ed_sy->setMinimumSize( 30, 10 );
    ed_sy->setMaximumSize( 100, 30 );
    QToolTip::add(ed_sy,"TV Y-scale (ArcSec/pixel)");
    lb = new QLabel( w );
    g->addWidget( lb, 2, 0 );
    lb->setText( "Y-scale:" );
    lb->setBuddy( ed_sy );
    lb->setMinimumSize( 30, 10 );
    lb->setMaximumSize( 100, 20 );
    QToolTip::add(lb,"TV Y-scale (ArcSec/pixel)");

    cb_fx = new QCheckBox( w );
    g->addWidget( cb_fx, 3, 0 );
    if (flipx)
       cb_fx->setText("Flip on X");
    else
       cb_fx->setText("No X flip");
    cb_fx->setMinimumSize( 30, 15 );
    cb_fx->setMaximumSize( 100, 30 );
    connect( cb_fx, SIGNAL(clicked()), SLOT(toggle_flipx()) );
    QToolTip::add(cb_fx, "Toggle horizontal flip" );
    cb_fy = new QCheckBox( w );
    g->addWidget( cb_fy, 3, 1 );
    if (flipy)
       cb_fy->setText("Flip on Y");
    else
       cb_fy->setText("No Y flip");
    cb_fy->setMinimumSize( 30, 15 );
    cb_fy->setMaximumSize( 100, 30 );
    connect( cb_fy, SIGNAL(clicked()), SLOT(toggle_flipy()) );
    QToolTip::add(cb_fy, "Toggle vertical flip" );

    // no stretch on the labels unless they have to
    g->setColStretch( 0, 0 );

    // finally insert page "screen" into the tab dialog and start GM
    td_opt->addTab( w, "&Screen" );
    g->activate();

    // set up 2-nd page  of the tab dialog
    w = new QWidget( td_opt, "acs" );
    g = new QGridLayout( w, 4, 3, 5 );

    ed_pa = new QLineEdit( w );
    g->addWidget( ed_pa, 0, 1 );
    sprintf(str,"%.2f",pa);
    ed_pa->setText( str );
    ed_pa->setMinimumSize( 30, 15 );
    ed_pa->setMaximumSize( 700, 30 );
    QToolTip::add(ed_pa,"Calculated P-angle (between Alp/Del and A/Z local sky axes)");
    lb = new QLabel( w );
    g->addWidget( lb, 0, 0 );
    lb->setText( "P-angle:" );
    lb->setBuddy( ed_pa );
    lb->setMinimumSize( 30, 10 );
    lb->setMaximumSize( 700, 20 );
    QToolTip::add(lb,"Calculated P-angle (between Alp/Del and A/Z local sky axes)");

    ed_p2 = new QLineEdit( w );
    g->addWidget( ed_p2, 1, 1 );
    sprintf(str,"%.2f",p2);
    ed_p2->setText( str );
    ed_p2->setMinimumSize( 30, 15 );
    ed_p2->setMaximumSize( 70, 30 );
    QToolTip::add(ed_p2,"Real P2-table position");
    lb = new QLabel( w );
    g->addWidget( lb, 1, 0 );
    lb->setText( "P2-pos:" );
    lb->setBuddy( ed_p2 );
    lb->setMinimumSize( 30, 10 );
    lb->setMaximumSize( 70, 20 );
    QToolTip::add(lb,"Real P2-table position");

    ed_zd = new QLineEdit( w );
    g->addWidget( ed_zd, 2, 1 );
    sprintf(str,"%.2f",zd);
    ed_zd->setText( str );
    ed_zd->setMinimumSize( 30, 15 );
    ed_zd->setMaximumSize( 70, 30 );
    QToolTip::add(ed_zd,"Current Zenith distance of telescope");
    lb = new QLabel( w );
    g->addWidget( lb, 2, 0 );
    lb->setText( "Z-dist:" );
    lb->setBuddy( ed_zd );
    lb->setMinimumSize( 30, 15 );
    lb->setMaximumSize( 70, 20 );
    QToolTip::add(lb,"Current Zenith distance of telescope");

    ed_nf = new QLineEdit( w );
    g->addWidget( ed_nf, 3, 1 );
    sprintf(str,"%d  - %s", nfoc, (nfoc==2)? "N2": ((nfoc==1)? "N1" : "CPF"));
    ed_nf->setText( str );
    ed_nf->setMinimumSize( 30, 15 );
    ed_nf->setMaximumSize( 70, 30 );
    QToolTip::add(ed_nf,"0 - CPF   1 - N1   2 - N2");
    lb = new QLabel( w );
    g->addWidget( lb, 3, 0 );
    lb->setText( "Tel.Focus:" );
    lb->setBuddy( ed_nf );
    lb->setMinimumSize( 30, 10 );
    lb->setMaximumSize( 70, 20 );
    QToolTip::add(lb,"Telescope focus used");

    QPushButton* bt = new QPushButton( "ACS get", w );
    g->addWidget( bt, 3, 2 );
    bt->setMaximumHeight(25);
    bt->setMaximumWidth(60);
    QToolTip::add(bt, "Get BTA parameters from ACS (once)" );
    connect( bt, SIGNAL(clicked()), SLOT(get_acs_opt()) );

    // finally insert page "acs" into the tab dialog and start GM
    td_opt->addTab( w, "&BTA" );
    g->activate();

    // set up 3-ed page  of the tab dialog
    w = new QWidget( td_opt, "guide" );
    g = new QGridLayout( w, 4, 2, 5 );

    ed_mn = new QLineEdit( w );
    g->addWidget( ed_mn, 0, 1 );
    sprintf(str,"%d",add_min);
    ed_mn->setText( str );
    ed_mn->setMinimumSize( 30, 15 );
    ed_mn->setMaximumSize( 100, 30 );
    QToolTip::add(ed_mn, "Number of TV-images to ADD" );
    lb = new QLabel( w );
    g->addWidget( lb, 0, 0 );
    lb->setText( "N add:" );
    lb->setBuddy( ed_mn );
    lb->setMinimumSize( 30, 10 );
    lb->setMaximumSize( 100, 20 );
    QToolTip::add(lb, "Number of TV-images to ADD" );

    ed_tm = new QLineEdit( w );
    g->addWidget( ed_tm, 1, 1 );
    sprintf(str,"%.2f",corr_fact);
    ed_tm->setText( str );
    ed_tm->setMinimumSize( 30, 15 );
    ed_tm->setMaximumSize( 100, 30 );
    QToolTip::add(ed_tm, "Multiplier of A/Z correction values before sending" );
    lb = new QLabel( w );
    g->addWidget( lb, 1, 0 );
    lb->setText( "Corr.Factor:" );
    lb->setBuddy( ed_tm );
    lb->setMinimumSize( 30, 10 );
    lb->setMaximumSize( 100, 20 );
    QToolTip::add(lb, "Multiplier of A/Z correction values before sending" );

    ed_dt = new QLineEdit( w );
    g->addWidget( ed_dt, 2, 1 );
    sprintf(str,"%.1f",corr_delay);
    ed_dt->setText( str );
    ed_dt->setMinimumSize( 30, 15 );
    ed_dt->setMaximumSize( 100, 30 );
    QToolTip::add(ed_dt, "Additional delay (sec) after correction" );
    lb = new QLabel( w );
    g->addWidget( lb, 2, 0 );
    lb->setText( "Corr.delay:" );
    lb->setBuddy( ed_dt );
    lb->setMinimumSize( 30, 10 );
    lb->setMaximumSize( 100, 20 );
    QToolTip::add(lb, "Additional delay (sec) after correction" );

    cb_snd = new QCheckBox( w );
    g->addWidget( cb_snd, 3, 0 );
    if (modes & SEND_CORR)
       cb_snd->setText("Send Corr.");
    else
       cb_snd->setText("Don't Send");
    cb_snd->setMinimumSize( 30, 15 );
    cb_snd->setMaximumSize( 100, 30 );
    connect( cb_snd, SIGNAL(clicked()), SLOT(toggle_send()) );
    QToolTip::add(cb_snd, "Toggle \"Really Sending Correction\" mode" );
    cb_acs = new QCheckBox( w );
    g->addWidget( cb_acs, 3, 1 );
    if (modes & ACS_GET) {
       cb_acs->setText("Get ASC");
       cb_acs->setChecked(TRUE);
    } else {
       cb_acs->setChecked(FALSE);
       cb_acs->setText("No ACS");
    }
    cb_acs->setMinimumSize( 30, 15 );
    cb_acs->setMaximumSize( 100, 30 );
    connect( cb_acs, SIGNAL(clicked()), SLOT(toggle_get()) );
    QToolTip::add(cb_acs, "Toggle \"Getting ACS BTA parameters\" mode" );

    // finally insert page "guide" into the tab dialog and start GM
    td_opt->addTab( w, "&Guide" );
    g->activate();

    // set up 4-th page  of the tab dialog
    w = new QWidget( td_opt, "center" );
    g = new QGridLayout( w, 4, 2, 5 );

    ed_lv = new QLineEdit( w );
    g->addWidget( ed_lv, 0, 1 );
    sprintf(str,"%.1f", *obj_lev);
    ed_lv->setText( str );
    ed_lv->setMinimumSize( 30, 15 );
    ed_lv->setMaximumSize( 100, 30 );
    QToolTip::add(ed_lv, "Object detection thresh-hold level (in \"sigma\" units)" );
    lb = new QLabel( w );
    g->addWidget( lb, 0, 0 );
    lb->setText( "Obj. level:" );
    lb->setBuddy( ed_lv );
    lb->setMinimumSize( 30, 10 );
    lb->setMaximumSize( 100, 20 );
    QToolTip::add(lb, "Object detection thresh-hold level (in \"sigma\" units)" );

    ed_np = new QLineEdit( w );
    g->addWidget( ed_np, 1, 1 );
    sprintf(str,"%d",*obj_pnt);
    ed_np->setText( str );
    ed_np->setMinimumSize( 30, 15 );
    ed_np->setMaximumSize( 100, 30 );
    QToolTip::add(ed_np, "Object detection \"neighbor points\" number (0-8)" );
    lb = new QLabel( w );
    g->addWidget( lb, 1, 0 );
    lb->setText( "N points:" );
    lb->setBuddy( ed_np );
    lb->setMinimumSize( 30, 10 );
    lb->setMaximumSize( 100, 20 );
    QToolTip::add(lb, "Object detection \"neighbor points\" number (0-8)" );

    ed_sq = new QLineEdit( w );
    g->addWidget( ed_sq, 2, 1 );
    sprintf(str,"%.1f", *obj_sqr);
    ed_sq->setText( str );
    ed_sq->setMinimumSize( 30, 15 );
    ed_sq->setMaximumSize( 100, 30 );
    QToolTip::add(ed_sq, "Object minimum whole square (in SquareArcSec) " );
    lb = new QLabel( w );
    g->addWidget( lb, 2, 0 );
    lb->setText( "Min. square:" );
    lb->setBuddy( ed_sq );
    lb->setMinimumSize( 30, 10 );
    lb->setMaximumSize( 100, 20 );
    QToolTip::add(lb, "Object minimum whole square (in SquareArcSec) " );

    // finally insert page "center" into the tab dialog and start GM
    td_opt->addTab( w, "&Center" );
    g->activate();

    // set up 5-th page  of the tab dialog
    w = new QWidget( td_opt, "bgd" );
    g = new QGridLayout( w, 4, 3, 5 );

    ed_bp = new QLineEdit( w );
    g->addWidget( ed_bp, 0, 1 );
    ed_bp->setText( "no" );
    ed_bp->setMinimumSize( 180, 15 );
    ed_bp->setMaximumSize( 200, 30 );
    QToolTip::add(ed_bp, "Name of programm for background operations" );
    bt = new QPushButton( "Program:", w );
    g->addWidget( bt, 0, 0 );
    bt->setMaximumHeight(20);
    bt->setMaximumWidth(60);
    QToolTip::add(bt, "Choose programm for background operations" );
    connect( bt, SIGNAL(clicked()), SLOT(find_bgd_prog()) );
    bt = new QPushButton( "No", w );
    g->addWidget( bt, 0, 2 );
    bt->setMaximumHeight(20);
    bt->setMaximumWidth(25);
    QToolTip::add(bt, "No background operations at all" );
    connect( bt, SIGNAL(clicked()), SLOT(no_bgd_prog()) );

    ed_bf = new QLineEdit( w );
    g->addWidget( ed_bf, 1, 1 );
    ed_bf->setText( "frame" );
    ed_bf->setMinimumSize( 180, 15 );
    ed_bf->setMaximumSize( 200, 30 );
    QToolTip::add(ed_bf, "Source of background image (TV-frame or FITS-file)");
    bt = new QPushButton( "Source:", w );
    g->addWidget( bt, 1, 0 );
    bt->setMaximumHeight(20);
    bt->setMaximumWidth(60);
    QToolTip::add(bt, "\"frame\" or background file name " );
    connect( bt, SIGNAL(clicked()), SLOT(find_bgd_file()) );
    bt = new QPushButton( "TV", w );
    g->addWidget( bt, 1, 2 );
    bt->setMaximumHeight(20);
    bt->setMaximumWidth(25);
    QToolTip::add(bt, "Use TV-frame as a background" );
    connect( bt, SIGNAL(clicked()), SLOT(no_bgd_file()) );

    cb_bc = new QCheckBox( w );
    g->addWidget( cb_bc, 2, 0 );
    cb_bc->setText("Shift");
    cb_bc->setChecked(bgd_shift);
    cb_bc->setMinimumSize( 30, 15 );
    cb_bc->setMaximumSize( 60, 20 );
    connect( cb_bc, SIGNAL(clicked()), SLOT(toggle_bgd_shift()) );
    QToolTip::add(cb_bc, "Determine X/Y-shift of the background" );
    QBoxLayout *lhb =  new QHBoxLayout();
    g->addLayout(lhb,2,1);
    cb_su = new QCheckBox( w );
    lhb->addWidget(cb_su, 1, AlignLeft);
    cb_su->setText("Use Shift");
    cb_su->setChecked(bgd_ushift);
    cb_su->setMinimumSize( 30, 15 );
    cb_su->setMaximumSize( 90, 20 );
    connect( cb_su, SIGNAL(clicked()), SLOT(toggle_bgd_ushift()) );
    QToolTip::add(cb_su, "Use X/Y-shift for coordinates" );
    cb_bs = new QCheckBox( w );
    lhb->addWidget(cb_bs, 1, AlignRight);
    cb_bs->setText("Subtruct Bgd");
    cb_bs->setChecked(bgd_sub);
    cb_bs->setMinimumSize( 60, 15 );
    cb_bs->setMaximumSize( 90, 20 );
    connect( cb_bs, SIGNAL(clicked()), SLOT(toggle_bgd_sub()) );
    QToolTip::add(cb_bs, "Subtruct background" );

    bt = new QPushButton( "Reload", w );
    g->addWidget( bt, 3, 0 );
    bt->setMaximumHeight(30);
    bt->setMaximumWidth(60);
    QToolTip::add(bt, "Reload New background frame" );
    connect( bt, SIGNAL(clicked()), SLOT(new_bgd_frame()) );
    bt = new QPushButton( "Zero X/Y-shift", w );
    g->addWidget( bt, 3, 1 );
    bt->setMaximumHeight(30);
    bt->setMaximumWidth(180);
    QToolTip::add(bt, "Set background shift to zero" );
    connect( bt, SIGNAL(clicked()), SLOT(zero_bgd_shift()) );
    bt = new QPushButton( "Try", w );
    g->addWidget( bt, 3, 2 );
    bt->setMaximumHeight(30);
    bt->setMaximumWidth(25);
    QToolTip::add(bt, "Do background operation once" );
    connect( bt, SIGNAL(clicked()), SLOT(use_bgd_frame()) );

    // finally insert page "bgd" into the tab dialog and start GM
    td_opt->addTab( w, "&Bgd" );
    g->activate();

    // we want Apply & Default & Cancel in tab.dialog
    td_opt->setApplyButton();
    connect( td_opt, SIGNAL(applyButtonPressed()), SLOT(apply_opt()) );
    td_opt->setDefaultButton("Load");
    connect( td_opt, SIGNAL(defaultButtonPressed()), SLOT(get_tv_conf()) );
    td_opt->setCancelButton();

    td_opt->show();
}
void TVguide::destr_opt()  // fix that "Options" dialog window was destroyed
{
    td_opt = NULL;
}

void TVguide::toggle_send()  // Toggle "Really Dending Correction" mode
{
    modes ^= SEND_CORR;
    if (modes & SEND_CORR) {
       cb_snd->setText("Send Corr.");
       if(passwd==NULL)
	  show_psw_win();
       cb_snd->setChecked(TRUE);
       if ((modes & ACS_GET) == 0)
	  toggle_get();
    } else {
       cb_snd->setText("Don't Send");
       cb_snd->setChecked(FALSE);
    }
}

void TVguide::toggle_get()  // Toggle "Getting ACS BTA parameters" mode
{
    modes ^= ACS_GET;
    if (modes & ACS_GET) {
       cb_acs->setText("Get ACS");
       cb_acs->setChecked(TRUE);
    } else {
       cb_acs->setChecked(FALSE);
       cb_acs->setText("No ACS");
    }
}

void TVguide::toggle_flipy()  // Toggle coordinates flipping
{
    if (flipy==0) {
       flipy = 1;
       cb_fy->setText("Flip on Y");
    } else {
       flipy = 0;
       cb_fy->setText("No Y flip");
    }
    cb_fy->setChecked(flipy == 1);

    xy_to_az(cor_x, cor_y, &cor_a, &cor_z);
    graph->next_center(cor_a, cor_z);
    emit new_image(); // do GDimage::update();
}
void TVguide::toggle_flipx()  // Toggle coordinates flipping
{
    if (flipx==0) {
       flipx = 1;
       cb_fx->setText("Flip on X");
    } else {
       flipx = 0;
       cb_fx->setText("No X flip");
    }
    cb_fx->setChecked(flipx == 1);

    xy_to_az(cor_x, cor_y, &cor_a, &cor_z);
    graph->next_center(cor_a, cor_z);
    emit new_image(); // do GDimage::update();
}
void TVguide::toggle_bgd_shift() { // Toggle "Determine backgr.shift" mode
    bgd_shift = !bgd_shift;
    graph->set_head("BgdShiftCalc",bgd_shift?"yes":"no");
}
void TVguide::toggle_bgd_ushift() { // Toggle "Use backgr.shift for coord." mode
    bgd_ushift = !bgd_ushift;
    graph->set_head("BgdShiftUse",bgd_ushift?"yes":"no");
}
void TVguide::toggle_bgd_sub() {   // Toggle "Do background subtraction" mode
    bgd_sub = !bgd_sub;
    graph->set_head("BgdSubtruct",bgd_sub?"yes":"no");
}

void TVguide::apply_opt()  // Use new (edited) values from "Options" dialog
{
    char *p,bp[40],bf[100];
    bool restart_bgd=FALSE;
    bool reload_bgd=FALSE;
    ha = atof( ed_ha->text() );
    sx = atof( ed_sx->text() );
    sy = atof( ed_sy->text() );
    zd = atof( ed_zd->text() );
    pa = atof( ed_pa->text() );
    p2 = atof( ed_p2->text() );
    nfoc = atoi( ed_nf->text() );
    add_min = atoi( ed_mn->text() );
    if (add_min<1) add_min=1;
    corr_fact = atof( ed_tm->text() );
    if (corr_fact<0.0) corr_fact=0.0;
    corr_delay = atof( ed_dt->text() );
    if (corr_delay<0.011) corr_delay=0.011;
    *obj_lev = atof( ed_lv->text() );
    *obj_pnt = atoi( ed_np->text() );
    *obj_sqr = atof( ed_sq->text() );
    strncpy(bp,ed_bp->text(),39);
    p = strtok(bp," ");
    if(p==NULL || strcmp(p,"no")==0) {
       restart_bgd=(bgd_proc!=NULL);
       bgd_proc=NULL;
    } else {
       if(bgd_proc==NULL || strcmp(p,bgd_proc)!=0) {
	  bgd_proc=strdup(p);
	  restart_bgd=TRUE;
       }
    }
    strncpy(bf,ed_bf->text(),99);
    p = strtok(bf," ");
    if(p==NULL || strcmp(p,"no")==0 || strcmp(p,"frame")==0) {
       reload_bgd=(bgd_file!=NULL);
       bgd_file=NULL;
    } else {
       if(bgd_file==NULL || strcmp(p,bgd_file)!=0) {
	  bgd_file=strdup(p);
	  reload_bgd=TRUE;
       }
    }

    turn_axes();                        // calculate A/Z & Alp/Del angles
    xy_to_az(cor_x, cor_y, &cor_a, &cor_z);

    graph->set_head("PosAngle", ha);
    graph->set_head("Xscale",   sx);
    graph->set_head("Yscale",   sy);
    graph->set_head("Nadd",     add_min);
    graph->set_head("CorrFactor",corr_fact);
    graph->set_head("CorrDelay", corr_delay);
    graph->set_head("ObjLevel", *obj_lev);
    graph->set_head("Npoints",  *obj_pnt);
    graph->set_head("MinSquare",*obj_sqr);
    if( bta == NULL || !bta->connected() ) {
       graph->set_head("ParAngle",pa);
       graph->set_head("P2",      p2);
       graph->set_head("ZenDist", zd);
       graph->set_head("TelFocus", (nfoc==2)? "N2": ((nfoc==1)? "N1" : "CPF"));
    }
    graph->set_head("BgdProg",bp);
    graph->set_head("BgdFile",bf);
    graph->next_center(cor_a, cor_z);

    save_tv_conf(dfltfile);

    emit new_image(); // do GDimage::update();

    if(restart_bgd) {
       if(bgd_proc==NULL)
	  stop_bgd_proc();
       else
	  start_bgd_proc();
    } else if(reload_bgd) {
       if(pipe_in==NULL)
	  start_bgd_proc();
       else {
	  if(bgd_file==NULL)
	     {fputs("file no\n",pipe_out); fflush(pipe_out);}
	  else
	     {fprintf(pipe_out,"file %s\n",bgd_file); fflush(pipe_out);}
	  fputs("reload\n",pipe_out); fflush(pipe_out);
       }
    }
}

void TVguide::get_tv_conf() {
  char confname[100];
  char *p, filename[40];
  QString fname;

  fname = QFileDialog::getOpenFileName(confdir, "tv*.conf",this);
  if ( fname.isEmpty() ) return;

  if((p=strrchr(fname.data(),'/'))==NULL)
     strncpy(filename,fname.data(),40);
  else
     strncpy(filename,p+1,40);
  if(strchr(filename,'.'))
    sprintf(confname,"%s", fname.data());
  else
    sprintf(confname,"%s.conf", fname.data());
  fname.truncate( fname.findRev( '/', fname.length()-1 ) );
  strncpy(confdir,fname.data(),100);

  if(load_tv_conf(confname))
     save_tv_conf(dfltfile);
}

bool TVguide::load_tv_conf(char *confname) {  // load parameters from TV-config. file
  char *p, tmp[100];
  char *nam=NULL, *val=NULL;
  FILE *fd;
  bool restart_bgd=FALSE;
  bool reload_bgd=FALSE;

   sprintf(tmp,"Reading ConfFile:\n  %s",confname);
   add_msg(tmp);
   if((fd=fopen(confname,"r")) == NULL) {
      printf("Read ConfFile:%s Failed!?\n",confname);
      add_msg("Failed!?\n");
      return(FALSE);
   } else
      printf("Loading Configuration from file: %s\n",confname);

   while (fgets(tmp,99,fd) != NULL) {
      if(tmp[0] != '#' && (val=strchr(tmp,'=')) != NULL) {
	  nam = tmp;
	  *val++ = '\0';
	  if((p=strchr(val,'\n')) != NULL) *p = '\0';
	  if(strcmp(nam,"TelFocus")!=0 || bta == NULL || !bta->connected() )
	     graph->set_head(nam, val);
	  if(strcmp(nam,"TVsystem")==0) {
	     char  cap[100];
	     sprintf(cap,"TVguide for %s", val);
	     setCaption(cap);
	  } else if(strcmp(nam,"PosAngle")==0) {
	     ha = atof(val);
	     ed_ha->setText(val);
	  } else if(strcmp(nam,"Xscale")==0) {
	     sx = atof(val);
	     ed_sx->setText(val);
	  } else if(strcmp(nam,"Yscale")==0) {
	     sy = atof(val);
	     ed_sy->setText(val);
	  } else if(strcmp(nam,"Nadd")==0) {
	     add_min = atoi(val);
	     ed_mn->setText(val);
	  } else if(strcmp(nam,"CorrFactor")==0) {
	     corr_fact = atof(val);
	     ed_tm->setText(val);
	  } else if(strcmp(nam,"CorrDelay")==0) {
	     corr_delay = atof(val);
	     ed_dt->setText(val);
	  } else if(strcmp(nam,"ObjLevel")==0) {
	     *obj_lev = atof(val);
	     ed_lv->setText(val);
	  } else if(strcmp(nam,"Npoints")==0) {
	     *obj_pnt = atoi(val);
	     ed_np->setText(val);
	  } else if(strcmp(nam,"MinSquare")==0) {
	     *obj_sqr = atof(val);
	     ed_sq->setText(val);
	  } else if(strcmp(nam,"TelFocus")==0) {
	     if(strchr(val,'1')!=NULL) {
		nfoc=1;
		ed_nf->setText("1 - N1");
	     } else if(strchr(val,'2')!=NULL) {
		nfoc=2;
		ed_nf->setText("2 - N2");
	     } else {
		nfoc=0;
		ed_nf->setText("0 - CPF");
	     }
	  } else if(strcmp(nam,"FlipX")==0) {
	     if(strncmp(val,"on",2)==0 || strncmp(val,"y",1)==0) {
		if (flipx==0)
		   toggle_flipx();
	     } else if(strncmp(val,"of",2)==0 || strncmp(val,"n",1)==0) {
		if (flipx==1)
		   toggle_flipx();
	     }
	  } else if(strcmp(nam,"FlipY")==0) {
	     if(strncmp(val,"on",2)==0 || strncmp(val,"y",1)==0) {
		if (flipy==0)
		   toggle_flipy();
	     } else if(strncmp(val,"of",2)==0 || strncmp(val,"n",1)==0) {
		if (flipy==1)
		   toggle_flipy();
	     }
	  } else if(strcmp(nam,"BgdProg")==0) {
	     ed_bp->setText(val);
	     p = strtok(val," ");
	     if(p==NULL || strcmp(p,"no")==0) {
		restart_bgd=(bgd_proc!=NULL);
		bgd_proc=NULL;
	     } else {
		if(bgd_proc==NULL || strcmp(p,bgd_proc)!=0) {
		   bgd_proc=strdup(p);
		   restart_bgd=TRUE;
		}
	     }
	  } else if(strcmp(nam,"BgdFile")==0) {
	     ed_bf->setText(val);
	     p = strtok(val," ");
	     if(p==NULL || strcmp(p,"no")==0 || strcmp(p,"frame")==0) {
		reload_bgd=(bgd_file!=NULL);
		bgd_file=NULL;
	     } else {
		if(bgd_file==NULL || strcmp(p,bgd_file)!=0) {
		   char *q = strrchr(p,'/');
		   bgd_file=strdup(p);
		   reload_bgd=TRUE;
		 /*
		  *char *q = strrchr(p,'/');
		  *if(q!=NULL && *p=='/' && p!=q) {
		  *    *q = '\0';
		  *    bgd_dir=strdup(p);
		  *}
		  */
		}
	     }
	  } else if(strcmp(nam,"BgdDir")==0) {
	     p = strtok(val," ");
	     if(p!=NULL)
		bgd_dir=strdup(p);
	  } else if(strcmp(nam,"SaveDir")==0) {
	     p = strtok(val," ");
	     if(p!=NULL)
		strncpy(graph->savedir,p,99);
	  }
      }
   }
   fclose(fd);
   add_msg("Ok!");

   if (modes & SEND_CORR) toggle_send();
   if( bta != NULL && bta->connected() ) {
      if ((modes & ACS_GET) == 0) toggle_get();
   } else {
      if (modes & ACS_GET) toggle_get();
   }
   turn_axes();                        // calculate A/Z & Alp/Del angles
   xy_to_az(cor_x, cor_y, &cor_a, &cor_z);
   graph->next_center(cor_a, cor_z);
   emit new_image(); // do GDimage::update();

   if(restart_bgd) {
      if(bgd_proc==NULL)
	 stop_bgd_proc();
      else
	 start_bgd_proc();
   } else if(reload_bgd) {
      if(pipe_in==NULL)
	 start_bgd_proc();
      else {
	 if(bgd_file==NULL)
	    {fputs("file no\n",pipe_out); fflush(pipe_out);}
	 else
	    {fprintf(pipe_out,"file %s\n",bgd_file); fflush(pipe_out);}
	 fputs("reload\n",pipe_out); fflush(pipe_out);
      }
   }
   return(TRUE);
}
bool TVguide::save_tv_conf(char *confname) {  // save parameters to TV-config. file
  int i,n;
  const char *name[] = {
     "TVsystem",
     "PosAngle",
     "Xscale",
     "Yscale",
     "Nadd",
     "CorrFactor",
     "CorrDelay",
     "ObjLevel",
     "Npoints",
     "MinSquare",
     "TelFocus",
     "FlipX",
     "FlipY",
     "BgdProg",
     "BgdFile",
     "BgdDir",
     "SaveDir",
     NULL
  };
  char  val[100],tmp[100];
  FILE *fd;

  sprintf(tmp,"Writing ConfFile:\n  %s",confname);
  add_msg(tmp);
  if((fd=fopen(confname,"w")) == NULL) {
     printf("Write ConfFile:%s Failed!?\n",confname);
     add_msg("Failed!?\n");
     return(FALSE);
  } else
     printf("Saving Configuration to file: %s\n",confname);
  fprintf(fd,"# Configuration file for TVguide\n");
  for(i=0; name[i]!=NULL; i++) {
     if(strcmp(name[i],"TelFocus")==0 && bta != NULL && bta->connected()) {
	//n = atoi(bta->Get("TelFoc"));
	n = atoi( ed_nf->text() );
	fprintf(fd,"TelFocus=%s\n", (n==2)? "N2": ((n==1)? "N1" : "CPF"));
     } else if(graph->get_head(name[i], val))
	fprintf(fd,"%s=%s\n",name[i],val);
  }
  fclose(fd);
  add_msg("Ok!");
  return(TRUE);
}

int TVguide::set_status(int new_stat)    // change current TVguide state
{
    int old_stat = status;
    status = new_stat;
    switch (status) {
       default:
       case GD_IDLE   :
	   stat_lab->setText( "  Idle" );
	   break;
       case ACS_IO    :
	   stat_lab->setText( "  ACS I/O" );
	   break;
       case WAIT_CORR :
	   stat_lab->setText( "  Correction" );
	   break;
       case ADD_IMAGE :
	   stat_lab->setText( "  Add TV-images" );
	   break;
    }
    return( old_stat );
}

void TVguide::get_acs_opt()  // Get values from ACS BTA for "Options" dialog
{
    char tmp[100];

    get_from_acs();

    if (td_opt != NULL) {
       sprintf(tmp,"%.2f",pa);
       ed_pa->setText( tmp );
       sprintf(tmp,"%.2f",p2);
       ed_p2->setText( tmp );
       sprintf(tmp,"%.2f",zd);
       ed_zd->setText( tmp );
       sprintf(tmp,"%d  - %s", nfoc, (nfoc==2)? "N2": ((nfoc==1)? "N1" : "CPF"));
       ed_nf->setText( tmp );
    }
    xy_to_az(cor_x, cor_y, &cor_a, &cor_z);
    graph->next_center(cor_a, cor_z);

    emit new_image(); // do GDimage::update();
}

void TVguide::get_from_acs()  // Get P-angle & etc. values from ACS BTA
{
    char *str;

    set_status(ACS_IO);
    if( bta != NULL && bta->ACSconnected() ) {
       int n;
       str = bta->Get("TelFoc");
       n = atoi(str);
       if(n>=0 && n<=2) nfoc=n;
       else             nfoc=0;
       str = bta->Get("ZenDist");
       zd  = atof(str);
       str = bta->Get("PA");
       pa  = atof(str);
       if(nfoc==0 || nfoc>2) {            // CPF?
	  str = bta->Get("P2");
	  p2  = atof(str);
       } else {                           // N1 or N2
	  p2 = 0.0;
       }
    }
    set_status(GD_IDLE);

    turn_axes();                        // calculate A/Z & Alp/Del angles
}

void TVguide::acs_io_tmout()
{
}

void TVguide::turn_axes()            // calculate A/Z & Alp/Del angles
{
    if(nfoc==0) {                            // CPF?
       aza = (p2 - ha) * M_PI / 180.;
       ada = ((p2 - ha) - pa) * M_PI / 180.;
       flipz = 0;
    } else if(nfoc==1) {                     // N1?
       aza = (180. + zd - ha) * M_PI / 180.;
       ada = aza - pa * M_PI / 180.;
       flipz = 0;
    } else if(nfoc==2) {                     // N2?
       aza = (270. + zd - ha) * M_PI / 180.;
       ada = aza - pa * M_PI / 180.;
       flipz = 1;
    } else {                                 // 70sm Guide BTA?
       aza = 270. * M_PI / 180.;
       ada = aza - pa * M_PI / 180.;
       flipz = 0;
    }
}

void TVguide::start_tv_add()    // start tv-image accumulation process
{
IFDBG( printf("start_tv_add(); "); )
    if(!vs56_ok()) {
       char tmp[80];
       int chip=vs_type();
       sprintf(tmp,"No 'vs%s' driver for:\n   %s\n",(chip>800)?"bt":(chip==56)?"56":(chip==60)?"60":"",vs_name());
       set_status(GD_IDLE);
       clear_msg();
       add_msg(tmp);
       return;
    }
    if (status != ADD_IMAGE) {
       int im_w = TV_FRAME_MAX_WIDTH;
       int im_h = TV_FRAME_MAX_HEIGHT;
       tv_chan_get_frame(&col0, &lin0, &fr_w, &fr_h);
       if(fr_w > im_w || fr_h > im_h) {
	  if(fr_w > im_w) {
	     col0 += (fr_w - im_w)/2;
	     fr_w = im_w;
	  }
	  if(fr_h > im_h) {
	     lin0 += (fr_h - im_h)/2;
	     fr_h = im_h;
	  }
	  tv_chan_frame(col0, lin0, fr_w, fr_h);
       }
       int ccd_mode = tv_chan_ccd_mode();
       nseq = tv_chan_seq_num();
       mseq = tv_chan_seq_msg();
       tv_chan_go(1,1,add_min,1,0,ccd_mode);
/*printf("%s-mode\n",ccd_mode? "CCD":"Vidicon");*/
       add_timer->start(add_min*40, TRUE); // call tvimage_ok() after Nadd*0.04sec
       set_status(ADD_IMAGE);
    }
}

void TVguide::tvimage_ok()     // normal stop of images accumulation
{
IFDBG( printf("tvimage_ok(); TVguide::status=%d\n",status); )
    if(!vs56_ok()) {
       char tmp[80];
       int chip=vs_type();
       sprintf(tmp,"No 'vs%s' driver for:\n   %s\n",(chip>800)?"bt":(chip==56)?"56":(chip==60)?"60":"",vs_name());
       set_status(GD_IDLE);
       clear_msg();
       add_msg(tmp);
       return;
    }
    if (status == ADD_IMAGE) {
	 if (mseq < tv_chan_seq_msg()) {
	     char *msg = tv_chan_msg();
	     if (*msg != 'F') {
		set_status(GD_IDLE);
		add_msg(msg);
		return;
	     }
	 }
	 mseq = tv_chan_seq_msg();
	 int n = tv_chan_seq_num();
	 int nadd = tv_chan_added();
	 if (n==nseq || nadd < add_min) {
	    add_timer->start( 40, TRUE); // recall tvimage_ok() after 0.04sec
	    return;
	 }
//printf("tvimage_ok()\n");
	 tv_chan_stop();
	 clear_msg();
	 if (modes & ACS_GET)
	    get_acs_opt();          // refresh ACS BTA parameters
	 set_status(GD_IDLE);

	 use_bgd_frame();           // get backgr.shift OR subtruct backgr.frame

	 emit tvframe_ok();         // start centering:  GDimage::get_tv_image()

	 if ( modes & SEND_CORR )
	    start_corr();           // send correction to ACS BTA
	 else if (modes & GD_RECYCLE) {
	    // restart tv-image accumulation process after delay
	    set_status(WAIT_CORR);
	    cor_timer->start( (int)(corr_delay*1000.), TRUE); // call acs_corr_ok() after corr_delay seconds
	 }
	 graph->flush_save();
    }
}

void TVguide::tvguide_start()  //start auto-guiding process
{
    if (status == GD_IDLE) {
	 modes |= GD_RECYCLE;
	 start_tv_add();       // start tv-image accumulation process
    }
}

void TVguide::tvguide_stop()     //stop guiding process
{
    if (status == ADD_IMAGE || status == WAIT_CORR) {
	 tv_chan_stop();
	 if (status == ADD_IMAGE)
	     set_status(GD_IDLE);
    }
    modes &= ~GD_RECYCLE;
}

void TVguide::tvguide_once()     //start one guiding action
{
    if (status == GD_IDLE) {
	 modes &= ~GD_RECYCLE;
	 start_tv_add();       // start tv-image accumulation process
    }
}

bool TVguide::tvguide_busy()   // return TRUE when TVguide working
{
     return(status != GD_IDLE || (modes & GD_RECYCLE) != 0);
}

void TVguide::xy_to_az(double dx, double dy, double *da, double *dz) //convert coor.values from sceen to telescope
{
    double sn = sin(aza);
    double cs = cos(aza);
    double x = dx*sx;
    double y = dy*sy;

    if (bgd_ushift && (fabs(dx)>0.01 || fabs(dy)>0.01 ) ) {
       x -= bgd_x;
       y -= bgd_y;
    }
    if (flipx) x = -x;
    if (flipy) y = -y;
    *da = cs*x - sn*y;
    *dz = sn*x + cs*y;
    if (flipz)        // N1 or N2?
       *dz = -*dz;
}

void TVguide::put_corr(double da, double dz)  // Send A/Z correctiont to ACS BTA
{
    double s = sqrt(da*da + dz*dz)*corr_fact;
    double spd = s/10.;
    double snz = sin(zd * M_PI / 180.);
    int nspd = 0;
    int ta, tz;
    char tmp[80];

    graph->next_corrtn(da, dz);

    if      (spd > 5.0) { spd = 5.0; nspd = 5;}
    else if (spd > 2.0) { spd = 2.0; nspd = 4;}
    else if (spd > 1.0) { spd = 1.0; nspd = 3;}
    else if (spd > 0.4) { spd = 0.4; nspd = 2;}
    else                { spd = 0.2; nspd = 1;}

    ta = (int)((corr_fact * da/snz) / spd * 10. + 0.5);
    tz = (int)(corr_fact * dz / spd * 10. + 0.5);

    if(ta == 0 && tz == 0) return;

    sprintf(tmp,"Sending A/Z-corr: %d %d %d\n", nspd, ta, tz);
    add_msg(tmp);

    set_status(ACS_IO);
    int stat = GD_IDLE;
    if( bta != NULL && bta->ACSconnected() ) {
       char *str = bta->Get("TelStat");
       int tel_st = atoi(str);
       if (tel_st<3)
	  add_msg("Warning! Really no correction!\n Telescope not in Tracking mode.");
       else {
	  if (nfoc!=0)
	     printf("Warning! Maybe wrong correction! You are not in Prime focus but in N%d.\n",nfoc);
	  char tmp[40];
	  sprintf(tmp,"%d %d %d", nspd, ta, tz);
//          bta->Put("Corr",tmp);
	  stat = WAIT_CORR;
	  ta = abs(ta);
	  tz = abs(tz);
	  double tcor = ((ta>tz)? ta : tz )/10. + corr_delay;
	  cor_timer->start( (int)(tcor*1000.), TRUE); // call acs_corr_ok() after tcor seconds
       }
    }
    set_status(stat);
}

void TVguide::put_azshift(double da, double dz)  // Send A/Z shift to ACS BTA
{
    double snz = sin(zd * M_PI / 180.);
    double sda,sdz;
    char tmp[160];

IFDBG( printf("put_azshift(%.2f,%.2f); TVguide::status=%d\n",da,dz,status); )

    sda = corr_fact * da;
    sdz = corr_fact * dz;

    set_status(ACS_IO);
    int stat = GD_IDLE;
    if( bta != NULL && bta->ACSconnected() ) {
       char *str = bta->Get("TelStat");
       int tel_st = atoi(str);
       if (tel_st<3)
	  add_msg("Warning! Really no correction!\n Telescope not in Tracking mode.");
/*else*/if(passwd==NULL)
	  add_msg("Warning! Can't send!\nPermission denied!\nSet BTA access level 3 password first!");
       else {
	  if(fabs(da) >= 0.1 || fabs(dz) >= 0.1) {
	     graph->next_corrtn(sda, sdz);
	     sda /= snz;
	     sprintf(tmp,"Sending A/Z-shift:\n     A:  %.2f*Factor/sin(Z) = %.2f\n     Z:  %.2f*Factor = %.2f",
		da,sda, dz,sdz);
	     add_msg(tmp);
	     if(bta->new_acs)                          //New ACS Ok??
		DoAZcorr(sda, sdz);
	     else if(bta->old_acs) {                   //try old one...
		sprintf(tmp,"%.2f %.2f", sda, sdz);
		bta->Put("AZshift",tmp);
	     }
	  }
	  stat = WAIT_CORR;
	  double tcor = ((sda > 1.5*sdz)? sda : 1.5*sdz)  + 4.;  // minimum 4sec delay!
	  if(tcor<corr_delay) tcor = corr_delay;
	  cor_timer->start( (int)(tcor*1000.), TRUE); // call acs_corr_ok() after tcor seconds
       }
    }
    set_status(stat);
}

void TVguide::set_corr(double dx, double dy ) //set coor.values: x/y->a/z
{
    cor_x = dx;
    cor_y = dy;
    xy_to_az(cor_x, cor_y, &cor_a, &cor_z);
    if(fabs(dx)>0.01 || fabs(dy)>0.01 )
       graph->next_center(cor_a, cor_z);
}

void TVguide::get_corr(double *da, double *dz) //get coor.values: da/dz
{
    *da = cor_a;
    *dz = cor_z;
}

void TVguide::start_corr()
{
IFDBG( printf("start_corr(); TVguide::status=%d\n",status); )
    if (status == GD_IDLE) {
       put_azshift(cor_a, cor_z);
       if(passwd==NULL)
	  show_psw_win();
    }
}

void TVguide::acs_corr_ok()
{
IFDBG( printf("acs_corr_ok() TVguide::status=%d\n",status); )
    if (status == WAIT_CORR) {
	 set_status(GD_IDLE);
	 if (modes & GD_RECYCLE)
	     start_tv_add();      // restart tv-image accumulation process
    }
}

void TVguide::start_bgd_proc() { // (re)start bgd-maint. process
   char ch[2];
   char str[100];

   if(bgd_proc==NULL)
       return;
IFDBG( fprintf(stderr,"start_bgd_proc()\n"); )
   if(pipe_in!=NULL) {
      stop_bgd_proc();
      sprintf(str,"Restart BGD-proc:\n    %s %s %s", bgd_proc, tv_drv, bgd_file?bgd_file:" ");
   } else {
      sprintf(str,"Start BGD-proc:\n   %s %s %s", bgd_proc, tv_drv, bgd_file?bgd_file:" ");
      //clear_msg();
   }
   sprintf(ch,"%1d",n_chan);

   if (pipe(pipe0) < 0) goto proc_err;
   if (pipe(pipe1) < 0) goto proc_err;
   if ((bgd_pid = fork()) < 0) goto proc_err;

   if (bgd_pid == 0) {
      if (dup2(pipe0[1],1) < 0) goto proc_stop;  // assigne stdout
      if (dup2(pipe1[0],0) < 0) goto proc_stop;  //     and stdin

      if(bgd_file==NULL)
	 execlp(bgd_proc, bgd_proc, tv_drv, NULL);
      else
	 execlp(bgd_proc, bgd_proc, tv_drv, bgd_file, NULL);
proc_stop:
      fprintf(stderr,"Can't start BGD:\n    %s %s %s", bgd_proc, tv_drv, bgd_file?bgd_file:" ");
      perror("Error:");
      write(1,"quit\n",5);
      exit(errno);
   } else {
      int status;
      int pid;
      pipe_in  = fdopen(pipe0[0],"r");
      pipe_out = fdopen(pipe1[1],"w");
      my_sleep(0.5);
      if((pid=waitpid(bgd_pid, &status, WNOHANG)) !=0 && pid==bgd_pid) {
	 fclose(pipe_in);
	 fclose(pipe_out);
proc_err:
	 sprintf(str,"Can't start:\n    %s %s %s", bgd_proc, tv_drv, bgd_file?bgd_file:" ");
	 pipe_in = pipe_out = NULL;
	 bgd_pid = -1;
      } else if(pid<0) {
	 perror("waitpid() failed");
      } else if(pid>0) {
	 fprintf(stderr,"Unexpected child PID=%d instead of %d\n",pid,bgd_pid);
      }
      tv_chan_get_frame(&bgd_col0, &bgd_lin0, &bgd_frmw, &bgd_frmh);// fix background frame
   }
IFDBG( fprintf(stderr,"%s\n",str);fflush(stderr);)
   add_msg(str);
}

void TVguide::check_bgd_proc() { // check bgd-maint. process
   char *p,bp[40];
   if(bgd_pid>0) {
      if(waitpid(-1, &status, WNOHANG) == bgd_pid || kill(bgd_pid,0)<0) {
	 char *msg = "CRASH of BGD-program?!\n";
	 add_msg(msg);
	 fputs(msg,stderr);
	 stop_bgd_proc();
      }
   }
   strncpy(bp,ed_bp->text(),39);
   p = strtok(bp," ");
   if(p==NULL || strcmp(p,"no")==0) {
      if(bgd_proc!=NULL)
	 stop_bgd_proc();
      bgd_proc=NULL;
   } else {
      if(bgd_proc==NULL || strcmp(p,bgd_proc)!=0) {
	 if(bgd_proc!=NULL)
	    stop_bgd_proc();
	 bgd_proc=strdup(p);
      }
   }
}

void TVguide::new_bgd_frame() {     // reload new backgr.frame
   char *p,bf[100];

IFDBG( printf("new_bgd_frame()\n"); )
   check_bgd_proc();

   if(pipe_in==NULL && bgd_proc!=NULL) {
      start_bgd_proc();
   } else {
      if(pipe_in==NULL) return;
      strncpy(bf,ed_bf->text(),99);
      p = strtok(bf," ");
      if(p==NULL || strcmp(p,"no")==0 || strcmp(p,"frame")==0) {
	 if(bgd_file!=NULL) {
	    bgd_file = NULL;
	    fputs("file no\n",pipe_out); fflush(pipe_out);
	 }
      } else {
	 if(bgd_file==NULL || strcmp(p,bgd_file)!=0) {
	    bgd_file=strdup(p);
	    fprintf(pipe_out,"file %s\n",bgd_file); fflush(pipe_out);
	 }
      }
      fputs("reload\n",pipe_out);                 // new bgd frame
      fflush(pipe_out);
      tv_chan_get_frame(&bgd_col0, &bgd_lin0, &bgd_frmw, &bgd_frmh);// fix background frame
   }
}

void TVguide::use_bgd_frame() {  // dummy OR get backgr.shift OR subtruct backgr.frame
   char *p,str1[20],str2[20],tmp[100];
   double dx,dy;
   bool reload = (bgd_file != NULL);

IFDBG( printf("use_bgd_frame() "); )
   check_bgd_proc();

   if(pipe_in==NULL && bgd_proc!=NULL) {
      start_bgd_proc();
      reload = FALSE;
   }
   if(pipe_in==NULL) return;
   tv_chan_get_frame(&col0, &lin0, &fr_w, &fr_h); // current  frame
   if(bgd_col0!=col0 || bgd_lin0!=lin0 ||
      bgd_frmw!=fr_w || bgd_frmh!=fr_h) {         // new frame?
      if(reload) {
	 fputs("reload\n",pipe_out);              // new bgd from file
	 fflush(pipe_out);
IFDBG( printf("bgd_reload "); )
      }
      bgd_col0 = col0;                            // fix new one...
      bgd_lin0 = lin0;
      bgd_frmw = fr_w;
      bgd_frmh = fr_h;
   }
   if(bgd_shift) {
      fputs("next\n",pipe_out); fflush(pipe_out);
      fgets(str1,20,pipe_in);
      if(str1[0]=='q') {
	 stop_bgd_proc();
	 return;
      }
      if(strncmp(str1,"dX=",3)==0) bgd_x=atof(&str1[3]);
      fgets(str2,20,pipe_in);
      if(str1[0]=='q') {
	 stop_bgd_proc();
	 return;
      }
      if(strncmp(str2,"dY=",3)==0) bgd_y=atof(&str2[3]);
      dx = bgd_x;
      dy = bgd_y;
      bgd_x *= sx;
      bgd_y *= sy;
      graph->next_bgdxy(bgd_x, bgd_y);
      sprintf(tmp, "BGD: dX=%.1f(%.2f) dY=%.1f(%.2f)\n", dx,bgd_x, dy,bgd_y);
      add_msg(tmp);
IFDBG( printf("bgd_shift "); )
   }
   if(bgd_sub) {
      int i, n = tv_chan_seq_num();
      fputs("sub\n",pipe_out); fflush(pipe_out);
      for(i=10; i>0; i--) {
	 my_sleep(0.2);
	 if(n != tv_chan_seq_num()) break;
      }
      if(i==0) {
	 sprintf(tmp,"BGD subtruction HANGs?!");
	 add_msg(tmp);
IFDBG( printf("%s ",tmp); )
      } else {
IFDBG( printf("bgd_sub "); )
      }
   }
IFDBG( printf("\n"); )
}

void TVguide::zero_bgd_shift() {     // set zero bgd.shift (bgd_x=bgd_y=0)
IFDBG( printf("zero_bgd_shift()\n"); )
   bgd_x=bgd_y=0.;
   graph->next_bgdxy(bgd_x, bgd_y);

   check_bgd_proc();
   if(pipe_in==NULL && bgd_proc!=NULL) {
      start_bgd_proc();
   } else {
      if(pipe_in==NULL) return;
      fputs("zero xy\n",pipe_out);                 // zero bgd.shift
      fflush(pipe_out);
   }
}

void TVguide::stop_bgd_proc() {  // stop bgd-maint. process
    char str[100];
    int pid,status;
    if(pipe_in!=NULL) {
IFDBG( printf("stop_bgd_proc()\n"); )
       fputs("quit\n",pipe_out);fflush(pipe_out);
       my_sleep(0.5);
       fclose(pipe_in);
       fclose(pipe_out);
       sprintf(str,"Stop BGD-process.");
       //clear_msg();
       add_msg(str);
       if((pid=waitpid(-1, &status, WNOHANG))==0) {
	  if(bgd_pid>0) kill(bgd_pid,15);
	  my_sleep(0.5);
	  pid=waitpid(-1, &status, WNOHANG);
       }
       if(pid<0) {
	  perror("waitpid() failed");
	  bgd_pid=-1;
       } else if(pid==0) {
       cant_kill:
	  sprintf(str,"Can't kill BGD-process!?");
	  fprintf(stderr,"%s\n",str);fflush(stderr);
	  add_msg(str);
       } else if(pid!=bgd_pid) {
	  fprintf(stderr,"Unexpected child PID=%d instead of %d\n",pid,bgd_pid);
	  if(bgd_pid>0) kill(bgd_pid,15);
	  my_sleep(0.5);
	  if(waitpid(-1, &status, WNOHANG) != bgd_pid)
	     goto cant_kill;
	  bgd_pid=-1;
       } else
	  bgd_pid=-1;
    }
    pipe_in = pipe_out = NULL;

}

void TVguide::find_bgd_prog() // Choose programm for background operations
{
  char *p;
  QString fname;

  fname = QFileDialog::getOpenFileName("bin", "*_bgd",this, ed_bp->text());
  if ( fname.isEmpty() ) return;
  if ((p = strrchr(fname.data(),'/'))==NULL)
     p=fname.data();
  else
     p++;
  ed_bp->setText(p);
}
void TVguide::no_bgd_prog()  // set NO background operations at all
{
     ed_bp->setText("no");
}

void TVguide::find_bgd_file()  // browse for background file
{
  char fitsname[80];
  char *p;
  QString fname;

  fname = QFileDialog::getOpenFileName(bgd_dir, "*.mt*",this, ed_bf->text());
  if ( fname.isEmpty() ) return;
  p=fname.data();
  if(strcmp(p,"no")==0 || strcmp(p,"frame")==0 ) {
     ed_bf->setText(p);
     return;
  }
  if(strstr(p,".mt"))
    sprintf(fitsname,"%s", p);
  else
    sprintf(fitsname,"%s.mt", p);
  ed_bf->setText(fitsname);

  fname.truncate( fname.findRev( '/', fname.length()-1 ) );
  if (strcmp(bgd_dir, fname.data())!=0) {
     bgd_dir = strdup(fname.data());
     graph->set_head("BgdDir",bgd_dir);
  }
}
void TVguide::no_bgd_file()
{
     ed_bf->setText("frame");
}

void TVguide::show_psw_win()
{
    if (wpsw != NULL) {
       set_max_size();
       if(wpsw->isVisible())
	  wpsw->raise();
       else {
	  wpsw->show();
       }
       return;
    }

    //wpsw = new QWidget( td_misc, "Passwd" );
    wpsw = new QFrame( td_misc, "Passwd" );
    wpsw->setFrameStyle(QFrame::Box|QFrame::Raised);
    wpsw->setGeometry( 40, 200, 200, 70 );

    QGridLayout * g = new QGridLayout( wpsw, 3, 1, 2 );

    ed_psw = new QLineEdit( wpsw );
    g->addWidget( ed_psw, 1, 0, AlignCenter );
    ed_psw->setEchoMode ( QLineEdit::Password );
    ed_psw->setMinimumSize( 30, 15 );
    ed_psw->setMaximumSize( 150, 20 );
    connect( ed_psw, SIGNAL(returnPressed()), SLOT(set_passwd()) );
    QToolTip::add(ed_psw, "Enter password for BTA access permition level 3 " );

    QLabel * lb = new QLabel( wpsw );
    g->addWidget( lb, 0, 0, AlignCenter );
    lb->setText( "BTA access Password:" );
    lb->setMinimumSize( 150, 10 );
    lb->setMaximumSize( 180, 20 );

    QPushButton* ok = new QPushButton( "Ok", wpsw );
    ok->resize(20,10);
    ok->setMinimumSize( 20, 10 );
    ok->setMaximumSize( 150, 20 );
    connect( ok, SIGNAL(clicked()), SLOT(set_passwd()) );
    QToolTip::add(ok, "Start password matching procedure" );
    g->addWidget(ok, 2, 0, AlignCenter);
    g->activate();

    //wpsw->setCaption( "TVguide Password" );
    //wpsw->resize(200,60);
    wpsw->show();
}
void TVguide::set_passwd() {
    char str[200];
    passwd = strdup(ed_psw->text());
    wpsw->hide();
    clear_msg();
    if( bta != NULL && bta->connected() ) {
       if(bta->new_acs) {
	  int nlev;
	  if((nlev=check_lev_passwd(passwd)) < 3) {
	     if(nlev==0)
		sprintf(str, "Warning! Wrong password(%s)!?\n",passwd);
	     else
		sprintf(str, "Warning! Password for Level %d!? ( >= 3 needs)\n",nlev);
	     passwd = NULL;
	  } else
	     sprintf(str, "Passwd - Ok! NewACS AccessLevel = %d.\n",nlev);
	  add_msg(str);
       }
       if(bta->old_acs) {
	  sprintf(str,"Passwd %s\n",passwd);
	  bta->write_str(str);
	  bta->read_str(str,200);
	  puts(str);
	  if(strstr(str,"permission level is 3")) {
	     sprintf(str,"Password for OldACS access level 3 OK!");
	  } else {
	     if(!bta->new_acs)
		passwd = NULL;
	     sprintf(str,"Wrong password for OldACS?\nObtain access level 3 failed!");
	  }
	  add_msg(str);
       }
       if(passwd != NULL)
	     sprintf(str,"You can send A/Z corrections.");
       else
	     sprintf(str,"Really can't send A/Z corrections!");
       add_msg(str);
    } else {
       sprintf(str,"No connection to bta_serv!\nCan't match password!\n");
       add_msg(str);
    }
}

void TVguide::show_msg_win()
{
    if (wmesg != NULL) {
       //set_max_size();
       if(wmesg->isVisible())
	  wmesg->raise();
       else {
	  wmesg->show();
       }
       return;
    }

    wmesg = new QMultiLineEdit( td_misc, "Text" );

    QColor fc,bc;
    fc.setNamedColor("DarkSlateBlue");
    bc.setNamedColor("Gray");
    QColor lc(bc.light());
    QColor mc(bc);
    QColor dc(bc.dark());
    QColorGroup gc(fc,bc,lc,dc,mc,fc,bc);
    QPalette p(gc,gc,gc);
    wmesg->setPalette(p);
    wmesg->setBackgroundMode(PaletteBase);
    wmesg->setFont( QFont( "Helvetica", 8, QFont::Bold ) );
    wmesg->setReadOnly ( TRUE );
    wmesg->setAutoUpdate( FALSE );
    wmesg->clear();

    wmesg->setAutoUpdate( TRUE );
    wmesg->resize(280,200);
    wmesg->repaint();

    //wmesg->setCaption( "TVguide Information" );
    wmesg->show();
}
void TVguide::add_msg(char *str)
{
    if (wmesg != NULL) {
       //set_max_size();
       if(wmesg->isVisible())
	  wmesg->raise();
       else {
	  wmesg->show();
       }
    } else
       show_msg_win();
    wmesg->append( str );
    wmesg->setCursorPosition(wmesg->numLines()-1, 0);
}
void TVguide::clear_msg()
{
    if (wmesg != NULL) {
       //set_max_size();
       if(wmesg->isVisible())
	  wmesg->raise();
       else {
	  wmesg->show();
       }
       wmesg->clear();
    } else
       show_msg_win();
}

GDimage::GDimage( QWidget *parent, const char *name )
    : QFrame( parent, name)
{
    double f,r,cx,cy;
    int i,pcol,plin;
    const int gdim_w = GDimageMaxWidth;
    const int gdim_h = GDimageMaxHeight;
    need_update = TRUE;
    im_w = 230;
    im_h = 230;
    tv_chan_get_frame(&col0, &lin0, &fr_w, &fr_h);
    if(fr_w > gdim_w) {
       col0 += (fr_w - gdim_w)/2;
       fr_w = gdim_w;
    }
    if(fr_h > gdim_h) {
       lin0 += (fr_h - gdim_h)/2;
       fr_h = gdim_h;
    }
    tv_chan_get_point(&pcol, &plin);
    if(pcol>col0 && pcol<col0+fr_w-1 && plin>lin0 && plin<lin0+fr_h-1) {
       x_c = pcol - col0 + (im_w-fr_w)/2;
       y_c = plin - lin0 + (im_h-fr_h)/2;
    } else {
       x_c = im_w/2;
       y_c = im_h/2;
    }
    tv_chan_get_center(&cx, &cy);
    if(cx>col0+1 && cx<col0+fr_w-1 && cy>lin0+1 && cy<lin0+fr_h-1) {
       xc = cx - col0 + (im_w-fr_w)/2;
       yc = cy - lin0 + (im_h-fr_h)/2;
       mouse_x = (int)(xc + 0.5);
       mouse_y = (int)(yc + 0.5);
    } else {
       xc = mouse_x = im_w/2;
       yc = mouse_y = im_h/2;
    }
    fwhm_x=0.;
    fwhm_y=0.;
    obj_lev = 3.;  // 3-sigma threshhold
    obj_pnt = 3;   // don't use single pixels
    obj_sqr = 0.5; // 0.5 SqArcSec min.obj.square

    mask = (char *)malloc(gdim_w*gdim_h);

    itab = (double *)calloc(1000,sizeof(double));
    for (f=r=0,i=0; i<1000; i++,r+=0.01) {
       double g = exp(-r*r/2.);
       f += r*r*r*g*0.01;
       itab[i] = (f + r*r*g) * M_PI;
    }

    setGeometry( 5, 5, im_w, im_h );

    QSlider * xsc = new QSlider(1,200,10,100,QSlider::Horizontal,this,"x_pos");
    connect( xsc, SIGNAL(sliderMoved(int)), this, SLOT(set_x_pos(int)) );
    xsc->setGeometry( 40, 215, 150, 15 );
    QToolTip::add( xsc, "X-position shift" );

    QSlider * ysc = new QSlider(1,200,10,100,QSlider::Vertical,this,"y_pos");
    connect( ysc, SIGNAL(sliderMoved(int)), this, SLOT(set_y_pos(int)) );
    ysc->setGeometry( 215, 40, 15, 150 );
    QToolTip::add( ysc, "Y-position shift" );

    QSlider * zsc = new QSlider(1,100,10,50,QSlider::Vertical,this,"xy_zoom");
    connect( zsc, SIGNAL(sliderMoved(int)), this, SLOT(set_xy_zoom(int)) );
    zsc->setGeometry( 0, 65, 15, 100 );
    QToolTip::add( zsc, "Change XY zoom" );

    gbt = new QButtonGroup(this);
    gbt->setGeometry (14, -2, 203, 18 );
    gbt->setFrameStyle( QFrame::Box | QFrame::Raised );
    rbt[0] = new QRadioButton("Frame",gbt);
    rbt[0]->setGeometry (3, 1, 60, 15 );
    rbt[0]->setChecked(TRUE);
    rbt[1] = new QRadioButton("SubFrm1",gbt);
    rbt[1]->setGeometry (70, 1, 60, 15 );
    rbt[2] = new QRadioButton("SubFrm2",gbt);
    rbt[2]->setGeometry (140, 1, 60, 15 );
    connect( gbt, SIGNAL( pressed(int)), SLOT(set_image_source(int)) );
    set_image_source(0);

    lin.begin(this);
    //lin.setBackgroundMode(OpaqueMode);
    QColor c;
    c.setNamedColor("White");
    lin.setPen(c);
    lin.setRasterOp (CopyROP);
    //lin.setFont( QFont( "times", 12, QFont::Bold ) );
    lin.setFont( QFont( "Symbol", 14, QFont::Normal ) );

    gdtime.start();
    gd_timer = new QTimer(this);
    connect( gd_timer, SIGNAL(timeout()), this, SLOT(cyclic_update()) );
    gd_timer->start( 1000, FALSE); // check need_update every second
}
GDimage::~GDimage()
{
    lin.end();
}

void GDimage::set_x_pos(int ix) {
    float x0 = mtx.dx();
    float x = ix - 100.;
    mtx.translate(x-x0,0.);
    rmtx.translate(x0-x,0.);
    update(0.2);
}
void GDimage::set_y_pos(int iy) {
    float y0 = mtx.dy();
    float y = iy - 100.;
    mtx.translate(0.,y-y0);
    rmtx.translate(0.,y0-y);
    update(0.2);
}
void GDimage::set_xy_zoom(int i) {
    float f = i/50.0;
    float z = 0.25*f*(f+1.)+0.5;
    float dz = z/mtx.m11();
    mtx.scale(dz,dz);
    rmtx.scale(1./dz,1./dz);
    tv_pm = tv_frame_pm.xForm( mtx );      // do zoom on image pixmap
    update(0.2);
//printf("z=%5.2f dz=%5.2f mtx.m11()=%5.2f rmtx.m11()=%5.2f\n",z,dz,mtx.m11(),rmtx.m11());
}

void GDimage::set_image_source(int nfrm) {
    if(im_src != nfrm) {
       if(tvg->tvguide_busy()) {
	  tvg->clear_msg();
	  tvg->add_msg("Don't switch frames!");
	  tvg->add_msg("Stop guiding first!");
       } else {
	  im_src = (nfrm<=0)? 0 : (nfrm<2)? 1 : 2;
       }
    }
}


void GDimage::get_tv_image()
{
    int l,x,y;
    char *m=NULL, *pm;
    const int gdim_w = GDimageMaxWidth;
    const int gdim_h = GDimageMaxHeight;

IFDBG( printf("GDimage::get_tv_image()\n"); )

    tvim_w = tv_chan_img_width();
    tvim_h = tv_chan_img_height();

    for(l=0;l<3;l++)
       rbt[l]->setChecked(l==im_src);
    switch(im_src) {
       int cols,lins;
       default:
       case 0: tv_chan_get_frame(&col0, &lin0, &fr_w, &fr_h);
	       break;
       case 1:
	       tv_chan_get_frame(&col0, &lin0, &fr_w, &fr_h);
	       tv_chan_get_subframe(&cols, &lins, &fr_w, &fr_h);
	       col0 += cols; lin0 += lins;
	       break;
       case 2:
	       tv_chan_get_frame(&col0, &lin0, &fr_w, &fr_h);
	       tv_chan_get_subframe2(&cols, &lins, &fr_w, &fr_h);
	       col0 += cols; lin0 += lins;
	       break;
    }
    if(fr_w > gdim_w) {
       col0 += (fr_w - gdim_w)/2;
       fr_w = gdim_w;
    }
    if(fr_h > gdim_h) {
       lin0 += (fr_h - gdim_h)/2;
       fr_h = gdim_h;
    }
    stat_tv_image();    // calc vmin, vmax, avr, med, dsp & lev

    if(obj_pnt>0)
       m = (char *)malloc(fr_w*fr_h);
    if((obj_pnt>0 && m==NULL) || mask==NULL) {
       printf("GDimage::get_tv_image() FAIL! Can't allocate memory\n");
       return;
    }
    QImage::Endian bitord = QImage::systemBitOrder();
    QImage tv_frame( fr_w, fr_h, 1, 2, bitord); // create frame image
    QColor c;
    c.setNamedColor("Gray");
    tv_frame.setColor( 0, c.rgb() );
    c.setNamedColor("Cyan");
    tv_frame.setColor( 1, c.rgb() );

    pm = (obj_pnt>0)? m : mask;

    for ( y=0; y<fr_h; y++ ) {            // setup object mask
	for ( x=0; x<fr_w; x++ ) {
	    int pix = tv_image[(lin0+y)*tvim_w + col0+x];
	    if(pix < lev3) *pm++ = -1;
	    else           *pm++ = 0;
	}
    }

    if(obj_pnt>0) {                      // setup "neighborhood" mask
       for ( y=0; y<fr_h; y++ ) {
	  for ( x=0; x<fr_w; x++ ) {
	      int k, n=y*fr_w + x;
	      mask[n]=m[n];
	      if(m[n]>=0) {
		 for(int j=-1; j<2; j++) {
		    int yj=y+j;
		    if(yj>=0 && yj<fr_h) {
		       for(int i=-1; i<2; i++) {
			  int xi=x+i;
			  if(xi>=0 && xi<fr_w && (i!=0 || j!=0)) {
			     k=yj*fr_w + xi;
			     if(m[k]>=0)
				mask[n]++;
			  }
		       }
		    }
		 }
	      }
	  }
       }
       free(m);
    }
    pm =  mask;
    for ( y=0; y<fr_h; y++ ) {                     // set frame image pixels
	uchar *p = tv_frame.scanLine(y);
	*p = 0;
	for ( x=0; x<fr_w; x++ ) {
	    int nbit = (bitord==QImage::LittleEndian)? x&7 : 7-(x&7);
	    int pix;
	    if(*pm++ < obj_pnt) pix=0;
	    else                 pix=1;
	    *p |= pix<<nbit;
	    if (x%8 == 7) {p++; *p = 0;}
	}
    }
    tv_frame_pm = tv_frame;                // convert frame image to pixmap
    tv_pm = tv_frame_pm.xForm( mtx );      // do zoom

    if (cent_tv_image()) {               // calc coords of image center
       mouse_x = (int)(xc + 0.5);
       mouse_y = (int)(yc + 0.5);

       tvg->set_corr(xc-x_c, yc-y_c);

       tv_chan_center(xc - (im_w-fr_w)/2 + col0, yc - (im_h-fr_h)/2 + lin0);

    } else
       tvg->set_corr(0., 0.);

    update();

}

void GDimage::stat_tv_image()    // do some statistics on tv_image frame:
{                                // calculate vmin, vmax, avr, med, dsp & lev
   short *img = tv_image;                     /* tv_image memory */
   int imgw=tvim_w;                           /* tv_image width */
   int width=fr_w, height=fr_h;               /* frame size    */
   int i,j,k,x,y,pix;
   int n,nbot,nmid,ntop;
   int hist0,hbot,htop;
   char tmp[80];

   n = width*height;

   nbot = (int)(n*0.03 + 0.5);     //  ~3%  ""  ""
   ntop = (int)(n*0.97 + 0.5);     //   
   nmid = n/2;

   for (j=0; j<6554; j++) hist[j]=0;
   hist0 = -32768;
   vmax = vmin = img[0];
   for (j=0; j<height; j++) {      //    =10
      y=lin0+j;
      for (i=0; i<width; i++) {
	 x=col0+i;
	 pix = img[imgw*y+x];
	 k = (pix-hist0)/10;
	 hist[k]++;
	 if(pix<vmin) vmin=pix;
	 if(pix>vmax) vmax=pix;
      }
   }
   for (i=j=0; j<6553 && i<nmid; j++, i += hist[j]);
   med = j*10+hist0;
   hist0 = med-3277;

   for (j=0; j<6554; j++) hist[j]=0;
   for (j=0; j<height; j++) {      //    =1
      y=lin0+j;
      for (i=0; i<width; i++) {
	 x=col0+i;
	 pix = img[imgw*y+x];
	 k = pix-hist0;
	 k = (k<0)? 0 : ((k>6553)? 6553 : k);
	 hist[k]++;
      }
   }
   hbot = 0;                       //  "" 
   htop = 6553;                    //    ~6% 
   for (i=j=0; j<6553; j++) {      //  ""  avr  dsp
      i += hist[j];
      if(i<=nbot) hbot = j;
      if(i<=nmid) med  = j+1;
      if(i<=ntop) htop = j+1;
   }
   if(med-hbot < htop-med) {       //   "" -
      htop = med + (med-hbot);
      if(htop>6553) htop=6553;
   } else {
      hbot = med - (htop-med);
      if(hbot<0) hbot=0;
   }
   med += hist0;
   for(k=0;k<7;k++) {
//printf("Range= %d - %d\n",hbot+hist0,htop+hist0);
      avr = 0.;
      for (i=0,j=hbot; j<=htop; j++) {
	 avr += (double)hist[j]*j;
	 i += hist[j];
      }
      if(i<1) i=1;
      avr /= i;
      for (dsp=0.,j=hbot; j<=htop; j++) {
	 double d = j-avr;
	 dsp += d*d*hist[j];
      }
      avr += hist0;
      dsp = sqrt(dsp/i);
//printf("Med=%d Avr=%.2f Dsp=%.2f\n",med,avr,dsp); fflush(stdout);
      htop = (int)(avr+2.0*(dsp<1.?1:dsp)+0.5) - hist0;
      if(htop>6553) htop=6553;
   }
   lev1 = (int)(avr+dsp+0.5);
   lev2 = (int)(avr+2.*dsp+0.5);
   lev3 = (int)(avr+obj_lev*dsp+0.5);
   sprintf(tmp,"Max=%d Med=%d Avr=%.2f Sigma=%.2f",vmax,med,avr,dsp);
//puts(tmp); fflush(stdout);
   tvg->add_msg(tmp);
}

bool GDimage::cent_tv_image()      // calc. center coordinates at tv_image frame
{
   short *img = tv_image;                     /* tv_image memory */
   char *pm =mask;
   int imgw=tvim_w;                           /* tv_image width */
   int width=fr_w, height=fr_h;               /* frame size */
   int i,j,n,x,y;
   double spx,spy,s,pix,ic,jc,dx,dy;
   double fx,fy,f,lev,kf,sgm,r,gi,icent;
   double fw1 = 2.*sqrt(2.*log(2.));           /* FW for Gigma=1 */
   char tmp[80];

   int nsq;          //min.sq. in pixels
   int nadd = tv_chan_added();
   if(nadd<1) nadd=1;

   double sx= tvg->sx;
   double sy= tvg->sy;

   if (sx<0.001) sx=sy;
   if (sy<0.001) sy=sx;
   if (sx<0.001) sx=0.1;
   if (sy<0.001) sy=0.1;
   nsq = (int)(obj_sqr/(sx*sy)+0.5);

   spx=spy=s=0.;
   pm = mask;
   for (n=j=0; j<height; j++) {
      y=lin0+j;
      for (i=0; i<width; i++) {
	 if(*pm++ >= obj_pnt) {
	    n++;
	    x=col0+i;
	    pix = img[imgw*y+x] - avr;
	    spx += i*pix;
	    spy += j*pix;
	    s += pix;
	 }
      }
   }
   if (n<nsq) {
      sprintf(tmp,"Lev=%d", lev3);
      tvg->add_msg(tmp);
      sprintf(tmp,"Object too small(%.3f < %.2f)\n",n*sx*sy, obj_sqr);
      tvg->add_msg(tmp);
      return(FALSE);
   }
   if(s<0.1 && s>-0.1) s = (s>=0.)? 0.1 : -0.1;
   ic = spx/s;
   jc = spy/s;
   xc = ic + (im_w-fr_w)/2;
   yc = jc + (im_h-fr_h)/2;
   sprintf(tmp,"Lev=%d Square=%.2f Int1=%.1f\n", lev3, n*sx*sy, s/nadd);
   tvg->add_msg(tmp);
   sprintf(tmp,"Xc=%.2f (%.1f)  Yc=%.2f (%.1f)\n",
		(xc-x_c)*sx, ic+col0, (yc-y_c)*sy, jc+lin0);
   tvg->add_msg(tmp);

   spx=spy=0.;
   pm = mask;
   for (j=0; j<height; j++) {
      y=lin0+j;
      dy = j-jc;
      for (i=0; i<width; i++) {
	 if(*pm++ >= obj_pnt) {
	    x=col0+i;
	    dx = i-ic;
	    pix = img[imgw*y+x] - avr;
	    spx += dx*dx*pix;
	    spy += dy*dy*pix;
	 }
      }
   }
   // relative obj.detect level  (0 -:- 1)
   lev = (lev3-avr)/(vmax-avr);
   if(lev>0.99)lev=0.99;
   if(lev<0.0001)lev=0.0001;

   // relative object radius
   r = sqrt(2.*log(1./lev));

   // expected integrall of gaussian restricted by r
   i = (int)(r*100.);
   if (i>=1000) i=999;
   if (i<1) i=1;
   gi = itab[i];
/*printf("Lev=%.3f R=%.2f i=%d G_Int=%.3f\n", lev, r, i, gi);*/

   // empiric koeff. for obj.cutting compensation
   if (lev>0.7)
      kf =  0.744*sqrt(1.-lev);
   else if (lev>0.3)
      kf = 0.9144-0.72727*lev;
   else
      kf = 1.1515-0.78945*sqrt(lev+0.03781);

   fx = sx * sqrt(fabs(spx/s)) * fw1 / kf;
   fy = sy * sqrt(fabs(spy/s)) * fw1 / kf;
   f = sqrt((fx*fx+fy*fy)/2.);
   sprintf(tmp,"FWHM_X=%.2f   FWHM_Y=%.2f",fx, fy);
   fwhm_x=fx; fwhm_y=fy;
   tvg->add_msg(tmp);
   graph->next_fwhmxy(fx,fy);

   sgm = f/fw1;
   icent = s/gi/((sgm/sx)*(sgm/sy));
   sprintf(tmp,"FWHM=%.2f   ICENT/Nadd=%.2f\n", f, icent/nadd );
   tvg->add_msg(tmp);
   graph->next_fwbr(f, icent/nadd/10.);

   return(TRUE);
}
void GDimage::fix_center()      // set new reference center from current coordinates
{
    x_c = mouse_x;
    y_c = mouse_y;
    tv_chan_point(x_c - (im_w-fr_w)/2 + col0, y_c - (im_h-fr_h)/2 + lin0);
    tvg->set_corr(xc-x_c, yc-y_c);
    update();
}

void GDimage::resizeEvent( QResizeEvent * e )
{
    int w = e->size().width();
    int h = e->size().height();
    x_c += (w-fr_w)/2 - (im_w-fr_w)/2;
    y_c += (h-fr_h)/2 - (im_h-fr_h)/2;
    xc += (double)(w-im_w)/2.;
    yc += (double)(h-im_h)/2.;
    mouse_x = (int)(xc + 0.5);
    mouse_y = (int)(yc + 0.5);
/*printf("GDimage::resizeEvent(): %d,%d -> %d,%d\n",im_w,im_h,w,h);*/
    if(w*h > im_w*im_h) {
       if(mask!=NULL) free(mask);
       mask = (char *)malloc(w*h);
    }
    im_w = w;
    im_h = h;
}

void  GDimage::update() {
    update(1.0);
}
void  GDimage::update(double sdt) {
    static int t_last = 0;
    int t_curr = gdtime.elapsed();
    int msdt = (int)(sdt*1000.);
    if(abs(t_curr-t_last) < msdt)
       need_update = TRUE;
    else {
       need_update = FALSE;
       repaint(TRUE);
       t_last = t_curr;
    }
}
void  GDimage::cyclic_update() {
    if(need_update) {
       need_update = FALSE;
       repaint(TRUE);
    }
}

void GDimage::paintEvent( QPaintEvent * )
{
    float z = mtx.m11();
    int im_w2 = im_w/2;
    int im_h2 = im_h/2;
    int x0 = (int)((x_c - im_w2)*z + im_w2 + mtx.dx() + 0.5);
    int y0 = (int)((y_c - im_h2)*z + im_h2 + mtx.dy() + 0.5);
    int x1 = (int)((mouse_x - im_w2)*z + im_w2 + mtx.dx() + 0.5);
    int y1 = (int)((mouse_y - im_h2)*z + im_h2 + mtx.dy() + 0.5);
    int sgx = (tvg->flipx)? -1 : 1;
    int sgy = (tvg->flipy)? -1 : 1;
    int sgz = (tvg->flipz)? -1 : 1;      // N1 or N2?
    double sx= tvg->sx;
    double sy= tvg->sy;
    if (sx<0.001) sx=sy;
    if (sy<0.001) sy=sx;
    if (sx<0.001) sx=0.1;
    if (sy<0.001) sy=0.1;
    int dsq = (int)(sqrt(4.*obj_sqr/(sx*sy)/M_PI)*z+0.5);
    int dfx=(int)(z*fwhm_x/sx+0.5);
    int dfy=(int)(z*fwhm_y/sy+0.5);
    int dx, dy;
    double aa;
    double da, dz;
    double sn, cs;
    int x,y, xa=width()/2, ya=height()/2, r=(width()+height())/5;
    QColor c;
    QPen p;

    if(fr_w>0 && fr_h>0) {
       x = (int)((im_w-fr_w*z)/2. + mtx.dx() + 0.5);
       y = (int)((im_h-fr_h*z)/2. + mtx.dy() + 0.5);
       bitBlt( this, x, y, &tv_pm);// copy pixmap to window
       if(im_src==0)
	  p.setColor(white);
       else {
	  int col, lin, frw, frh;
	  tv_chan_get_frame(&col, &lin, &frw, &frh);
	  if(frw!=fr_w || frh!=fr_h) {
	     if(im_src==1) p.setColor(red);
	     else          p.setColor(blue);
	  }
       }
       lin.setPen(p);
       lin.drawRect(x, y, (int)(fr_w*z), (int)(fr_h*z));
    }
    sn = sin(tvg->aza);
    cs = cos(tvg->aza);

    c.setNamedColor("Red");
    p.setStyle(SolidLine);
    p.setColor(c);
    lin.setPen(p);

    x = (int)(xa + sgx*r*cs);                // +A axis
    y = (int)(ya - sgy*r*sn);
    lin.drawLine(xa,ya,x,y);
    dx = (int)(-7*cs - 3*sn)*sgx;
    dy = (int)(7*sn - 3*cs) *sgy;
    lin.drawLine(x+dx,y+dy,x,y);
    dx = (int)(-7*cs + 3*sn)*sgx;
    dy = (int)(7*sn + 3*cs) *sgy;
    lin.drawLine(x+dx,y+dy,x,y);
    dx = -(int)(5*cs + 5*sn) *sgx;
    dy = -(int)(-5*sn + 5*cs)*sgy;
    lin.drawText(x+dx, y+dy, "A");

    c.setNamedColor("Magenta");
    p.setStyle(SolidLine);
    p.setColor(c);
    lin.setPen(p);

    x = (int)(xa + sgz*sgx*r*sn);                // +Z axis
    y = (int)(ya + sgz*sgy*r*cs);
    lin.drawLine(xa,ya,x,y);
    dx = (int)(-3*cs - 7*sn)*sgz*sgx;
    dy = (int)(3*sn - 7*cs) *sgz*sgy;
    lin.drawLine(x+dx,y+dy,x,y);
    dx = (int)(3*cs - 7*sn) *sgz*sgx;
    dy = (int)(-3*sn - 7*cs)*sgz*sgy;
    lin.drawLine(x+dx,y+dy,x,y);
    dx = (int)(5*cs - 5*sn) *sgz*sgx;
    dy = (int)(-5*sn - 5*cs)*sgz*sgy;
    lin.drawText(x+dx, y+dy, "Z");

    if(tvg->flipz)
      aa = 2. * tvg->aza - tvg->ada;
    else
      aa = tvg->ada;
    sn = sin(aa);
    cs = cos(aa);

    c.setNamedColor("Black");
    p.setStyle(SolidLine);
    p.setColor(c);
    lin.setPen(p);

    x = (int)(xa - sgx*r*cs);                // +Alpha axis
    y = (int)(ya + sgy*r*sn);
    lin.drawLine(xa,ya,x,y);
    dx = (int)(7*cs - 3*sn) *sgx;
    dy = (int)(-7*sn - 3*cs)*sgy;
    lin.drawLine(x+dx,y+dy,x,y);
    dx = (int)(7*cs + 3*sn) *sgx;
    dy = (int)(-7*sn + 3*cs)*sgy;
    lin.drawLine(x+dx,y+dy,x,y);
    dx = -(int)(3*cs + 5*sn) *sgx;
    dy = -(int)(-3*sn + 5*cs)*sgy;
    lin.drawText(x+dx, y+dy, "a");

    x = (int)(xa - sgz*sgx*r*sn);                // +Delta axis
    y = (int)(ya - sgz*sgy*r*cs);
    lin.drawLine(xa,ya,x,y);
    dx = (int)(-3*cs + 7*sn)*sgz*sgx;
    dy = (int)(3*sn + 7*cs) *sgz*sgy;
    lin.drawLine(x+dx,y+dy,x,y);
    dx = (int)(3*cs + 7*sn) *sgz*sgx;
    dy = (int)(-3*sn + 7*cs)*sgz*sgy;
    lin.drawLine(x+dx,y+dy,x,y);
    dx = (int)(3*cs - 5*sn) *sgz*sgx;
    dy = (int)(-3*sn - 5*cs)*sgz*sgy;
    lin.drawText(x+dx, y+dy, "d");

    tvg->get_corr(&da, &dz);

    c.setNamedColor("Blue");
    p.setStyle(DashLine);
    p.setColor(c);
    lin.setPen(p);
    lin.drawLine(0,y0,width(),y0);
    lin.drawLine(x0,0,x0,height());
    if(dsq>2) {
       p.setStyle(SolidLine);
       lin.setPen(p);
       lin.drawEllipse(x0-dsq/2,y0-dsq/2,dsq,dsq);
    }
    c.setNamedColor("White");
    p.setStyle(DashDotLine);
    p.setColor(c);
    lin.setPen(p);
    lin.drawLine(0,y1,width(),y1);
    lin.drawLine(x1,0,x1,height());
    if(dfx>2&&dfy>2) {
//     p.setStyle(SolidLine);
       lin.setPen(p);
       lin.drawEllipse(x1-dfx/2,y1-dfy/2,dfx,dfy);
    }
    c.setNamedColor("Red");
    lin.setPen(c);
    char tmp[40];
    sprintf(tmp,"%+.2f",da);
    x = x1+2;
    y = y1+6;
    if (x>im_w-30) x=im_w-30;
    if (y>im_h-7) y=im_h-7;
    if (y<20) y=20;
    lin.drawText(x, y-8, tmp);

    c.setNamedColor("Magenta");
    lin.setPen(c);
    sprintf(tmp,"%+.2f",dz);
    lin.drawText(x, y+8, tmp);

}

void GDimage::mousePressEvent( QMouseEvent *e )
{
    float z = mtx.m11();
    int im_w2 = im_w/2;
    int im_h2 = im_h/2;
    int mouse_button = e->button();
    if ( mouse_button == RightButton ) {
       x_c = (int)((e->x() - im_w2 - mtx.dx())/z + im_w2 + 0.5);
       y_c = (int)((e->y() - im_h2 - mtx.dy())/z + im_h2 + 0.5);
       tv_chan_point(x_c - (im_w-fr_w)/2 + col0, y_c - (im_h-fr_h)/2 + lin0);

    } else if ( mouse_button == LeftButton ) {
       xc = mouse_x = (int)((e->x() - im_w2 - mtx.dx())/z + im_w2 + 0.5);
       yc = mouse_y = (int)((e->y() - im_h2 - mtx.dy())/z + im_h2 + 0.5);
    } else
       return;
    tvg->set_corr(xc-x_c, yc-y_c);

    //printf("\nmousePressEvent: %d %d %d\n",mouse_button,mouse_x,mouse_y);
    update();
}

GDgraph::GDgraph( QWidget *parent, const char *name )
    : QFrame( parent, name)
{
    t_scale = 5.0;   // 5sec -> 1pix
    v_scale = 0.08;  // 1"    -> 12.5pix

    need_update = TRUE;

    daz_pnt = 0;
    da_cur = dz_cur = 0.;
    caz_pnt = 0;
    ca_cur = cz_cur = 0.;
    bxy_pnt = 0;
    bx_cur = by_cur = 0.;
    fwxy_pnt = 0;
    fx_cur = fy_cur = 0.;
    fwbr_pnt = 0;
    fw_cur = br_cur = 0.;
    time_cur = 0.;
    time_open = 0.;
    time_head = 0.;
    save_fd  = NULL;
    strcpy(savedir,".");
    bta=NULL;
    init_head();

    for(int i=0; i<GD_BUFF_SIZE; i++) {
       bdaz[i].time = 0;
       bcaz[i].time = 0;
       bbxy[i].time = 0;
       bfwxy[i].time = 0;
       bfwbr[i].time = 0;
    }
    type = DAZ;

    sfile = new QLabel( this );
    sfile->setText( "File: No" );
    sfile->setGeometry( 0, 0, 200, 15 );

    lin.begin(this);
    //lin.setBackgroundMode(OpaqueMode);
    QColor c;
    c.setNamedColor("White");
    lin.setPen(c);
    lin.setRasterOp (CopyROP);
    //lin.setFont( QFont( "times", 12, QFont::Bold ) );
    lin.setFont( QFont( "Symbol", 14, QFont::Normal ) );

    QSlider * tsc = new QSlider(1,100,1,50,QSlider::Horizontal,this,"t_scale");
    connect( tsc, SIGNAL(sliderMoved(int)), this, SLOT(set_t_scale(int)) );
    tsc->setGeometry( 25, 340, 175, 17 );
    QToolTip::add( tsc, "Change Time-axis scaling" );
    QLabel * lb = new QLabel( tsc, "T:", this );
    lb->setGeometry( 10, 340, 15, 20 );

    QSlider * vsc = new QSlider(1,100,1,8,QSlider::Vertical,this,"v_scale");
    connect( vsc, SIGNAL(sliderMoved(int)), this, SLOT(set_v_scale(int)) );
    vsc->setGeometry( 0, 110, 17, 160 );
    QToolTip::add( vsc, "Change Values-axis scaling" );
    lb = new QLabel( vsc, "A:", this );
    lb->setGeometry( 0, 90, 20, 20 );
    lb = new QLabel( vsc, "Z:", this );
    lb->setGeometry( 0, 270, 20, 20 );

    cbtype = new QComboBox( FALSE, this, "graph type" );
    cbtype->setGeometry( 80, 15, 130, 25 );
    cbtype->insertItem( "dA / dZ" );
    cbtype->insertItem( "Corr.A / Corr.Z" );
    cbtype->insertItem( "X/Y Backgr.Shift" );
    cbtype->insertItem( "FWHM_X / FWHM_Y" );
    cbtype->insertItem( "FWHM / Icent*0.1" );
    connect( cbtype, SIGNAL(activated(int)), SLOT(set_type(int)) );
    QToolTip::add( cbtype, "Change Type of graphic Values" );
    set_type(DAZ);

    QPopupMenu *act = new QPopupMenu();
    CHECK_PTR( act );
    act->insertItem( "&Reset graphics",this, SLOT(clear_buf()));
    act->insertItem( "&Open new file",this, SLOT(open_save()));
    act->insertItem( "&Open stdout",this, SLOT(open_stdout()));
    act->insertItem( "&Close file",this, SLOT(close_save()));

    QMenuBar *menu = new QMenuBar( this );
    CHECK_PTR( menu );
    menu->insertItem( "&GraphMenu", act );
    menu->setMaximumWidth(75);
    menu->setMaximumHeight(25);
    menu->setGeometry( 5, 15, 70, 25 );

    grtime.start();
    gr_timer = new QTimer(this);
    connect( gr_timer, SIGNAL(timeout()), this, SLOT(cyclic_update()) );
    gr_timer->start( 1000, FALSE); // check need_update every second
				   // and call update() every 5 seconds
}

GDgraph::~GDgraph()
{
    lin.end();
}

int GDgraph::graph_time_elapsed() {
static int curr_ms = 0;
    int ms = grtime.elapsed();
    while (ms < curr_ms) ms += 24*3600;      // fix a bag in QTime::elapsed()
    return (curr_ms = ms);
}

void GDgraph::next_center(double da, double dz)
{
    time_cur=bdaz[daz_pnt].time = graph_time_elapsed();
    da_cur = bdaz[daz_pnt].da = da;
    dz_cur = bdaz[daz_pnt].dz = dz;
    daz_pnt++;
    if(daz_pnt>=GD_BUFF_SIZE) daz_pnt=0;
    bdaz[daz_pnt].time = 0;
    update();
}

void GDgraph::next_corrtn(double ca, double cz)
{
    time_cur=bcaz[caz_pnt].time = graph_time_elapsed();
    ca_cur += ca;
    cz_cur += cz;
    bcaz[caz_pnt].da = ca_cur;
    bcaz[caz_pnt].dz = cz_cur;
    caz_pnt++;
    if(caz_pnt>=GD_BUFF_SIZE) caz_pnt=0;
    bcaz[caz_pnt].time = 0;
    update();
}

void GDgraph::next_bgdxy(double bx, double by)
{
    time_cur=bbxy[bxy_pnt].time = graph_time_elapsed();
    bx_cur = bbxy[bxy_pnt].da = bx;
    by_cur = bbxy[bxy_pnt].dz = by;
    bxy_pnt++;
    if(bxy_pnt>=GD_BUFF_SIZE) bxy_pnt=0;
    bbxy[bxy_pnt].time = 0;
    update();
}

void GDgraph::next_fwhmxy(double fwx, double fwy) {
    time_cur=bfwxy[fwxy_pnt].time = graph_time_elapsed();
    fx_cur = bfwxy[fwxy_pnt].da = fwx;
    fy_cur = bfwxy[fwxy_pnt].dz = fwy;
    fwxy_pnt++;
    if(fwxy_pnt>=GD_BUFF_SIZE) fwxy_pnt=0;
    bfwxy[fwxy_pnt].time = 0;
    update();
}

void GDgraph::next_fwbr(double fw,   double br) {
    time_cur=bfwbr[fwbr_pnt].time = graph_time_elapsed();
    fw_cur = bfwbr[fwbr_pnt].da = fw;
    br_cur = bfwbr[fwbr_pnt].dz = br;
    fwbr_pnt++;
    if(fwbr_pnt>=GD_BUFF_SIZE) fwbr_pnt=0;
    bfwbr[fwbr_pnt].time = 0;
    update();
}

void GDgraph::open_save() {
  char savename[100];
  char *p, filename[40];
  QString fname;

  fname = QFileDialog::getOpenFileName(savedir, "*.tgd",this);
  if ( fname.isEmpty() ) return;

  if((p=strrchr(fname.data(),'/'))==NULL)
     strncpy(filename,fname.data(),40);
  else
     strncpy(filename,p+1,40);
  if(strchr(filename,'.'))
    sprintf(savename,"%s", fname.data());
  else
    sprintf(savename,"%s.tgd", fname.data());
  fname.truncate( fname.findRev( '/', fname.length()-1 ) );
  strncpy(savedir,fname.data(),100);
  set_head("SaveDir",savedir);

  open_file(savename);
}

void GDgraph::open_stdout() {
  open_file("stdout");
}

void GDgraph::open_file(char *savename) {
  char filename[40];
  char *p, tmp[100];

  if(save_fd != NULL)
     close_save();
  if(strcmp(savename,"stdout")==0) {
     strncpy(filename,savename,40);
     save_fd = stdout;
  } else {
     if((p=strrchr(savename,'/'))==NULL)
	strncpy(filename,savename,40);
     else
	strncpy(filename,p+1,40);
     if((save_fd=fopen(savename,"w")) == NULL) {
	sprintf(tmp, "File: Failed!?");
	sfile->setText( tmp );
	return;
     }
  }
  sprintf(tmp, "File: %s", filename);
  sfile->setText( tmp );

  print_head();
  fprintf(save_fd,"============================\n");
  if(bta!=NULL && bta->connected()) {
     print_acs();
     fprintf(save_fd,"============================\n");
  }
  fprintf(save_fd,"Time\tAcent\tZcent\tAcorr\tZcorr\tFWHM_X\tFWHM_Y\tFWHM\tIcent\tBgd_X\tBgd_Y\n");

  time_open = time_head = graph_time_elapsed();
}

void GDgraph::init_head() {
   int n;
   struct ParamList *hlp;

   for (n=0,hlp=head; n<30; n++, hlp++)
      hlp->name[0] = hlp->val[0] = 0;
}

void GDgraph::set_head(char *name, char *val) {
   int n;
   struct ParamList *hlp;

   for (n=0,hlp=head; n<30 && hlp->name[0]; n++, hlp++) {
      if (strcmp(hlp->name, name) == 0) {
	 strncpy(hlp->val,val,99);
	 return;
      }
   }
   if (n<30) {
       strncpy(hlp->name,name,15);
       strncpy(hlp->val,val,99);
   }
}
void GDgraph::set_head(char *name, int ival) {
   char tmp[32];
   sprintf(tmp,"%d",ival);
   set_head(name,tmp);
}
void GDgraph::set_head(char *name, double fval) {
   char tmp[32];
   sprintf(tmp,"%.3f",fval);
   set_head(name,tmp);
}
char *GDgraph::get_head(const char *name, char val[]) {
   int n;
   struct ParamList *hlp;

   for (n=0,hlp=head; n<30 && hlp->name[0]; n++, hlp++) {
      if (strcmp(hlp->name, name) == 0) {
	 strncpy(val,hlp->val,99);
	 return(&val[0]);
      }
   }
   return(NULL);
}

void GDgraph::print_head() {
   int n;
   struct ParamList *hlp;

   for (n=0,hlp=head; n<30 && hlp->name[0]; n++, hlp++)
      fprintf(save_fd,"%s=%s\n",hlp->name,hlp->val);
}

void GDgraph::print_acs() {
  int n;
  if(bta!=NULL && bta->connected()) {
     fprintf(save_fd,"ACS=%s\n",bta->Get("ASU"));
     n = atoi(bta->Get("TelStat"));
     if(n<0 || n>4)  n=0;
     fprintf(save_fd,"TelStat=%s\n", (n<2)? "Stopping": ((n==2)? "Pointing" : "Tracking"));
     n = atoi(bta->Get("TelFoc"));
     if(n<0 || n>2)  n=0;
     fprintf(save_fd,"TelFocus=%s\n", (n==2)? "N2": ((n==1)? "N1" : "CPF"));
     fprintf(save_fd,"Alpha=%s\n",bta->Get("Alpha"));
     fprintf(save_fd,"Delta=%s\n",bta->Get("Delta"));
     fprintf(save_fd,"Azimuth=%s\n",bta->Get("Azimuth"));
     fprintf(save_fd,"ZenDist=%s\n",bta->Get("ZenDist"));
     fprintf(save_fd,"A_Speed=%s\n",bta->Get("AzVel"));
     fprintf(save_fd,"Z_Speed=%s\n",bta->Get("ZdVel"));
     fprintf(save_fd,"ParAngle=%s\n",bta->Get("PA"));
     fprintf(save_fd,"P2=%s\n",bta->Get("P2"));
     fprintf(save_fd,"Mtime=%s\n",bta->Get("Mtime"));
     fprintf(save_fd,"Stime=%s\n",bta->Get("Stime"));
     fprintf(save_fd,"Focus=%s\n",bta->Get("Focus"));
     fprintf(save_fd,"Dome=%s\n",bta->Get("Dome"));
     fprintf(save_fd,"Press=%s\n",bta->Get("Press"));
     fprintf(save_fd,"Tempr=%s\n",bta->Get("T-out"));
  }
}

void GDgraph::close_save() {
    if(save_fd != NULL) {
       fprintf(save_fd,"============================\n");
       print_acs();
       if (save_fd != stdout)
	  fclose(save_fd);
       save_fd = NULL;
       sfile->setText( "File: No" );
    }
}

void GDgraph::flush_save() {
    if(save_fd != NULL) {
       if((time_cur-time_head)/1000./60. > 5. && bta!=NULL && bta->connected()) {
	  fprintf(save_fd,"============================\n");
	  print_acs();
	  fprintf(save_fd,"============================\n");
	  time_head = time_cur;
       }
       fprintf(save_fd,"%.1f\t%.2f\t%.2f\t%.2f\t%.2f\t%.1f\t%.1f\t%.1f\t%.1f\t%.2f\t%.2f\n",
	       (time_cur-time_open)/1000.,
	       da_cur,dz_cur,ca_cur,cz_cur,fx_cur,fy_cur,fw_cur,br_cur,bx_cur,by_cur);
       fflush(save_fd);
    }
}

void GDgraph::set_t_scale(int scr_val)
{
    t_scale = (double)scr_val/10.;
    update(0.2);
}
void GDgraph::set_v_scale(int scr_val)
{
    v_scale = (double)scr_val/100.;
    update(0.2);
}
void GDgraph::set_type(int new_type) {
    type = new_type;
    update();
}
void GDgraph::clear_buf() {
    daz_pnt = 0;
    da_cur=dz_cur=0.;
    caz_pnt = 0;
    ca_cur=cz_cur=0.;
    bxy_pnt = 0;
    bx_cur=by_cur=0.;
    fwxy_pnt = 0;
    fx_cur=fy_cur=0.;
    fwbr_pnt = 0;
    fw_cur=br_cur=0.;
    for(int i=0; i<GD_BUFF_SIZE; i++) {
       bdaz[i].time = 0;
       bcaz[i].time = 0;
       bbxy[i].time = 0;
       bfwxy[i].time = 0;
       bfwbr[i].time = 0;
    }
    update();
}


void  GDgraph::update() {
    update(0.5);
}
void  GDgraph::update(double sdt) {
    static int t_last = 0;
    int t_curr = graph_time_elapsed();
    int msdt = (int)(sdt*1000.);
    if(t_curr-t_last < msdt)
       need_update = TRUE;
    else {
       need_update = FALSE;
       repaint(TRUE);
       t_last = t_curr;
    }
}
void  GDgraph::cyclic_update() {
    static int t_last = 0;
    int t_curr = graph_time_elapsed();
    if(need_update || t_curr-t_last>5000) {
       need_update = FALSE;
       repaint(TRUE);
       t_last = t_curr;
    }
}

void  GDgraph::paintEvent( QPaintEvent * )
{
    draw_graph( 127, TRUE);
    draw_graph( 252, FALSE);
}

void  GDgraph::draw_graph( int y0,  bool az)
{
    int x,x0,xe,y;
    int pnt,bpnt;
    int t_curr = graph_time_elapsed();
    double v;
    struct GDbuff_val *buff;

    switch (type) {
       default:
       case DAZ:   buff = bdaz;
		   bpnt = daz_pnt;
		   break;
       case CORAZ: buff = bcaz;
		   bpnt = caz_pnt;
		   break;
       case BGDXY: buff = bbxy;
		   bpnt = bxy_pnt;
		   break;
       case FWXY:  buff = bfwxy;
		   bpnt = fwxy_pnt;
		   break;
       case FWBR:  buff = bfwbr;
		   bpnt = fwbr_pnt;
		   break;
    }

    QColor c;
    QPen p;
    c.setNamedColor("Black");
    lin.setPen(c);
    p.setColor(c);
    x0 = width()-2;
    xe = 20;
    for(v=-5.; v<5.1; v+=1.) {
       y = y0 - (int)(v/v_scale);
       p.setStyle(SolidLine);
       lin.setPen(p);
       lin.drawLine(xe-1, y, xe+2, y);
       lin.drawLine(xe, y-3, xe, y+3);
       lin.drawLine(x0-2, y, x0+1, y);
       lin.drawLine(x0, y-3, x0, y+3);
       p.setStyle(DotLine);
       lin.setPen(p);
       lin.drawLine(xe, y, x0, y);
    }
    lin.setPen(c);
    lin.drawLine(xe,y0,x0,y0);
    for (int t=(t_curr/30000)*30000; ; t -= 30000) {
       x = x0 - (int)((t_curr - t)/1000./t_scale + 0.5);
       if (x<xe) break;
       if ((t%600000)==0)
	  lin.drawLine(x, y0-5, x, y0+5);
       else if ((t%60000)==0)
	  lin.drawLine(x, y0-2, x, y0+2);
       else
	  lin.drawLine(x, y0-1, x, y0+1);
    }
    if (az)
       c.setNamedColor("Red");
    else
       c.setNamedColor("Magenta");
    lin.setPen(c);
    pnt = bpnt-1;
    if(pnt<0) pnt = GD_BUFF_SIZE-1;
    for(int i=0; i<GD_BUFF_SIZE; i++) {
       if (pnt == bpnt || buff[pnt].time == 0)
	  break;
       x = x0 - (int)((t_curr - buff[pnt].time)/1000./t_scale + 0.5);
       v = (az)? buff[pnt].da: buff[pnt].dz;
       y = y0 - (int)(v/v_scale);
       if (i==0) {
	  lin.moveTo(x0, y);
	  lin.lineTo(x, y);
       } else
	  lin.lineTo(x, y);
       if (x<xe)
	  break;
       pnt--;
       if(pnt<0) pnt = GD_BUFF_SIZE-1;
    }
}

#include <string.h>
#include <sys/ioctl.h>
#include <sys/file.h>
#include <sys/types.h>          /* generic definitions */
#include <sys/socket.h>         /* sockets definitions */
#include <sys/stat.h>           /* stat(2v); */
#include <netinet/in.h>         /* struct sockaddr_in (Internet domain) */
#include <arpa/inet.h>          /* inet_ntoa() & etc...(Internet address manipulation) */
#include <netdb.h>              /* getservbyname(3n), gethostbyname(3n); */
#include <sys/time.h>           /* struct timeval timeout */


const char *serv_name_local = "tb";
const char *serv_name_full  = "tb.bta.sao.ru";
const char *serv_ip_addr    = "192.168.3.11";
const char *port  = "7654";    /* default TCP-port number or a service name */
const int  netmask = 0xfffffe00;   /* 255.255.254.0 for 192.168.2+192.168.3 */

/* open socket & connect to bta_serv*/
BTAserv::BTAserv(QObject * parent, const char * name)
    : QObject( parent, name)
{
  char  myhost[128];
  int   my_ip, serv_ip;
  struct hostent     *hp,hs;
  struct servent     *sp,ss;
  struct protoent    *ps;
  struct sockaddr_in serv;
  char  str[100];
//extern int connect(int  sockfd, struct sockaddr *serv_addr, int addrlen );

  IniList();
  mtime=M_time;
  sleep(1);
  new_acs = fabs(M_time-mtime)>0.01;
  old_acs = FALSE;
  if(new_acs)
     printf("New ACS Connection Ok!\n");

  if(gethostname(myhost, sizeof(myhost)) < 0) {
      printf("Can't get my own host name!?.\n");
      exit(1);
  }
  if ((hp = gethostbyname(myhost)) != NULL) {
      struct in_addr in;
      my_ip = in.s_addr = *(int*)hp->h_addr;
      printf("My host: %s (%s)\n", myhost, inet_ntoa(in));
  } else {
      printf("Can't get my own host IP-addr!?.\n");
      exit(1);
  }
  if ((hp = gethostbyname(serv_name_local)) != NULL) {
      serv_ip = *(int*)hp->h_addr;
      serv_name = serv_name_local;
  } else if ((hp = gethostbyname(serv_name_full)) != NULL) {
      serv_ip = *(int*)hp->h_addr;
      serv_name = serv_name_full;
  } else {
      serv_ip = inet_addr(serv_ip_addr);
      if ((hp = gethostbyaddr((char *)&serv_ip, 4, AF_INET)) != NULL) {
	 if (hp->h_name != NULL)
	    serv_name = hp->h_name;
	 else
	    serv_name = serv_ip_addr;
      } else {
	 hs.h_name = serv_name_full;
	 hs.h_aliases = NULL;
	 memcpy(hs.h_addr, &serv_ip, 4);
	 hs.h_length = 4;
	 hs.h_addrtype = AF_INET;
	 hp = &hs;
	 serv_name = serv_ip_addr;
      }
  }

  /* "port" could indicate a port number or a service */
  if ( (ss.s_port = atoi(port)) == 0) {
     /* get service reference */
     if ((sp = getservbyname(port,"tcp")) == NULL) {
	printf("Unable to find service '%s'\n",port);
	exit(1);
     }
  } else {
     ss.s_proto  = "tcp";
     ss.s_aliases= NULL;
     ss.s_name   = "bta_serv";
     ss.s_port   = htons(ss.s_port);
     sp = &ss;
  }

  if (!(ps = getprotobyname(sp->s_proto))) {
     printf("Unable to recognize protocol '%s'\n",port);
     exit(1);
  }

  if((ns = socket(hp->h_addrtype, SOCK_STREAM, ps->p_proto)) < 0) {
     printf("Could not create socket!\n");
     exit(1);
  }

  memset((char *)&serv, 0, sizeof(serv));
  memcpy(&serv.sin_addr, hp->h_addr, hp->h_length);
  serv.sin_family = hp->h_addrtype;
  serv.sin_port = sp->s_port;
  if (::connect(ns, (struct sockaddr *)&serv, sizeof(serv))<0) {
     struct in_addr in;
     in.s_addr = serv_ip;
     printf("Could not connect to Old ACS server at %s (%s)\n",serv_name, inet_ntoa(in));
     close(ns);
     ns = -1;
  }
  if (ns>=0) {
     printf("Old ACS Connection to BTAserver: %s\n",serv_name);
     write_str("Ver\n");
     if(read_str(str, 200)<1) {
	printf("Really NO Connection!\n Please set ACS data values manually!\n");
	return;
     }
     puts(str);

     sprintf(str,"Log TVguide from %s\n", myhost);
     write_str(str);
     read_str(str,200);
/*
 *    write_str("Passwd camac\n");
 *    read_str(str,200);
 *    puts(str);
 */
     old_acs = TRUE;
  }
  if(new_acs || old_acs) {
     timer = new QTimer(this);
     connect( timer, SIGNAL(timeout()), this, SLOT(GetList()) );
     //timer->start( 2000, TRUE);             // call GetList() in 2 second
     GetList();
  }
}

/* disconnect from bta_serv & close socket*/
BTAserv::~BTAserv()
{
    if(ns>=0) {
       close(ns);
       printf("DisConnecting from bta_serv...\n");
       ns = -1;
    }
    if(new_acs) {
       printf("DisConnecting from New ACS...\n");
    }
}

bool BTAserv::connected()
{
    if(check_shm_block(&sdat)) return(TRUE);
    return( ns >= 0 );
}

/* write to socket */
int BTAserv::write_blk(char *buf, int len)
{
  int nb, nfd = 1;
  struct timeval timeout;
  double tout = 5.0;
  fd_set writefds;

/*printf("BTAserv::write_blk(%s,%d)\n",buf, len);*/
    if (!connected())
	    return(-1);
    if(len <= 0) return(0);
    do{
do_write:
	nb = write(ns, buf, len);
	if(nb < 0) {
	    if(errno == EINTR)                   /* maybe SIGALRM occure? */
		goto do_write;
	    if(errno != EAGAIN){
		printf("Could not send block. %s\n",strerror(errno));
		return(-1);
	    }
	    nb = 0;
	} else if(nb > 0) {
	   /*    printf("Sending %d bytes to socket <",nb);
		fflush(stdout);
		write(1, buf, nb);
		printf(">\n");
	    */
	    len -= nb;
	    buf += nb;
	}
do_write_select:
	FD_ZERO(&writefds);
	FD_SET(ns, &writefds);
	timeout.tv_sec = (int)tout;
	timeout.tv_usec = (int)((tout-timeout.tv_sec)*1000000.);

	nfd = select(ns+1, (fd_set *)NULL, &writefds, (fd_set *)NULL, &timeout);
	if(nfd <= 0) {
	    if(nfd < 0) {                   /* nfd < 0 -> error */
		if(errno != EINTR) {
		    printf("Error in select. %s\n",strerror(errno));
		    sleep(1);
		} else                      /* maybe SIGALRM occure? */
		    goto do_write_select;
	    } else                          /* nfd == 0 -> timeout */
		printf("Timeout writing to socket.\n");
	    return(-1);
	} else {                            /* nfd > 0 -> can write */
	    if(!FD_ISSET(ns, &writefds))
		printf("Paranoia check: Select but can't write?!\n");
	}
    } while(len>0 && nfd>0);
    return(0);
}
int BTAserv::write_str(char *str) {
    return( write_blk(str, strlen(str)) );
}

/* read from socket */
int BTAserv::read_blk(char *buff, int len)
{
  static char rbuf[2000];
  static int nbuf=0;
  int i,l,nb, nfd = 1;
  struct timeval timeout;
  double tout;
  fd_set readfds;

    if (!connected())
	    return(-1);

   if (nbuf>0 && len<=nbuf)  goto write_out;

   tout = (nbuf>0)? 0.1 : 6.33;

   for( nb=0; nb<=0; ) {
do_read_select:
      FD_ZERO(&readfds);
      FD_SET(ns, &readfds);
      timeout.tv_sec = (int)tout;
      timeout.tv_usec = (int)((tout-timeout.tv_sec)*1000000.);

      nfd = select(ns+1, &readfds, (fd_set *)NULL, (fd_set *)NULL, &timeout);
      if(nfd <= 0) {
	  if(nfd < 0) {                   /* nfd < 0 -> error */
	      if(errno != EINTR) {
		  printf("Error in select?! %s\n",strerror(errno));
		  return(-1);
	      } else                      /* maybe SIGALRM occure? */
		  goto do_read_select;
	  } else {                        /* nfd == 0 -> timeout */
	      if (nbuf > 0)
		 goto write_out;
	      printf("Read socket timeout! Connection closed...\n");
	      close(ns);
	      ns = -1;
	      return -1;
	 }
      } else {                            /* nfd > 0 -> can read */
	  if(!FD_ISSET(ns, &readfds))
	      printf("Paranoia check: Select but can't read?!\n");
      }
do_read:
      l = 1000-nbuf;
      nb = read(ns, &rbuf[nbuf], l);
      if ( nb < 0 ) {
	 if (errno == EINTR) goto do_read;    /* maybe SIGALRM occure? */
	 if (errno != EPIPE)
	    printf("Read socket error (%s)\n", strerror(errno));
	 nb=0;
	 //if (errno == EAGAIN||errno == EBUSY) goto do_read_select; /* ???? */
      }
      if ( nb == 0 ) {
	 printf("Connection closed by foreign host (%s)\n", serv_name);
	 close(ns);
	 ns = -1;
	 return -1;
      }
   }
   nbuf += nb;
write_out:
   if (nbuf < len) len = nbuf;
   for(l=0; l<len; )
       *buff++ = rbuf[l++];
   for(i=0; l<nbuf;)
       rbuf[i++] = rbuf[l++];
   nbuf -= len;
   return len;
}

int BTAserv::read_str(char *buff, int len)
{
  int cnt,n;
/*char *b=buff;*/
   for (cnt=0; (n=read_blk(buff, 1)) == 1 && cnt<len-1; cnt++,buff++)
      if( *buff == '\0' || *buff == '\n') break;
   *buff = '\0';
/*printf("BTAserv::read_str(%s)=%d\n",b, cnt);*/
   return(cnt>0? cnt : n);
}

void BTAserv::IniList() {
   struct ParamList *blp=par_list;                                        //[20]
   strcpy(blp->name,"ASU");      strcpy(blp->val,"Disconn.");   blp++;  // 0
   strcpy(blp->name,"TelStat");  strcpy(blp->val,"0");          blp++;  // 1
   strcpy(blp->name,"TelFoc");   strcpy(blp->val,"0");          blp++;  // 2
   strcpy(blp->name,"Alpha");    strcpy(blp->val,"00 00 00.0"); blp++;  // 3
   strcpy(blp->name,"Delta");    strcpy(blp->val,"+00 00 00.0");blp++;  // 4
   strcpy(blp->name,"Azimuth");  strcpy(blp->val,"0.0");        blp++;  // 5
   strcpy(blp->name,"ZenDist");  strcpy(blp->val,"80.0");       blp++;  // 6
   strcpy(blp->name,"AzVel");    strcpy(blp->val,"000 00 00.0");blp++;  // 7
   strcpy(blp->name,"ZdVel");    strcpy(blp->val,"00 00 00.0"); blp++;  // 8
   strcpy(blp->name,"PA");       strcpy(blp->val,"0.0");        blp++;  // 9
   strcpy(blp->name,"P2");       strcpy(blp->val,"0.0");        blp++;  // 10
   strcpy(blp->name,"Mtime");    strcpy(blp->val,"00 00 00.0"); blp++;  // 11
   strcpy(blp->name,"Stime");    strcpy(blp->val,"00 00 00.0"); blp++;  // 12
   strcpy(blp->name,"Focus");    strcpy(blp->val,"0.0");        blp++;  // 13
   strcpy(blp->name,"Dome");     strcpy(blp->val,"0.0");        blp++;  // 14
   strcpy(blp->name,"Press");    strcpy(blp->val,"585.0");      blp++;  // 15
   strcpy(blp->name,"T-out");    strcpy(blp->val,"0.0");        blp++;  // 16
   blp->name[0] = '\0';          blp->val[0] = '\0';                    // 17
}

bool BTAserv::GetList()
{
   char buff[2000];
   char str[1000], *par, *p;
   struct ParamList *blp;

   if(!connected())  return(FALSE);

   new_acs = fabs(M_time-mtime)>0.01;

   if(new_acs) {                    //New ACS??
      int tel_st=Tel_State;
      if(tel_st!=0) tel_st++;
      if(Sys_Mode==SysTrkOk) tel_st++;
      for (blp=par_list; blp->name[0]; blp++) {
	 if(strcmp(blp->name,"ASU")==0)          strcpy(blp->val,"Connect");
	 else if(strcmp(blp->name,"TelStat")==0) sprintf(blp->val,"%1d",tel_st);
	 else if(strcmp(blp->name,"TelFoc")==0)  sprintf(blp->val,"%1d",Tel_Focus);
	 else if(strcmp(blp->name,"Alpha")==0)   time_asc(CurAlpha, blp->val);
	 else if(strcmp(blp->name,"Delta")==0)   angle_asc(CurDelta, blp->val);
	 else if(strcmp(blp->name,"Azimuth")==0) sprintf(blp->val,"%+7.2f",val_A/3600.);
	 else if(strcmp(blp->name,"ZenDist")==0) sprintf(blp->val,"%5.2f",val_Z/3600.);
	 else if(strcmp(blp->name,"AzVel")==0)   angle_asc(vel_A, blp->val);
	 else if(strcmp(blp->name,"ZdVel")==0)   angle_asc(vel_Z, blp->val);
	 else if(strcmp(blp->name,"PA")==0)      sprintf(blp->val,"%6.2f",tag_P/3600.);
	 else if(strcmp(blp->name,"P2")==0)      sprintf(blp->val,"%6.2f",val_P/3600.);
	 else if(strcmp(blp->name,"Mtime")==0)   time_asc(M_time, blp->val);
	 else if(strcmp(blp->name,"Stime")==0)   time_asc(S_time, blp->val);
	 else if(strcmp(blp->name,"Focus")==0)   sprintf(blp->val,"%6.2f",val_F);
	 else if(strcmp(blp->name,"Dome")==0)    sprintf(blp->val,"%+7.2f",val_D/3600.);
	 else if(strcmp(blp->name,"Press")==0)   sprintf(blp->val,"%5.1f",val_B);
	 else if(strcmp(blp->name,"T-out")==0)   sprintf(blp->val,"%+5.1f",val_T1);
      }
   } else if(old_acs) {               //try OLD ACS...
      strcpy(str,"Get");
      for (blp=par_list; blp->name[0]; blp++) {
	 strcat(str, " ");
	 strcat(str, blp->name);
      }
      strcat(str, "\n");
      write_str(str);
      if(read_str(buff, 2000)<1)  return(FALSE);
      blp=par_list;
      for(par=strtok(buff," \t\n"); par!=NULL; par=strtok(NULL," \t\n")) {
	 if ((p=strchr(par, '=')) != NULL) {
	    *p++ = '\0';
	    if (strcmp(par,"Error") == 0) {
printf("BTA Get Error: %s\n",p);
	       return(FALSE);
	    } else if (strcmp(blp->name, par) == 0) {
	       strncpy(blp->val, p, 24);
	    } else {
	       for (blp=par_list; blp->name[0]; blp++) {
		  if (strcmp(blp->name, par) == 0) {
		     strncpy(blp->val, p, 20);
		     break;
		  }
	       }
	    }
	    for(p=blp->val; *p != '\0'; p++)
	       if(*p == '_')  *p = ' ';
/*printf("%s=%s\n",blp->name,blp->val);*/
	    blp++;
	 }
      }
   }
   timer->start( 2000, TRUE);             // call GetList() every 2 second
   mtime=M_time;
   return(TRUE);
}

char * BTAserv::Find(const char *name)
{
   struct ParamList *blp=par_list;

   for (blp=par_list; blp->name[0]; blp++) {
      if (strcmp(blp->name, name) == 0)
	 return(blp->val);
   }
   return(NULL);
}

char * BTAserv::Get(const char *name)
{
   static char buff[100];
   char str[200], *p;

   if((p=Find(name)) != NULL )
      return(p);
   strcpy(str,"Get ");
   strcat(str, name);
   strcat(str, "\n");
   write_str(str);
   read_str(str, 200);
   if ((p=strchr(str, '=')) != NULL)
      strcpy(buff, strtok(p,"= \n"));
   else
      strcpy(buff, "?????");
   for(p=buff; *p != '\0'; p++)
      if(*p == '_')  *p = ' ';
   return(buff);
}

bool BTAserv::Put(const char *name, const char *value)
{
   char str[200];
   sprintf(str,"Put %s %s\n", name, value);
   write_str(str);
   read_str(str, 200);
   if (strncmp(str,"Ok",2)) {
      printf("Sending %s fault (%s)\n",name, str);
      return(FALSE);
   } else
      return(TRUE);
}
bool BTAserv::ACSconnected()
{
   if (!connected()) {
      printf("No connection with `bta_serv` at %s (Network problems!?)\n",serv_name_full);
      return(FALSE);
   }
   if(fabs(M_time-mtime)>0.01)        //New ACS??
      return(TRUE);
   if (strncmp(Get("ASU"),"Conn",4)) {
      printf("No connection with ACS BTA (maybe turned off...)\n");
      return(FALSE);
   } else
      return(TRUE);
}

char * BTAserv::time_asc(double t, char *lin)
{
    int h, min;
    double sec;
    h   = (int)(t/3600.);
    min = (int)((t - (double)h*3600.)/60.);
    sec = t - (double)h*3600. - (double)min*60.;
    sprintf(lin, "%02d %02d %05.2f", h,min,sec);
    return lin;
}

char * BTAserv::angle_asc(double a, char *lin)
{
    char s;
    int d, min;
    double sec;
    if (a >= 0.)
       s = '+';
    else {
       s = '-'; a = -a;
    }
    d   = (int)(a/3600.);
    min = (int)((a - (double)d*3600.)/60.);
    sec = a - (double)d*3600. - (double)min*60.;
    if(d>99)
       sprintf (lin, "%c%3d %02d %04.1f", s,d,min,sec);
    else
       sprintf (lin, "%c%02d %02d %04.1f", s,d,min,sec);
    return lin;
}

double BTAserv::asc_time(char *str) {
 int h=0, min=0;
 double val=0.0, sec=0.0;
   if (str != NULL) {
      if(sscanf(str,"%d%d%lf",&h,&min,&sec) > 0) {
	 if (h >= 0) {
	    val = (double)h*3600. + (double)min*60. + sec;
	    if (val >= 24.*3600.)  val -= 24.*3600.;
	    if (val < 0.) val += 24.*3600.;
	 }
      }
   }
   return (val);
}

double  BTAserv::asc_angle(char *str) {
 int sgn, gr=0, min=0;
 double val=0.0, sec=0.0;
   if (str != NULL) {
      if(sscanf (str,"%d%d%lf",&gr,&min,&sec) > 0) {
	 if (gr > 0)
	    sgn = 1;
	 else if (gr == 0) {
	    if (strchr(str, '-'))
		sgn = -1;
	    else
		sgn = 1;
	 } else {
	    sgn = -1; gr = -gr;
	 }
	 val = ((double)gr*3600. + (double)min*60. + sec) * (double)sgn;
      }
   }
   return (val);
}


int main( int argc, char **argv )
{
    QApplication app( argc, argv );

    get_shm_block( &sdat, ClientSide);
    get_cmd_queue( &ucmd, ClientSide);

    BTAserv bta;

    if(bta.connected()) {
       if( bta.ACSconnected() ) printf("ACS BTA - Ok!\n");
    } else {
       printf("Can't connect to BTAserver\n");
    }

    TVguide tvg;

    tvg.bta = &bta;     // connect BTAserv to TVguide
    tvg.graph->bta = &bta;

    app.setMainWidget( &tvg );
    tvg.show();

    return app.exec();
}


#include "tvguide.moc"                  // include metadata generated by the moc
