/*
 *       TV- 
 *     VS56  VS60.
 *    QT-.
 */
#include <qwidget.h>
#include <qmenubar.h>
#include <qpainter.h>
#include <qstring.h>
#include <qlabel.h>
#include <qpixmap.h>
#include <qslider.h>
#include <qtabdlg.h>
#include <qlined.h>
#include <qchkbox.h>

// maximum image sizes
#define maxW  768
#define maxH  576
#define maxL  maxW*maxH

// mouse usage modes
#define NO_MOUSE  0
#define MOD_CUTS  1
#define SET_FRAME 2
#define SET_LINE  3
#define SET_POINT 4

class TVimage : public QWidget
{
    Q_OBJECT
public:
    TVimage( QWidget *parent=0, const char *name=0 );
    ~TVimage();

public slots:
    void        setChan( int n);
    void        setTagWnd( QWidget * );
    void        show_help();
    void        html_manual();
    void        save_to_file(int file_format);
    void        set_bias_name();
    void        set_flat_name();
    void        subtr_bias();
    void        div_by_flat();
    void        bias_flat_red();
    void        save_fits();
    void        load_fits();
    void        single();
    void        get10();
    void        cont();
    void        stop();
    void        add1();
    void        add10();
    void        add50();
    void        add100();
    void        addcont();
    void        init();
    void        clear();
    void        write();
    void        calc_flat_row();
    void        use_flat_row(int oper);
    void        togle_ccd_mode();
    void        set_ccd_mode(bool new_mode);
    void        togleFormat();
    void        togleBar();
    void        resetFrame(int);
    void        setMouseMode(int);
    void        getLUT();
    void        getLUT(const char *lname);
    void        getITT();
    void        getITT(const char *iname);
    void        startExp();
    void        startElev();
    void        startElev2();

private slots:
    void        tmout();
    void        getExp(int);
    void        stopExp();
    void        getElev(int);
    void        getElev1(int);
    void        getElev2(int);
    void        stopElev();
    void        stopElev1();
    void        stopElev2();
    void        start_cycle();
    void        step_cycle();
    void        destr_opt();    // fix that "Options" dialog window was destroyed
    void        apply_opt();    // Use new (edited) options
    void        setup_opt();    // Show "Options" dialog panel
    void        toggle_bias();
    void        toggle_flat();
    void        stat_frame();
    void        calc_cuts();
    void        togle_autocuts_mode();
    void        edit_fits();
    void        destr_fits();
    void        edit_fits_ok();
    void        edit_fits_apply();
    void        edit_fits_cancel();
    void        edit_fits_refresh();

protected:
    void  paintEvent( QPaintEvent * );
    void  mousePressEvent( QMouseEvent * );
    void  mouseReleaseEvent( QMouseEvent * );
    void  mouseMoveEvent( QMouseEvent * );

private:
    void nodrv_mesg();
    void show_text(const char *);
    void draw_graph(void);
    void show_image(bool);
    void copy_image(void);
    void stat_image(bool set_cuts);
    void set_autocuts_mode(bool new_mode);
    bool get_fits(char *, short *, int *, int *, int *);
    int nchan;
    int cutTop;        // Up&Low visualisation CUTs
    int cutBot;
    bool autocuts_mode;
    int autocuts_id;
    int mouse_mode;    // MOD_CUTS or SET_FRAME or ...
    int mouse_button;
    int mouse_x0;
    int mouse_y0;
    int cutTop0;
    int cutBot0;
    int mouse_x;
    int mouse_y;
    int image_line0;
    int image_w, image_h;
    short *tv_image;
    int format;
    int frame_x, frame_y, frame_w, frame_h;
    int line_x0, line_y0, line_x1, line_y1;
    int point_x, point_y, point_z;
    bool force_showing;
    bool show_bar;
    bool ccd_mode;
    int ccd_id;
    void stop_cycle();
    int cycle_stage;
    int cycle_n;
    int cycle_cnt;
    int cycle_nadd;
    bool cycle_bias;
    bool cycle_flat;
    double cycle_delay;
    int cycle_add_ok;
    QWidget *wtag;
    QTimer *tvtimer;
    QMenuBar *menu;
    QPopupMenu *file;
    QPopupMenu *save;
    QPopupMenu *oper;
    QPopupMenu *mmod;
    QPopupMenu *opts;
    QPopupMenu *ccd;
    QPopupMenu *math;
    QPopupMenu *help;
    QLabel     *mesg;
    QPainter  lin;
    QPainter  graph;
    QString lut_name;
    QRgb lut[256];
    QString itt_name;
    int itt[256];
    QPixmap image_pm;
    QPixmap bar_pm;
    QSlider *expos;
    QSlider *elev,*elev1,*elev2;
    char dfltfile[100];
    bool load_tv_dflt(char *dfltname);  // load defaults from  file
    bool save_tv_dflt(char *dfltname);  // save defaults to  file
    char savename[200];
    char *savedir;
    char *servdir;
    char *biasname;
    char *flatname;
    QTabDialog *td_opt;
    QLineEdit *ed_c_nadd;
    QLineEdit *ed_c_dt;
    QLineEdit *ed_c_n;
    QCheckBox *cb_bg;
    QCheckBox *cb_fl;
    QLineEdit *ed_bcut;
    QLineEdit *ed_tcut;
    QLineEdit *ed_f_x,*ed_sf_x,*ed_sf2_x;
    QLineEdit *ed_f_y,*ed_sf_y,*ed_sf2_y;
    QLineEdit *ed_f_w,*ed_sf_w,*ed_sf2_w;
    QLineEdit *ed_f_h,*ed_sf_h,*ed_sf2_h;
    QLineEdit *ed_p_x;
    QLineEdit *ed_p_y;
    QLabel    *lb_p_v;
    QLineEdit *ed_exp;
    QLineEdit *ed_elref1,*ed_elref2;
    QWidget   *fits_hdr;
#define FITS_UNKNOWN 0
#define FITS_INT     1
#define FITS_REAL    2
#define FITS_STRING  3
    int       fits_typ[37];
    QLabel    *fits_nam[37];
    QLineEdit *fits_val[37];
    QLabel    *fits_com[37];
};
