/*
 * (c) 1998,99 Gerd Knorr
 *
 *    capture a image, compress as jpeg and upload to the webserver
 *    using ftp
 *
 */
/*
 * Oct2000 VSher (SAO RAS)
 *   webcam  video4linux
 *   JPEG- FITS- ...
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <math.h>
#include <fcntl.h>
#include <sys/time.h>
#include <sys/ioctl.h>

#include "jpeglib.h"

/* ---------------------------------------------------------------------- */
/* configuration                                                          */

/*#define FTP_HOST          "mail"
 *#define FTP_USER          "webcam"
 *#define FTP_PASS          "webcam@sao.ru"
 *#define FTP_DIRECTORY     "WWW"
 */
#define FTP_HOST          "ann"
#define FTP_USER          "vsher"
#define FTP_PASS          "VSher"
#define FTP_DIRECTORY     "tmp"

#define FTP_FILE          "%s.jpeg"
#define FTP_TMPFILE       "mtuploading.jpeg"
#define FTP_INIT_ONCE     "pass"		/* call once at start */
#define FTP_INIT_CONN     "bin"			/* per connection */

/*#define JPEG_FILE         "/tmp/mtwebcam.jpeg"  */
#define JPEG_FILE         "mtwebcam.jpeg"
#define JPEG_QUALITY      75

#define GRAB_ZOOM         4
#define GRAB_TEXT         "BTA %d.%m.%Y %H:%M:%S"        /* strftime */
#define TEXT_HEIGHT       13

/* ---------------------------------------------------------------------- */

static int ftp_debug = 0;

/* ---------------------------------------------------------------------- */
/* jpeg stuff                                                             */

int
write_jpeg(char *filename, char *data, int width, int height)
{
    struct jpeg_compress_struct cinfo;
    struct jpeg_error_mgr jerr;
    FILE *fp;
    int i;
    unsigned char *line;

    if (NULL == (fp = fopen(filename,"w"))) {
	fprintf(stderr,"can't open %s for writing: %s\n",
		filename,strerror(errno));
	return -1;
    }

    cinfo.err = jpeg_std_error(&jerr);
    jpeg_create_compress(&cinfo);
    jpeg_stdio_dest(&cinfo, fp);
    cinfo.image_width = width;
    cinfo.image_height = height;
    cinfo.input_components = 1;
    cinfo.in_color_space = JCS_GRAYSCALE;
    jpeg_set_defaults(&cinfo);
    jpeg_set_quality(&cinfo, JPEG_QUALITY, TRUE);
    jpeg_start_compress(&cinfo, TRUE);

    for (i = 0, line = data; i < height; i++, line += width*1)
	jpeg_write_scanlines(&cinfo, &line, 1);
    
    jpeg_finish_compress(&(cinfo));
    jpeg_destroy_compress(&(cinfo));
    fclose(fp);

    return 0;
}

/* ---------------------------------------------------------------------- */
/* FTP stuff                                                              */

static int ftp_pty, ftp_pid,ftp_connected;

static char pty_name[32];
static char tty_name[32];

void ftp_send(char *command);
int  ftp_recv();

int open_pty()
{
    static char s1[] = "pqrs";
    static char s2[] = "0123456789abcdef";
    
    char *p1,*p2;
    int pty;

    for (p1 = s1; *p1; p1++) {
        for (p2 = s2; *p2; p2++) {
            sprintf(pty_name,"/dev/pty%c%c",*p1,*p2);
            sprintf(tty_name,"/dev/tty%c%c",*p1,*p2);
	    if (-1 == access(tty_name,R_OK|W_OK))
		continue;
            if (-1 != (pty = open(pty_name,O_RDWR)))
                return pty;
        }
    }
    return -1;
}

void
ftp_init()
{
    static char *argv[] = { "ftp", "-n", NULL };
    static char *init_commands[] = { FTP_INIT_ONCE, NULL };
    int i;

    if (-1 == (ftp_pty = open_pty())) {
	fprintf(stderr,"can't grab pty\n");
        exit(1);
    }
    switch (ftp_pid = fork()) {
    case -1:
        perror("fork");
	exit(1);
    case 0:
        /* child */
	close(ftp_pty);
        close(0); close(1); close(2);
	setsid();
        open(tty_name,O_RDWR); dup(0); dup(0);
        execvp(argv[0],argv);
        perror("execvp");
        exit(1);
    default:
        /* parent */
        break;
    }
    ftp_recv();

    /* initialisation */
    for (i = 0; init_commands[i] != NULL; i++) {
	ftp_send(init_commands[i]);
	ftp_recv();
    }
    return;
}

void
ftp_send(char *command)
{
    char line[128];
    int length;

    length = sprintf(line,"%s\n",command);
    if (ftp_debug)
	fprintf(stderr,">> %s",line);
    if (length != write(ftp_pty,line,length)) {
	fprintf(stderr,"ftp: write error\n");
	exit(1);
    }
}

int
ftp_recv()
{
    char line[512],*p,*n;
    int length, done, status, ret=0;
    fd_set set;

    for (done = 0; !done;) {
	FD_ZERO(&set);
	FD_SET(ftp_pty,&set);
	select(ftp_pty+1,&set,NULL,NULL,NULL);
	
	switch (length = read(ftp_pty,line,511)) {
	case -1:
	    perror("ftp: read error");
	    exit(1);
	case 0:
	    fprintf(stderr,"ftp: EOF\n");
	    exit(1);
	}
	line[length] = 0;

	for (p=line; p && *p; p = n) {
	    /* split into lines */
	    if (NULL != (n = strchr(p,'\n')) || NULL != (n = strchr(p,'\r')))
		*(n++) = 0;
	    else
		n = NULL;
	    if (ftp_debug)
		fprintf(stderr,"<< %s\n",p);

	    /* prompt? */
	    if (NULL != strstr(p,"ftp>")) {
		done = 1;
	    }

	    /* line dropped ? */
	    if (NULL != strstr(p,"closed connection")) {
		fprintf(stderr,"ftp: lost connection\n");
		ftp_connected = 0;
	    }
	    if (NULL != strstr(p,"Not connected")) {
		if (ftp_connected)
		    fprintf(stderr,"ftp: lost connection\n");
		ftp_connected = 0;
	    }

	    /* status? */
	    if (1 == sscanf(p,"%d",&status)) {
		ret = status;
	    }
	}
    }
    return ret;
}

void
ftp_connect()
{
    int i,delay = 0;
    static char *init_commands[] = { FTP_INIT_CONN, NULL };

    for (;;) {
	/* Wiederholungsversuche mit wachsendem Intervall, 10 min max. */
	if (delay) {
	    fprintf(stderr,"ftp: connect failed, sleeping %d sec\n",delay);
	    sleep(delay);
	    delay *= 2;
	    if (delay > 600)
		delay = 600;
	} else {
	    delay = 5;
	}

	/* (re-) connect */
	ftp_send("close");
	ftp_recv();
	ftp_send("open " FTP_HOST);
	if (220 != ftp_recv())
	    continue;

	fprintf(stderr,"ftp: connected to " FTP_HOST "\n");
	ftp_connected = 1;

	/* login */
	ftp_send("user " FTP_USER " " FTP_PASS);
	if (230 != ftp_recv()) {
	    if (!ftp_connected)
		continue;
	    fprintf(stderr,"ftp: login incorrect\n");
	    exit(1);
	}
	
	/* set directory */
	ftp_send("cd " FTP_DIRECTORY);
	if (250 != ftp_recv()) {
	    if (!ftp_connected)
		continue;
	    fprintf(stderr,"ftp: login incorrect\n");
	    exit(1);
	}

	/* initialisation */
	for (i = 0; init_commands[i] != NULL; i++) {
	    ftp_send(init_commands[i]);
	    ftp_recv();
	}

	/* ok */
	break;
    }
}

static char *objname;

void
ftp_upload()
{
    char ftpcom[100];
    ftp_send("put " JPEG_FILE " " FTP_TMPFILE);
    ftp_recv();
    sprintf(ftpcom,"rename " FTP_TMPFILE " " FTP_FILE,objname);
    ftp_send(ftpcom);
    ftp_recv();
}


/* ---------------------------------------------------------------------- */
/* capture stuff  - vs56                                                  */


static int                      out_w,out_h,out_size;
static unsigned char            *out_data=NULL;

#define MAXSIZE 1200*1200
boolean get_fits(char *fitsname, int *width, int *height, short *array, int maxsize);
static short *img;                     /* FITS image memory */
static int imgw, imgh;                 /* FITS image size   */
static int col0=0, lin0=0, frmw=0, frmh=0; /* =0 - no frame */


static void stat_tv_image(int init);
static int vmin,vmax,med;              /* minimum, maximum, median     */
static double avr, dsp;                /* average level and dispersion */
#define BOTLEV (-5.)                   /* CUT-levels (in Sigma) */
#define TOPLEV (+20.)
static int vbot,vtop;                  /* real CUT-levels */

void
grab_init()
{
    double zoom=GRAB_ZOOM;
    out_w = imgw/zoom;
    if(out_w<300) {zoom=imgw/300.; out_w=300;}
    out_h = imgh/zoom;
    if(out_h<100) {zoom=imgh/100.; out_h=100; out_w = imgw/zoom;}
    out_size = out_w * (out_h+TEXT_HEIGHT);
    fprintf(stderr,"image out: using %dx(%d+%d) => %d byte\n",
	    out_w,out_h,TEXT_HEIGHT,out_size);
    if(out_data!=NULL)
	free(out_data);
    if (NULL == (out_data = malloc(out_size))) {
	fprintf(stderr,"out of virtual memory\n");
	exit(1);
    }
}

unsigned char*
grab_one(int *width, int *height)
{
    int i,j,x0,y0,ix,ix1,iy,iy1,mx,in;
    double xstep,ystep,x,y;

    x0 = 0;
    xstep = (double)frmw/out_w;
    y0 = 0;
    ystep = (double)frmh/out_h;

    for(i=0,iy=y0,y=y0; i<out_h; i++,y+=ystep) {
       iy1=(int)(y+ystep+0.00000001);
       for(j=0; j<out_w; j++)
	  out_data[i*out_w+j] = 0;
       for(;iy<iy1;iy++) {
	  for(j=0,ix=x0,x=x0; j<out_w; j++,x+=xstep) {
	     ix1=(int)(x+xstep+0.00000001);
	     mx=out_data[i*out_w+j];
	     for(;ix<ix1;ix++) {
		in = (int)((double)(img[iy*imgw+ix]-vbot)*255.0/(vtop-vbot)+0.5);
		if(in>255) in=255;
		if(in<0) in=0;
		if(in>mx) mx=in;
	     }
	     out_data[i*out_w+j]=mx;
	  }
       }
    }
    for(i=out_h; i<out_h+TEXT_HEIGHT; i++)
       for(j=0; j<out_w; j++)
	  out_data[i*out_w+j] = 0;
    *width = out_w;
    *height = out_h;
    return out_data;
}


/* ---------------------------------------------------------------------- */

#if 0
# define CHAR_HEIGHT  8
# define CHAR_WIDTH   8
# define CHAR_START   1
# include "font_8x8.h"
#else
# define CHAR_HEIGHT  11
# define CHAR_WIDTH   6
# define CHAR_START   4
# include "font_6x11.h"
#endif

void
add_text(char *image, int width, int height)
{
    time_t      t;
    struct tm  *tm;
    char        line[128],tmp[20],*ptr;
    int         i,x,y,f,len;

    time(&t);
    tm = localtime(&t);
    len = strftime(line,127,GRAB_TEXT,tm);

    sprintf(tmp,"   FITS:%s",objname);
    strncat(line,tmp,127);
    len = strlen(line);

    fprintf(stderr,"%s\n",line);

    for (y = 0; y < CHAR_HEIGHT; y++) {
	ptr = image + 1 * width * (height-CHAR_HEIGHT+y) + 12;
	for (x = 0; x < len; x++) {
	    f = fontdata[line[x] * CHAR_HEIGHT + y];
	    for (i = CHAR_WIDTH-1; i >= 0; i--) {
		if (f & (CHAR_START << i)) {
		    ptr[0] = 255;
		}
		ptr += 1;
	    }
	}
    }
}

int
main(int argc, char *argv[])
{
    char *fitsname, *p;
    unsigned char *image;
    int i, ret, width, height;

    if (argc<2) {
       fprintf(stderr,"Usage: mtwebcam FITSfile.mt\n");
       exit(1);
    }
    fitsname = argv[1];
    objname = strdup(fitsname);

    if((p=strchr(objname,'.'))!=NULL) *p = '\0';

    fprintf(stderr,"Original video4linux webcam v1.1 - (c) 1998,99 Gerd Knorr\n");
    fprintf(stderr,"FITS-files presentation mtwebcam  -  Oct2000 VSher\n");
    fprintf(stderr,"JPEG quality %d\n", JPEG_QUALITY);
    fprintf(stderr,"ftp config:\n  " FTP_USER "@" FTP_HOST ":" FTP_DIRECTORY
		   "\n  " FTP_TMPFILE " => " FTP_FILE "\n",objname);
    
    img = (short *)calloc(MAXSIZE, 2);
    if(img == NULL) {
	fprintf(stderr, "Can't allocate %d-word FITS-image buffer\n", MAXSIZE);
	exit(1);
    }
    ftp_init();
    ftp_connect();

    ret=0;
    for(i=1;i<argc;i++) {
	fitsname = argv[i];
	objname = strdup(fitsname);
	if((p=strchr(objname,'.'))!=NULL) *p = '\0';

	if(get_fits(fitsname, &imgw, &imgh, img, MAXSIZE)) {
	    grab_init();

	    if (!ftp_connected)
	       ftp_connect();

	    stat_tv_image(1);

	    image = grab_one(&width,&height);
	    add_text(image, width, height+TEXT_HEIGHT);
	    write_jpeg(JPEG_FILE, image, width, height+TEXT_HEIGHT);

	    ftp_upload();

	    frmw=frmh=0; /* =0 - no frame */
	} else
	    ret=1;
    }
    exit(ret);
}

/*----VSher--------------------------------------------------------------------*/

static int hist[6554];              /* histogram array 6554 ~= 2^16 / 10 */
				    /* do some statistics on tv_image frame */
static void stat_tv_image(int init) /* if(init) => full; else partial */
{                             /* calculate vmin, vmax, avr, med, dsp, vbot,vtop  */
   int i,j,k,x,y,pix;
   int xstep,ystep;
   int n,nbot,nmid,ntop;
   int hist0,hbot,htop;

   if(init) {
      n = frmw*frmh;
      nbot = (int)(n*0.05 + 0.5);    /* ~5%  ""  "" */
      ntop = (int)(n*0.95 + 0.5);    /*   */
      nmid = n/2;

      for (j=0; j<6554; j++) hist[j]=0;
      hist0 = -32768;
      vmax = vmin = img[0];
      for (j=0; j<frmh; j++) {   /*    =10 */
	 y=lin0+j;
	 for (i=0; i<frmw; i++) {
	    x=col0+i;
	    pix = img[imgw*y+x];
	    k = (pix-hist0+5)/10;
	    hist[k]++;
	    if(pix<vmin) vmin=pix;
	    if(pix>vmax) vmax=pix;
	 }
      }
      for (i=j=0; j<6553 && i<nmid; j++, i += hist[j]);
      med = j*10+hist0;
      hist0 = med-3277;

      for (j=0; j<6554; j++) hist[j]=0;
      for (j=0; j<frmh; j++) {   /*    =1 */
	 y=lin0+j;
	 for (i=0; i<frmw; i++) {
	    x=col0+i;
	    pix = img[imgw*y+x];
	    k = pix-hist0;
	    k = (k<0)? 0 : ((k>6553)? 6553 : k);
	    hist[k]++;
	 }
      }
   } else {   /*   - -   -  */
      xstep = frmw/150+1;
      ystep = frmh/150+1;
      hist0 = (int)(avr+0.5)-3277;
      for (j=0; j<6554; j++) hist[j]=0;
      vmax = vmin = img[imgw*(lin0+frmh/2)+(col0+frmw/2)];
      n=0;
      for (j=0; j<frmh; j += ystep) {   /*    =1 */
	 y=lin0+j;
	 for (i=0; i<frmw; i += xstep ) {
	    x=col0+i;
	    pix = img[imgw*y+x];
	    k = pix-hist0;
	    k = (k<0)? 0 : ((k>6553)? 6553 : k);
	    hist[k]++;
	    n++;
	    if(pix<vmin) vmin=pix;
	    if(pix>vmax) vmax=pix;
	 }
      }
      nbot = (int)(n*0.05 + 0.5);    /* ~5%  ""  "" */
      ntop = (int)(n*0.95 + 0.5);    /*   */
      nmid = n/2;

   }
   hbot = 0;                    /*  ""    */
   htop = 6553;                 /*    ~10%  */
   for (i=j=0; j<6553; j++) {   /*  ""  avr  dsp */
      i += hist[j];
      if(i<=nbot) hbot = j;
      if(i<=nmid) med  = j;
      if(i<=ntop) htop = j;
   }

   if(med-hbot < htop-med) {     /*   "" - */
      htop = med + (med-hbot);
      if(htop>6553) htop=6553;
   } else {
      hbot = med - (htop-med);
      if(hbot<0) hbot=0;
   }
   med += hist0;
   for(k=0;k<7;k++) {
     /* printf("Range= %d - %d\n",hbot+hist0,htop+hist0);*/
      avr = 0.;
      for (i=0,j=hbot; j<htop; j++) {
	 avr += hist[j]*j;
	 i += hist[j];
      }
      avr /= i;
      for (dsp=0.,j=hbot; j<htop; j++) {
	 double d = j-avr;
	 dsp += d*d*hist[j];
      }
      avr += hist0;
      dsp = sqrt(dsp/i);
      htop = (int)(avr+2.0*(dsp<1.?1:dsp)+0.5) - hist0;
   }
   vbot = (int)(avr+BOTLEV*dsp+0.5);
   vtop = (int)(avr+TOPLEV*dsp+0.5);
   if(vbot<vmin) vbot=vmin;
   if(vtop>vmax) vtop=vmax;
/*printf("Max=%d Min=%d Med=%d Avr=%.2f Dsp=%.2f Cuts=%d,%d\n",
 *       vmax,vmin,med,avr,dsp,vbot,vtop);
 */
}


/* Load frame [col0,lin0,frmw,frmh] from FITS-file */
boolean get_fits(char *fitsname, int *width, int *height, short *array, int maxsize)
{
  int k,l,x,y,lin,n;
  char tmp[2880];
  FILE *fd;
  boolean simple = FALSE;
  int bitpix = 8;
  int naxis = 0;
  int xsize = 0;
  int ysize = 0;
  double bscale = 1.;
  double bzero = 0.;
  int mn=32767, mx=-32768;
  boolean pipe = FALSE;
  short *lbuf = NULL;

  if(strstr(fitsname,".gz")||strstr(fitsname,".bz")) {
     char command[100];
     if(strstr(fitsname,".gz"))
	sprintf(command,"zcat %s",fitsname);
     else
	sprintf(command,"bzcat %s",fitsname);
     if((fd = popen(command,"r")) == NULL) {
	fprintf(stderr, "Can't unZIP FITS-file (%s)\n", command);
	return FALSE;
     }
     pipe = TRUE;
  } else {
     if((fd=fopen(fitsname, "r")) == NULL) {
	fprintf(stderr, "Can't read FITS-file %s\n", fitsname);
	return FALSE;
     }
     pipe = FALSE;
  }
  for (lin=36*5; lin>0; lin--) {
     char *pname = tmp;
     char *pval = tmp+9;
     char *pcom ;
     if (fread(tmp,1,80,fd) <= 0 ) goto wrong_fits;
     pcom = strchr(pval, '/');
     if(pcom != NULL) *pcom++ = '\0';

     if(       strncmp(pname,"END      ",9)==0) break;
     else if(  strncmp(pname,"SIMPLE  =",9)==0) {
	if( strstr(pval," T") == NULL ) goto wrong_fits;
	simple = TRUE;
     } else if(strncmp(pname,"BITPIX  =",9)==0) {
	if( strstr(pval," 16")  == NULL) goto wrong_fits;
	bitpix = 16;
     } else if(strncmp(pname,"NAXIS   =",9)==0) {
	if( strstr(pval," 2") == NULL) goto wrong_fits;
	naxis = 2;
     } else if(strncmp(pname,"NAXIS1  =",9)==0) {
	xsize = atoi(pval);
     } else if(strncmp(pname,"NAXIS2  =",9)==0) {
	ysize = atoi(pval);
     } else if(strncmp(pname,"BSCALE  =",9)==0) {
	bscale = atof(pval);
     } else if(strncmp(pname,"BZERO   =",9)==0) {
	bzero = atof(pval);
     }
  }
/*
 *printf("%dlines simple=%s bitpix=%d naxis=%d\n",36*5-lin,simple?"T":"F",bitpix,naxis);
 *printf("xsize=%d  ysize=%d bscale=%f bzero=%f\n",xsize,ysize,bscale,bzero);
 */
  if(lin==0 || !simple || bitpix!=16 || naxis!=2 || xsize<1 || ysize<1) goto wrong_fits;

  if(frmw==0) frmw=xsize;
  if(frmh==0) frmh=ysize;
  if(col0+frmw>xsize) frmw=xsize-col0;
  if(lin0+frmh>ysize) frmh=ysize-lin0;
  col0=lin0=0;
  if(frmw*frmh > maxsize) {
	fprintf(stderr, "FIST-image too long! (%dx%d points)\n", frmw,frmh);
	return FALSE;
  }

  lbuf = (short *)calloc(xsize, 2);
  if(lbuf == NULL) {
	fprintf(stderr, "Can't allocate %d-byte buffer\n", xsize*2);
	return FALSE;
  }
  if(pipe) {
     lin = lin%36;
     while(--lin>0)
	if (fread(tmp,1,80,fd) <= 0 ) goto wrong_fits;
     for ( y=0; y<lin0; y++ )
	if (fread(lbuf,2,xsize,fd) <= 0 ) goto wrong_fits;
  } else
     fseek(fd, ((ftell(fd)-1)/2880+1)*2880 + lin0*xsize*2, SEEK_SET);

  for ( l=0,y=lin0; y<ysize && l<frmh; y++,l++ ) {
     if(pipe) {
	if (col0>0) {
	   if (fread(lbuf,2,col0,fd) <= 0 ) goto wrong_fits;
	}
     } else
	fseek(fd, col0*2, SEEK_CUR);
     for ( k=0,x=col0; x<xsize && k<frmw; x++,k++ ) {
	union {char b[2]; short i;} pix;
	int i;
	if ((i=fgetc(fd))==EOF)  goto wrong_fits;
	pix.b[1] = i;
	if ((i=fgetc(fd))==EOF)  goto wrong_fits;
	pix.b[0] = i;
	n = l*frmw+k;
	i = (short)(pix.i*bscale + bzero + 0.5);
	if(n<maxsize) array[n] = i;
	if(mn>i) mn = i;
	if(mx<i) mx = i;
     }
     if (x<xsize) {
	if(pipe) {
	   if (fread(lbuf,2,xsize-x,fd) <= 0 ) goto wrong_fits;
	} else
	   fseek(fd, (xsize-x)*2, SEEK_CUR);
     }
  }
  fclose(fd);
  if(lbuf!=NULL) free(lbuf);
  fprintf(stderr, "Load FITS from %s  %d x %d \n",fitsname,xsize,ysize);
/*fprintf(stderr, "frame: %d,%d %d x %d  min=%d max=%d\n",col0,lin0,frmw,frmh,mn,mx);*/
  *width=frmw;
  *height=frmh;
  return TRUE;

wrong_fits:
  fprintf(stderr, "Can't recognize FITS-format (%s)\n", fitsname);
  fclose(fd);
  if(lbuf!=NULL) free(lbuf);
  return FALSE;
}
