/*
 * (c) 1998,99 Gerd Knorr
 *
 *    capture a image, compress as jpeg and upload to the webserver
 *    using ftp
 *
 */
/*
 * Oct2000 VSher (SAO RAS)
 *   webcam  video4linux
 *   TV-  VS56/VS60 ...
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <math.h>
#include <fcntl.h>
#include <sys/time.h>
#include <sys/ioctl.h>

#include "jpeglib.h"

#include "tv_image_lib.h"  /* use VS56/VS60-tools to get images */

/* ---------------------------------------------------------------------- */
/* configuration                                                          */

#define FTP_HOST          "www"
#define FTP_USER          "webcam"
#define FTP_PASS          "xxxxxx"
#define FTP_DIRECTORY     "WWW"
#define FTP_FILE          "tvwebcam_%s_%s_%d.jpeg"
static char ftp_file[100];
#define FTP_FILE_MINI     "tvwebcam_%s_%s_%d_mini.jpeg"
static char ftp_file_mini[100];
#define FTP_TMPFILE       "tvupload_%s_%s_%d.jpeg"
static char ftp_tmpfile[100];
#define FTP_INIT_ONCE     "pass"		/* call once at start */
#define FTP_INIT_CONN     "bin"			/* per connection */

#define JPEG_FILE         "/tmp/tvwebcam_%s_%d.jpeg"
static char jpeg_file[100];
#define JPEG_QUALITY      75

#define GRAB_WIDTH        324
#define GRAB_HEIGHT       243
#define GRAB_TEXT         "webcam %d.%m.%Y %H:%M:%S"        /* strftime */

#define GRAB_DELAY        3

/* ---------------------------------------------------------------------- */

static int ftp_debug = 0;

/* ---------------------------------------------------------------------- */
/* jpeg stuff                                                             */

int
write_jpeg(char *filename, char *data, int width, int height)
{
    struct jpeg_compress_struct cinfo;
    struct jpeg_error_mgr jerr;
    FILE *fp;
    int i;
    unsigned char *line;

    if (NULL == (fp = fopen(filename,"w"))) {
	fprintf(stderr,"can't open %s for writing: %s\n",
		filename,strerror(errno));
	return -1;
    }

    cinfo.err = jpeg_std_error(&jerr);
    jpeg_create_compress(&cinfo);
    jpeg_stdio_dest(&cinfo, fp);
    cinfo.image_width = width;
    cinfo.image_height = height;
    cinfo.input_components = 1;
    cinfo.in_color_space = JCS_GRAYSCALE;
    jpeg_set_defaults(&cinfo);
    jpeg_set_quality(&cinfo, JPEG_QUALITY, TRUE);
    jpeg_start_compress(&cinfo, TRUE);

    for (i = 0, line = data; i < height; i++, line += width*1)
	jpeg_write_scanlines(&cinfo, &line, 1);
    
    jpeg_finish_compress(&(cinfo));
    jpeg_destroy_compress(&(cinfo));
    fclose(fp);

    return 0;
}

/* ---------------------------------------------------------------------- */
/* FTP stuff                                                              */

static int ftp_pty, ftp_pid,ftp_connected;

static char pty_name[32];
static char tty_name[32];

void ftp_send(char *command);
int  ftp_recv();

int open_pty()
{
    static char s1[] = "pqrs";
    static char s2[] = "0123456789abcdef";
    
    char *p1,*p2;
    int pty;

    for (p1 = s1; *p1; p1++) {
        for (p2 = s2; *p2; p2++) {
            sprintf(pty_name,"/dev/pty%c%c",*p1,*p2);
            sprintf(tty_name,"/dev/tty%c%c",*p1,*p2);
	    if (-1 == access(tty_name,R_OK|W_OK))
		continue;
            if (-1 != (pty = open(pty_name,O_RDWR)))
                return pty;
        }
    }
    return -1;
}

void
ftp_init()
{
    static char *argv[] = { "ftp", "-n", NULL };
    static char *init_commands[] = { FTP_INIT_ONCE, NULL };
    int i;

    if (-1 == (ftp_pty = open_pty())) {
	fprintf(stderr,"can't grab pty\n");
        exit(1);
    }
    switch (ftp_pid = fork()) {
    case -1:
        perror("fork");
	exit(1);
    case 0:
        /* child */
	close(ftp_pty);
        close(0); close(1); close(2);
	setsid();
        open(tty_name,O_RDWR); dup(0); dup(0);
        execvp(argv[0],argv);
        perror("execvp");
        exit(1);
    default:
        /* parent */
        break;
    }
    ftp_recv();

    /* initialisation */
    for (i = 0; init_commands[i] != NULL; i++) {
	ftp_send(init_commands[i]);
	ftp_recv();
    }
    return;
}

void
ftp_send(char *command)
{
    char line[128];
    int length;

    length = sprintf(line,"%s\n",command);
    if (ftp_debug)
	fprintf(stderr,">> %s",line);
    if (length != write(ftp_pty,line,length)) {
	fprintf(stderr,"ftp: write error\n");
	exit(1);
    }
}

int
ftp_recv()
{
    char line[512],*p,*n;
    int length, done, status, ret=0;
    fd_set set;

    for (done = 0; !done;) {
	FD_ZERO(&set);
	FD_SET(ftp_pty,&set);
	select(ftp_pty+1,&set,NULL,NULL,NULL);
	
	switch (length = read(ftp_pty,line,511)) {
	case -1:
	    perror("ftp: read error");
	    exit(1);
	case 0:
	    fprintf(stderr,"ftp: EOF\n");
	    exit(1);
	}
	line[length] = 0;

	for (p=line; p && *p; p = n) {
	    /* split into lines */
	    if (NULL != (n = strchr(p,'\n')) || NULL != (n = strchr(p,'\r')))
		*(n++) = 0;
	    else
		n = NULL;
	    if (ftp_debug)
		fprintf(stderr,"<< %s\n",p);

	    /* prompt? */
	    if (NULL != strstr(p,"ftp>")) {
		done = 1;
	    }

	    /* line dropped ? */
	    if (NULL != strstr(p,"closed connection")) {
		fprintf(stderr,"ftp: lost connection\n");
		ftp_connected = 0;
	    }
	    if (NULL != strstr(p,"Not connected")) {
		if (ftp_connected)
		    fprintf(stderr,"ftp: lost connection\n");
		ftp_connected = 0;
	    }

	    /* status? */
	    if (1 == sscanf(p,"%d",&status)) {
		ret = status;
	    }
	}
    }
    return ret;
}

void
ftp_connect()
{
    int i,delay = 0;
    static char *init_commands[] = { FTP_INIT_CONN, NULL };
    int status;

    for (;;) {
	/* Wiederholungsversuche mit wachsendem Intervall, 10 min max. */
	if (delay) {
	    fprintf(stderr,"ftp: connect failed, sleeping %d sec\n",delay);
	    sleep(delay);
	    delay *= 2;
	    if (delay > 600)
		delay = 600;
	} else {
	    delay = 5;
	}

	/* (re-) connect */
	ftp_send("close");
	ftp_recv();
	ftp_send("open " FTP_HOST);
	status=ftp_recv();
	if (status != 220 && status != 530)
	    continue;

	fprintf(stderr,"ftp: connected to " FTP_HOST "\n");
	ftp_connected = 1;

	/* login */
	ftp_send("user " FTP_USER " " FTP_PASS);
	if (230 != ftp_recv()) {
	    if (!ftp_connected)
		continue;
	    fprintf(stderr,"ftp: login incorrect\n");
	    exit(1);
	}
	
	/* set directory */
	ftp_send("cd " FTP_DIRECTORY);
	if (250 != ftp_recv()) {
	    if (!ftp_connected)
		continue;
	    fprintf(stderr,"ftp: login incorrect\n");
	    exit(1);
	}

	/* initialisation */
	for (i = 0; init_commands[i] != NULL; i++) {
	    ftp_send(init_commands[i]);
	    ftp_recv();
	}

	/* ok */
	break;
    }
}

static int nchan = 0;                  /* TV-chan. number */

void
ftp_upload(char *remote_file)
{
    char ftpcom[100];
    sprintf(ftpcom, "put %s %s", jpeg_file, ftp_tmpfile);
    ftp_send(ftpcom);
    ftp_recv();
    sprintf(ftpcom,"rename %s %s", ftp_tmpfile, remote_file);
    ftp_send(ftpcom);
    ftp_recv();
}


/* ---------------------------------------------------------------------- */
/* capture stuff  - vs56                                                  */


static int                      out_w,out_h,out_size;
static unsigned char            *out_data, *out_mini;

static short *img;                     /* VS image memory */
static int imgw, imgh;                 /* VS image size   */
//static int nchan = 0;                /* TV-chan. number */
static int frmc0, frml0, frmw, frmh;   /* frame: area of interest */


static void stat_tv_image(int init);
static int vmin,vmax,med;              /* minimum, maximum, median     */
static double avr, dsp;                /* average level and dispersion */
#define BOTLEV (-5.)                   /* CUT-levels (in Sigma) */
#define TOPLEV (+20.)
static int vbot,vtop;                  /* real CUT-levels */

static int nseq, nseq_o, mseq, mseq_o, nadded, nadded_o;
static int col0_o, lin0_o, width_o, height_o;

void
grab_init()
{
    get_vs56_sh_mem();

    out_w = GRAB_WIDTH;
    out_h = GRAB_HEIGHT;
    out_size = out_w * out_h;
    fprintf(stderr,"grabber out: using %dx%d => %d byte\n",
	    out_w,out_h,out_size);
    if (NULL == (out_data = malloc(out_size)) ||
	NULL == (out_mini = malloc(out_size/4))) {
	fprintf(stderr,"out of virtual memory\n");
	exit(1);
    }
    img = tv_chan_image (nchan);
    imgw = tv_chan_img_width();
    imgh = tv_chan_img_height();
    fprintf(stderr,"grabber inp: using %dx%dx2 => %d byte\n",
	    imgw,imgh,imgw*imgh*2);
    nseq = nseq_o = tv_chan_seq_num();
    mseq = mseq_o = tv_chan_seq_msg();
    col0_o = lin0_o = width_o = height_o = 0;
    nadded = nadded_o = 0;
}

unsigned char*
grab_one(int *width, int *height)
{
    int i,j,x0,y0,ix,ix1,iy,iy1,mx,in,xbnd0,xbnd1,ybnd0,ybnd1;
    double xstep,ystep,x,y;

    x0 = frmc0;  xbnd0 = 0;  xbnd1 = out_w-1;
    if(frmw>out_w) {
       xstep = (double)frmw/out_w;
    } else if(frmw==out_w) {
       xstep = 1.0;
    } else {
       xstep = 1.0;
       x0 = frmc0 - (out_w-frmw)/2;
       if(x0<0) x0=0;
       if(x0+out_w>=imgw) x0=imgw-out_w-1;
       xbnd0 = frmc0-x0;
       xbnd1 = xbnd0+frmw-1;
       frmw=out_w;
    }
    y0 = frml0;  ybnd0 = 0;  ybnd1 = out_h-1;
    if(frmh>out_h) {
       ystep = (double)frmh/out_h;
    } else if(frmh==out_h) {
       ystep = 1.0;
    } else {
       ystep = 1.0;
       y0 = frml0 - (out_h-frmh)/2;
       if(y0<0) y0=0;
       if(y0+out_h>=imgh) y0=imgh-out_h-1;
       ybnd0 = frml0-y0;
       ybnd1 = ybnd0+frmh-1;
       frmh=out_h;
    }
    for(i=0,iy=y0,y=y0; i<out_h; i++,y+=ystep) {
       iy1=(int)(y+ystep+0.00000001);
       for(j=0; j<out_w; j++)
	  out_data[i*out_w+j] = 0;
       for(;iy<iy1;iy++) {
	  for(j=0,ix=x0,x=x0; j<out_w; j++,x+=xstep) {
	     ix1=(int)(x+xstep+0.00000001);
	     mx=out_data[i*out_w+j];
	     for(;ix<ix1;ix++) {
		in = (int)((double)(img[iy*imgw+ix]-vbot)*255.0/(vtop-vbot)+0.5);
		if(in>255) in=255;
		if(in<0) in=0;
		if(in>mx) mx=in;
	     }
	     out_data[i*out_w+j]=mx;
	  }
       }
    }
    for(i=ybnd0; i<ybnd1; i++) {
       j=i*out_w;
       out_data[j+xbnd0] = out_data[j+xbnd1] = 255;
    }
    for(j=xbnd0; j<xbnd1; j++)
       out_data[ybnd0*out_w+j] = out_data[ybnd1*out_w+j] = 255;

    tv_chan_get_point(&ix, &iy);
    if(ix>frmc0 && ix<frmc0+frmw-1 &&
       iy>frml0 && iy<frml0+frmh-1) {

       j=(int)((ix-frmc0)/xstep+0.5)+xbnd0;
       for(i=ybnd0; i<ybnd1; i++)
	   out_data[i*out_w+j] = ((i%6)>2)? 0 : 255;

       i=(int)((iy-frml0)/ystep+0.5)+ybnd0;
       for(j=xbnd0; j<xbnd1; j++)
	   out_data[i*out_w+j] = ((j%6)>2)? 0 : 255;
    }
    *width = out_w;
    *height = out_h;
    return out_data;
}

void
make_mini_image(unsigned char *image, int width, int height, int fact, unsigned char *mini_image)
{
    int i,ii,j,jj,pixi,pixo;
    int h = height/fact;
    int w = width/fact;

    for(j=0;j<h;j++) for(i=0;i<w;i++) {
       pixo=0;
       for(jj=0;jj<fact;jj++) for(ii=0;ii<fact;ii++) {
	  pixi = image[(j*fact+jj)*width + i*fact+ii];
	  if(pixi>pixo) pixo=pixi;
       }
       mini_image[j*w+i] = pixo;
    }
}

/* ---------------------------------------------------------------------- */

#if 0
# define CHAR_HEIGHT  8
# define CHAR_WIDTH   8
# define CHAR_START   1
# include "font_8x8.h"
#else
# define CHAR_HEIGHT  11
# define CHAR_WIDTH   6
# define CHAR_START   4
# include "font_6x11.h"
#endif

void
add_text(char *image, int width, int height, char *alt_text)
{
    time_t      t;
    struct tm  *tm;
    char        line[128],tmp[20],*ptr;
    int         i,x,y,f,len;

    if(alt_text==NULL) {
       time(&t);
       tm = localtime(&t);
       len = strftime(line,110,GRAB_TEXT,tm);
    } else
       strncpy(line,alt_text,110);

    if(vs_type()<800)
      sprintf(tmp," VS%d/Ch%d",vs_type(),nchan);
    else
      sprintf(tmp," BT%d/Ch%d",vs_type(),nchan);
    strncat(line,tmp,127);
    len = strlen(line);
    if((len+1)*CHAR_WIDTH > width) len = width/CHAR_WIDTH-1;

    fprintf(stderr,"%s\n",line);

    for (y = 0; y < CHAR_HEIGHT; y++) {
	ptr = image + 1 * width * (height-CHAR_HEIGHT-2+y) + CHAR_WIDTH;
	for (x = 0; x < len; x++) {
	    f = fontdata[line[x] * CHAR_HEIGHT + y];
	    for (i = CHAR_WIDTH-1; i >= 0; i--) {
		if (f & (CHAR_START << i)) {
		    ptr[0] = 255;
		}
		ptr += 1;
	    }
	}
    }
}

int
main(int argc, char *argv[])
{
    unsigned char *image;
    int width, height;
    char tv_driver[20] = "";
    char hname[80],tvname[20],str[80];
    char *hostname=NULL;
    FILE *fd;
    double t,t_o;

    if(argc>1) {
       char *p = strchr(argv[1],':');
       if(p!=NULL || argv[1][0] == 'v' || argv[1][0] == 'V' || argv[1][0] == 'b') {
	  strncpy(tv_driver,argv[1],19);
	  if(p!=NULL) nchan = atoi(p+1);
       } else
	  nchan = atoi(argv[1]);
    }
    if(strncmp(tv_driver,"vs56",4)==0 || strncmp(tv_driver,"VS56",4)==0) get_vs56_sh_mem();
    if(strncmp(tv_driver,"vs60",4)==0 || strncmp(tv_driver,"VS60",4)==0) get_vs60_sh_mem();
    if(strncmp(tv_driver,"vsbt",4)==0 || strncmp(tv_driver,"bttv",4)==0) get_vsbt_sh_mem();

    if (argc > 2)
	ftp_debug = 1;

    if((fd=popen("hostname","r"))!=NULL) {
       hostname=fgets(hname,39,fd);
       pclose(fd);
    }
    if(hostname==NULL) {
       char *p;
       if((p=getenv("HOSTNAME"))!=NULL)
	  strcpy(hname,p);
       else
	  strcpy(hname,"X");
    }
    hostname = strtok(hname,".\n");
    if(vs_type()<800)
       sprintf(tvname,"vs%d", vs_type());
    else
       strcpy(tvname,"vsbt");
    sprintf(str,"%s:",hostname);

    sprintf( ftp_file,    FTP_FILE,    hostname,tvname,nchan);
    sprintf( ftp_file_mini,FTP_FILE_MINI,hostname,tvname,nchan);
    sprintf( ftp_tmpfile, FTP_TMPFILE, hostname,tvname,nchan);
    sprintf( jpeg_file,   JPEG_FILE,   tvname,nchan);


    fprintf(stderr,"Original video4linux webcam v1.1 - (c) 1998,99 Gerd Knorr\n");
    fprintf(stderr,"VS56/VS60/bttv tvwebcam  -  Oct2000 VSher\n");
    fprintf(stderr,"grabber config: size %dx%d, jpeg quality %d\n",
		   GRAB_WIDTH,GRAB_HEIGHT,JPEG_QUALITY);
    fprintf(stderr,"ftp config:\n  " FTP_USER "@" FTP_HOST ":" FTP_DIRECTORY
		   "\n   %s => %s\n",ftp_tmpfile,ftp_file);

    grab_init();

    ftp_init();
    ftp_connect();

    t = t_o = my_time();
    nseq_o = 0;
    for (;;) {
      if (!ftp_connected)
	    ftp_connect();

	tv_chan_get_frame(&frmc0, &frml0, &frmw, &frmh);
	nseq = tv_chan_seq_num();
	if( nseq!=0 && (frmc0!=col0_o || frml0!=lin0_o ||
	    frmw!=width_o || frmh!=height_o || nseq!=nseq_o) ) {

	    nadded = tv_chan_added();
	    t = my_time();

	    if( frmc0!=col0_o || frml0!=lin0_o ||
		frmw!=width_o || frmh!=height_o ||
		nadded != nadded_o || t-t_o > 30. ) {

		stat_tv_image(1);

		col0_o=frmc0; lin0_o=frml0;
		width_o=frmw; height_o=frmh;
		nadded_o=nadded;
	    } else
		stat_tv_image(0);

	    image = grab_one(&width, &height);
	    make_mini_image(image, width, height, 3, out_mini);
	    add_text(image, width, height,NULL);
	    add_text(out_mini, width/3, height/3, str);

	    write_jpeg(jpeg_file, image, width, height);
	    ftp_upload(ftp_file);

	    write_jpeg(jpeg_file, out_mini, width/3, height/3);
	    ftp_upload(ftp_file_mini);

	    nseq_o=nseq;
	    t_o = t;
	}

#if GRAB_DELAY > 0
	sleep(GRAB_DELAY);
#endif
    }
    return 0;
}

/*----VSher--------------------------------------------------------------------*/

static int hist[6554];              /* histogram array 6554 ~= 2^16 / 10 */
				    /* do some statistics on tv_image frame */
static void stat_tv_image(int init) /* if(init) => full; else partial */
{                             /* calculate vmin, vmax, avr, med, dsp, vbot,vtop  */
   int i,j,k,x,y,pix;
   int xstep,ystep;
   int n,nbot,nmid,ntop;
   int hist0,hbot,htop;

   if(init) {
      n = frmw*frmh;
      nbot = (int)(n*0.05 + 0.5);    /* ~5%  ""  "" */
      ntop = (int)(n*0.95 + 0.5);    /*   */
      nmid = n/2;

      for (j=0; j<6554; j++) hist[j]=0;
      hist0 = -32768;
      vmax = vmin = img[0];
      for (j=0; j<frmh; j++) {   /*    =10 */
	 y=frml0+j;
	 for (i=0; i<frmw; i++) {
	    x=frmc0+i;
	    pix = img[imgw*y+x];
	    k = (pix-hist0+5)/10;
	    hist[k]++;
	    if(pix<vmin) vmin=pix;
	    if(pix>vmax) vmax=pix;
	 }
      }
      for (i=j=0; j<6553 && i<nmid; j++, i += hist[j]);
      med = j*10+hist0;
      hist0 = med-3277;

      for (j=0; j<6554; j++) hist[j]=0;
      for (j=0; j<frmh; j++) {   /*    =1 */
	 y=frml0+j;
	 for (i=0; i<frmw; i++) {
	    x=frmc0+i;
	    pix = img[imgw*y+x];
	    k = pix-hist0;
	    k = (k<0)? 0 : ((k>6553)? 6553 : k);
	    hist[k]++;
	 }
      }
   } else {   /*   - -   -  */
      xstep = frmw/150+1;
      ystep = frmh/150+1;
      hist0 = (int)(avr+0.5)-3277;
      for (j=0; j<6554; j++) hist[j]=0;
      vmax = vmin = img[imgw*(frml0+frmh/2)+(frmc0+frmw/2)];
      n=0;
      for (j=0; j<frmh; j += ystep) {   /*    =1 */
	 y=frml0+j;
	 for (i=0; i<frmw; i += xstep ) {
	    x=frmc0+i;
	    pix = img[imgw*y+x];
	    k = pix-hist0;
	    k = (k<0)? 0 : ((k>6553)? 6553 : k);
	    hist[k]++;
	    n++;
	    if(pix<vmin) vmin=pix;
	    if(pix>vmax) vmax=pix;
	 }
      }
      nbot = (int)(n*0.05 + 0.5);    /* ~5%  ""  "" */
      ntop = (int)(n*0.95 + 0.5);    /*   */
      nmid = n/2;

   }
   hbot = 0;                    /*  ""    */
   htop = 6553;                 /*    ~10%  */
   for (i=j=0; j<6553; j++) {   /*  ""  avr  dsp */
      i += hist[j];
      if(i<=nbot) hbot = j;
      if(i<=nmid) med  = j+1;
      if(i<=ntop) htop = j+1;
   }

   if(med-hbot < htop-med) {     /*   "" - */
      htop = med + (med-hbot);
      if(htop>6553) htop=6553;
   } else {
      hbot = med - (htop-med);
      if(hbot<0) hbot=0;
   }
   med += hist0;
   for(k=0;k<7;k++) {
     /* printf("Range= %d - %d\n",hbot+hist0,htop+hist0);*/
      avr = 0.;
      for (i=0,j=hbot; j<=htop; j++) {
	 avr += (double)hist[j]*j;
	 i += hist[j];
      }
      if(i<1) i=1;
      avr /= i;
      for (dsp=0.,j=hbot; j<=htop; j++) {
	 double d = j-avr;
	 dsp += d*d*hist[j];
      }
      avr += hist0;
      dsp = sqrt(dsp/i);
      htop = (int)(avr+2.0*(dsp<1.?1:dsp)+0.5) - hist0;
      if(htop>6553) htop=6553;
   }
   vbot = (int)(avr+BOTLEV*dsp+0.5);
   vtop = (int)(avr+TOPLEV*dsp+0.5);
   if(vbot<vmin) vbot=vmin;
   if(vtop>vmax) vtop=vmax;
/*printf("Max=%d Min=%d Med=%d Avr=%.2f Dsp=%.2f Cuts=%d,%d\n",
 *       vmax,vmin,med,avr,dsp,vbot,vtop);
 */
}

