
/*       - */

#define MaxNumCh  4    /* max.possible TV-channels for grabber card */
#define VSNumCh   4    /* number of TV-channels at VS56/60 */
#define BTNumCh   3    /* number of BT848/878 TV-channels */
#define MaxWidth  768
#define MaxHeight 576
#define MinFrame  8    /* minimun 8x8 frame size */

/* commands */
#define INULL     0    /* nothing to do */
#define ICLEAR    1    /* clear image before exposure */
#define IADD      2    /* do exposure and add it to image */
#define IEXP      (ICLEAR|IADD)  /* do exposure */
#define ICCD      4    /* read image with online CCD "line-reduction" */
#define IWRITE    8    /* write low bytes of image back to vs56/2 */
#define IINIT     16   /* re-intializate vs56 */
#define IQUICK    32   /* quick ADD: no intermediate showing! but after completion! */

/* one channel "driver<->client" communication area */
struct TVchan {
   int num;         /* OUT channel number */
   int vs_type;     /* OUT type of VS-grabber: 56/60 or BT-grabber: 848/878 */
   int tandem;      /* OUT "Tandem" chan. flag */
   double exposure; /* IN  sec, for "Tandem" only! */
   int activ;       /* OUT last set_chan(this_channel,) */
   int seqn;        /* OUT sequential number of current image */
   double tstart;   /* OUT time() of 1-st ADDed image */
   double tcurr;    /* OUT time() of last ADDed (current) image */
   int pid;         /* IN  last active client PID or 0 */
   int command;     /* IN  actions to do (CLEAR,EXP,ADD and so on) */
   int nexp;        /* IN  number of times to execute command (-1 - infinit)*/
   int nadd;        /* IN  number of images to ADD (-1 - infinit)*/
   int nadded;      /* OUT number of images really  ADDed */
   int addlev;      /* OUT zero level already substructed from ADDed images */
   int next_form;   /* IN  to change format (1-768x576, 2-512x512) */
   int form;        /* OUT current format */
   union {
     int ref_lev;       /* IN VS56 electric reference level (0-63) */
     struct {
	int white_lev;  /* IN VS60 level of "White" (0-255) */
	int black_lev;  /* IN VS60 level of "Black" (0-255) */
     } vs60;
     struct {
	int brightness; /* IN hight-byte of BTTV brightness (0-255) */
	int contrast;   /* IN hight-byte of BTTV contrast (0-255) */
     } bttv;
   } ctrl;
   int img_width;   /* OUT image sizes for current format */
   int img_height;  /* OUT */
   int frm_width;   /* IN  frame size */
   int frm_height;  /* IN  */
   int frm_col0;    /* IN  frame position */
   int frm_lin0;    /* IN  */
   int subfrm_width; /* TRN  soft. subframe size */
   int subfrm_height;/* TRN   (initially = frm_width,frm_height) */
   int subfrm_col0;  /* TRN  soft. subframe position */
   int subfrm_lin0;  /* TRN   (relative to frm_col0,frm_lin0; initially=0,0)  */
   int subfrm2_width; /* TRN  2-nd soft. subframe size */
   int subfrm2_height;/* TRN   (initially = frm_width,frm_height) */
   int subfrm2_col0;  /* TRN  2-nd soft. subframe position */
   int subfrm2_lin0;  /* TRN   (relative to frm_col0,frm_lin0; initially=0,0)  */
   int pnt_lin;     /* TRN  pointer position */
   int pnt_col;     /* TRN    (x/y of manually selected pixel) */
   double center_x; /* TRN  image center */
   double center_y; /* TRN   (calculated x/y position) */
   int seqmsg;      /* OUT number of last message */
   char msg[100];   /* OUT last message buffer */
   int bias;           /* IN CCD bias level */
   int flat[MaxWidth]; /* IN CCD "Line" FlatField <<8 */
   char fits_header[2880]; /* TRN up to 36 FITS-lines for header of FITS-file */
   short image[MaxWidth*MaxHeight];/* OUT image itself */
};

/* TV memory: whole "driver<->client" communication area */
struct TVmem {
   int size;            /* size of TV memory segment */
   char grab_name[32];  /* grabber card name */
   int numch;           /* actual number of chans (1 <= numch <= MaxNumCh) */
   int drv_pid;         /* actual drivers PID or 0 */
   struct TVchan chmem[1]; /*memory for the 1-st channel (should be extended to numch) */
};

extern struct TVmem *tvmem;
extern struct TVchan *tvch[MaxNumCh], *curch;

#define NumCh  (tvmem->numch)
