
/*   VS60     VS56 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <signal.h>
#include <time.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/times.h>
#include <fcntl.h>
#include <sched.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/mman.h>

#include "vs60_io.h"
#include "pci_amcc.h"
#include "vs60.h"

#include "vs56_shm.h"    /* to use all VS56 image tools */
#include "vs56_shm_key.h"
#include "vs60_shm_key.h"


static void vs60_init();
static unsigned short find_pci_port(unsigned short default_port);
static void init_io();
static void get_sh_mem();
static void chan_msg( char *new_msg);
static void wall_msg( char *new_msg);
static void mydelay(double dt);
static double mytime();
static void myabort(int);
static void vs60_exit(int n);
void load_init_file(char *file_name_list[]);
void reset_card();
int  write_reg(unsigned int reg, unsigned int val);
void set_fifo_mode(int mode);
void transfer_block(unsigned char *buff, int buff_size);
unsigned long read_status();
int  set_window(int col0, int lin0, int ncol, int nlin);
int  wait_tv_sync(int sync_type);
static unsigned set_bt252_byte(unsigned int addr, unsigned char byte);
int  set_channel(unsigned int chan);
int  load_itt(unsigned char *itt);
int  set_elref(unsigned char black, unsigned char white);
int  set_bank_mode(int bank_mode);
int  single_cadr(unsigned char *image_buffer, int buffer_size, int force_sync);
int  next_cadr(unsigned char *image_buffer, int buffer_size);

char *myname;
char *init_file_list[] = {
  "vs60bt2.rbf", "spe/vs60bt2.rbf", "../spe/vs60bt2.rbf", NULL};

static unsigned short Port = 0x0;         /* VS60 PCI base i/o address */

static unsigned char *shm_addr;           /* shm-block with buffers   */
static struct TVmem  *tvmem;              /* TV-memory structure      */
static struct TVchan *tvch[MaxNumCh];     /* for TV-channels & images */
static struct TVchan *curch=NULL;         /* current working channel  */

static unsigned char *tv_cadr;            /* one TV-cadr memory    */
static    int cadr_w, cadr_h;             /* cadr sizes   */
static double cadr_time = 0.;             /* time of the last cadr */

static union {
   struct BMCSR reg;
   unsigned long ul;
} BusCS;

int main (int argc, char *argv[])
{
   short *img;                     /* image memory */
   int imgw, imgh;                 /* image sizes   */
   int col0, lin0, width, height;  /* frame: area of interest */
   int nch, i,j, ncadr,ne,na, x,y, fsize, nadd, seqn;
   int ccd_mode, ccd_favr, *ccd_f;
   int quick_mode, quick_add_mode, slow_add_mode;
   unsigned char *p;
   unsigned short *q;
   char tmp[100];
   int first[VSNumCh], lev[VSNumCh], w_lev[VSNumCh], b_lev[VSNumCh];
   int hist[256];
   int old_col0, old_lin0, old_width, old_height;
   double tdelay = 0.1;

   myname = argv[0];

   vs60_init();

   for(nch=0; nch<NumCh; nch++) {
      first[nch] = 1;
      lev[nch] = 0;
      w_lev[nch] = tvch[nch]->ctrl.vs60.white_lev;
      b_lev[nch] = tvch[nch]->ctrl.vs60.black_lev;
   }
   old_col0  = old_lin0 = 0;
   old_width = cadr_w;
   old_height= cadr_h;

   ncadr=ne=0;
   while(1) {
     na=0;
     for(nch=0; nch<NumCh; nch++) {
       if(tvch[nch]->command == INULL) {
	   continue;
       }
       if(tvch[nch]->command == IINIT) {
printf("Chan%d Init\n",nch);
	  tvch[nch]->command = INULL;
	  load_init_file(init_file_list);

	  if(set_elref(b_lev[nch], w_lev[nch]))
	     fprintf(stderr,"BLACK/WHITE levels setting FAULT!\n");

	  if(set_channel(nch))
	     fprintf(stderr,"TV Channel setting FAULT!\n");
	  else
	     curch=tvch[nch];

	  if(load_itt(NULL))
	     fprintf(stderr,"Normal ITT loading FAULT!\n");

	  if(set_window(VS_CAM_BT_CAMERA__XB, VS_CAM_BT_CAMERA__YB, cadr_w, cadr_h))
	     fprintf(stderr,"Windiow setting FAULT!\n");
	  else {
	     old_col0  = old_lin0 = 0;
	     old_width = cadr_w;
	     old_height= cadr_h;
printf("Chan%d Window:%dx%d+0+0\n",nch,cadr_w,cadr_h);
	  }
	  wall_msg("VS60 Init");
	  continue;
       }
       if(tvch[nch]->command&ICLEAR) {
	  first[nch] = 1;
	  lev[nch] = 0;
	  tvch[nch]->nadded = 0;
	  tvch[nch]->addlev = 0;
	  tvch[nch]->tstart = tvch[nch]->tcurr = mytime();
	  if((tvch[nch]->command&(IADD|IQUICK)) == 0) {
	     for(q=tvch[nch]->image,j=0; j<MaxWidth*MaxHeight; j++) *q++ = 0;
	     tvch[nch]->seqn++;
	     tvch[nch]->command &= ~ICLEAR;
	  }
       }
       if(tvch[nch]->command&(IADD|IQUICK)) {
	  int force_sync=0;
	  if(tvch[nch]->nadd==0 || tvch[nch]->nexp==0) {
	     tvch[nch]->command = INULL;
	     continue;
	  }
	  if(curch!=tvch[nch] ||
	     w_lev[nch] != tvch[nch]->ctrl.vs60.white_lev ||
	     b_lev[nch] != tvch[nch]->ctrl.vs60.black_lev) {
	     int wl = tvch[nch]->ctrl.vs60.white_lev;
	     int bl = tvch[nch]->ctrl.vs60.black_lev;
	     if(curch!=tvch[nch])
		fprintf(stderr,"Chan%d --> Chan%d\n",curch->num,nch);
	     if(b_lev[curch->num] != bl)
		fprintf(stderr,"Chan%d set level of 'BLACK' = %d\n",nch,bl);
	     if(w_lev[curch->num] != wl)
		fprintf(stderr,"Chan%d set level of 'WHITE' = %d\n",nch,wl);
	     if(set_elref(bl, wl))
		fprintf(stderr,"BLACK/WHITE levels setting FAULT!\n");
	     else {
		b_lev[nch] = bl;
		w_lev[nch] = wl;
	     }
	     if(set_channel(nch)) {
		fprintf(stderr,"TV-chan%d setting FAULT!\n",nch);
		tvch[nch]->command = INULL;
		continue;
	     } else
		curch=tvch[nch];
	     force_sync=1;
	  }
/*printf("Chan%d Add: Nexp=%d Nadd=%d\n",curch->num,curch->nexp,curch->nadd);*/
	  quick_mode = (curch->command&IQUICK)? 1 : 0;
	  quick_add_mode = ((curch->command&IQUICK) && curch->nadd>1)? 1 : 0;
	  slow_add_mode = (((curch->command&IQUICK)==0 && curch->nadd>1)||(curch->nadd<0))? 1 : 0;
	  ccd_mode = (curch->command&ICCD)? 1 : 0;

	  ccd_f = curch->flat;
	  nadd = curch->nadd;
	  seqn = curch->seqn;
	  img  = curch->image;
	  imgw = curch->img_width;
	  imgh = curch->img_height;
	  col0 = curch->frm_col0;
	  lin0 = curch->frm_lin0;
	  width= curch->frm_width;
	  height=curch->frm_height;
	  lin0 = (lin0/2)*2;
	  width = (width<1)? 4 : ((width+3)/4)*4;
	  if((col0+width)>VS_CAM_BT_CAMERA__XS)
	     width = ((VS_CAM_BT_CAMERA__XS-col0)/4)*4;
	  if((lin0+height)>VS_CAM_BT_CAMERA__YS)
	     height = ((VS_CAM_BT_CAMERA__YS-lin0)/2)*2;
	  fsize = width*height;
	  if(old_col0 != col0 || old_lin0 != lin0 ||
	     old_width != width || old_height != height) {
	     if(set_window(VS_CAM_BT_CAMERA__XB+col0, VS_CAM_BT_CAMERA__YB+lin0/2, width, height))
		fprintf(stderr,"Windiow setting FAULT!\n");
	     else {
		old_col0 = col0;
		old_lin0 = lin0;
		old_width = width;
		old_height = height;
printf("Chan%d Window:%dx%d+%d+%d\n",nch,width, height, col0,lin0);
	     }
	     force_sync=1;
	  }
	  if(ccd_mode) {
	     ccd_favr = 1;
	     for(i=0; i<width; i++) {
		x=col0+i;
		if(ccd_f[x]<1) ccd_f[x]=1;
		ccd_favr += ccd_f[x];
	     }
	     ccd_favr = (ccd_favr + width/2) / width;
	  }
	  if(single_cadr(tv_cadr, fsize, force_sync)) {
	     fprintf(stderr,"Getting single cadr FAULT!\n");
	     curch->command &= ~(IEXP|IQUICK);
	     curch->nadd = curch->nexp = 0;
	     continue;
	  } else {
	     int lev0 = (first[nch] || quick_mode)? 0 : lev[nch];

	     tvch[nch]->tcurr = cadr_time;

	     if(first[nch] && !quick_mode)
		for(i=0; i<256; i++) hist[i]=0;

	     p=tv_cadr;
	     for (j=0; j<height; j++) {
		y=lin0+j;
		q = &img[imgw*y+col0];
		if(ccd_mode && !quick_add_mode) {
		   int *cfp = ccd_f+col0;
		   if(first[nch] && !quick_mode)
		      for (i=0; i<width; i++) {
			 hist[*p]++;
			 *q++ = *p++ - ((*cfp++ - ccd_favr)>>8);
		      }
		   else if(first[nch])
		      for (i=0; i<width; i++)
			 *q++ = *p++ - ((*cfp++ - ccd_favr)>>8);
		   else
		      for (i=0; i<width; i++)
			 *q++ += *p++ - ((*cfp++ - ccd_favr)>>8) - lev0;

		} else if(first[nch] && !quick_mode)
		   for (i=0; i<width; i++) {
		      hist[*p]++;
		      *q++ = *p++;
		   }
		else if(first[nch])
		   for (i=0; i<width; i++)
		      *q++ = *p++;
		else
		   for (i=0; i<width; i++)
		      *q++ += *p++ - lev0;
	     }
	     ncadr++;
	     curch->nadded++;
	     if(!first[nch])
		curch->addlev += lev0;
	     if(first[nch] && !quick_mode) {
		int i05=0;
		int i95=255;
		int n05 = (int)(fsize*0.05 + 0.5);
		int n95 = (int)(fsize*0.95 + 0.5);
		double havr;
		for(i=j=0; i<256; i++) {
		   int j0 = j;
		   j += hist[i];
		   if(j0<n05 && n05<=j) i05=i;
		   if(j0<n95 && n95<=j) i95=i;
		}
		for(havr=0.,j=0,i=i05; i<=i95; i++) {
		   havr += hist[i]*i;
		   j += hist[i];
		}
		if(j>1) havr /= j;
		lev[nch] = (int)(havr+0.5);
	     }
	     if(!quick_add_mode) {
		curch->seqn++;
		sprintf(tmp,"Frame%d (%s%d SubLev=%d)",
		  curch->seqn,(first[nch])?"Get":"Add",curch->nadded,curch->addlev);
		chan_msg(tmp);
	     } else
		seqn++;
	     if(curch->nadd>0)
		nadd--;
	     first[nch] = 0;
	  }
	  na=1;
	  while(curch->nadd!=0 && nadd!=0) {
	     if(next_cadr(tv_cadr, fsize)) {
		fprintf(stderr,"Getting next cadr FAULT!\n");
		curch->command &= ~(IEXP|IQUICK);
		curch->nadd = curch->nexp = 0;
		break;
	     } else {
		int lev0 = lev[nch];
		tvch[nch]->tcurr = cadr_time;

		p=tv_cadr;
		for (j=0; j<height; j++) {
		   y=lin0+j;
		   q = &img[imgw*y+col0];
		   if(ccd_mode && !quick_add_mode) {
		      int *cfp = ccd_f+col0;
		      for (i=0; i<width; i++) {
			 *q++ += *p++ - ((*cfp++ - ccd_favr)>>8) - lev0;
		      }
		   } else if(slow_add_mode)
		      for (i=0; i<width; i++) {
			 *q++ += *p++ - lev0;
		      }
		   else
		      for (i=0; i<width; i++) {
			 *q++ += *p++;
		      }
		}
		ncadr++;
		curch->nadded++;
		if(slow_add_mode)
		   curch->addlev += lev0;
		if(!quick_add_mode) {
		   curch->seqn++;
		   sprintf(tmp,"Frame%d (Add%d SubLev=%d)",
		      curch->seqn,curch->nadded,curch->addlev);
		   chan_msg(tmp);
		   if(nadd!=1) {
		      if(quick_mode) {
			 if(na%5==0)
			    mydelay(tdelay);
		      } else
			 mydelay(tdelay);
		   }
		} else {                  /* "quick" mode */
		   seqn++;
		   if(na%25==0)
		      mydelay(tdelay);
		}
		if(curch->nadd>0)
		   nadd--;
		na++;
	     }
	  }
	  if(ccd_mode && quick_add_mode) {
	     for (j=0; j<height; j++) {
		int *cfp = ccd_f+col0;
		y=lin0+j;
		q = &img[imgw*y+col0];
		for (i=0; i<width; i++,q++,cfp++) {
		   *q -= ((*cfp-ccd_favr)*na+128)>>8;
		}
	     }
	  }
	  if(quick_add_mode) {
	     curch->seqn = seqn;
	     sprintf(tmp,"Frame%d (Add%d SubLev=%d)",
		  curch->seqn,curch->nadded,curch->addlev);
	     chan_msg(tmp);
	  }
	  if(curch->nexp>0)
	     curch->nexp--;
	  if(curch->nexp==0) {
/*printf("Chan%d: Nadd=%d dTime=%.3f\n",curch->num, curch->nadd, curch->tcurr-curch->tstart);*/
	     curch->command &= ~(IEXP|IQUICK);
	     curch->nadd = 0;
	     if(quick_add_mode)
		curch->seqn = seqn;
	     if(curch->pid != 0) {  /* client needs ACK? */
		if(kill(curch->pid, SIGUSR1)<0)
		   perror("Sending ACK-signal to client");
		curch->pid = 0;
	     }
	  }
	  if(quick_add_mode) {
	     curch->seqn = seqn;
	     sprintf(tmp,"Frame%d (Add%d SubLev=%d)",
		  curch->seqn,curch->nadded,curch->addlev);
	     chan_msg(tmp);
	  }
       }
     }
     if(na) ne++;
     if(quick_mode) {
	if(ne%5==0)
	   mydelay(tdelay);
     } else
	mydelay(tdelay);

     quick_mode = 0;
   }
   vs60_exit(0);
}

static void vs60_init() {
    static struct sched_param p;
    int pid=0;

    get_sh_mem();

  /* allocate cadr buffer */
   cadr_w = VS_CAM_BT_CAMERA__XS;
   cadr_h = VS_CAM_BT_CAMERA__YS;
   tv_cadr = (unsigned char *)malloc(cadr_w*cadr_h);
   if(tv_cadr==NULL) {
      fprintf(stderr,"Can't allocate memory for cadr buffer!\n");
      exit(1);
   }
   Port = find_pci_port(0);
   init_io();

   signal(SIGHUP, myabort);
   signal(SIGINT, myabort);
   signal(SIGQUIT,myabort);
   signal(SIGILL, myabort);
   signal(SIGABRT,myabort);
   signal(SIGFPE, myabort);
   signal(SIGSEGV,myabort);
   signal(SIGPIPE,myabort);
   signal(SIGTERM,myabort);
   signal(SIGUSR1,myabort);
   signal(SIGUSR2,myabort);
/***
   p.sched_priority = 10;
   if (sched_setscheduler(pid, SCHED_FIFO, &p)) {
      perror("Can't set scheduler priority");
      vs60_exit(errno);
   }
***/fprintf(stderr,"Don't set REALTIME priority!\n");

   if (mlockall(MCL_CURRENT)) {
      perror("Can't lock process memory");
      vs60_exit(errno);
   }

   load_init_file(init_file_list);

   if(set_elref(0, 255)) {
      fprintf(stderr,"El.Ref. setting FAULT!\n");
      vs60_exit(1);
   }

   if(set_channel(0)) {
      fprintf(stderr,"TV Channel setting FAULT!\n");
      vs60_exit(1);
   }

   if(load_itt(NULL)) {
      fprintf(stderr,"Normal ITT loading FAULT!\n");
      vs60_exit(1);
   }
   if(set_window(VS_CAM_BT_CAMERA__XB, VS_CAM_BT_CAMERA__YB, cadr_w, cadr_h)) {
      fprintf(stderr,"Windiow setting FAULT!\n");
      vs60_exit(1);
   }
}

#define VS60_VENDOR_ID 0xaa10
#define VS60_DEVICE_ID 0xbb01

static unsigned short find_pci_port(unsigned short default_port)
{
   char tmp[100], *bp, *p;
   static int vs60ok=0;
   unsigned int vend_id, dev_id;
   FILE *fd;
   if((fd=fopen("/proc/pci","r"))==NULL) goto cant_find;
   vs60ok=0;
   while(fgets(tmp,99,fd)!=NULL) {
      for(bp=tmp; *bp==' '; bp++);
      if(strstr(bp,"Vendor id=") && strstr(bp,"Device id=")) {
	 p=strchr(bp,'=');
	 vend_id=strtoul(p+1,NULL,16);
	 p=strrchr(bp,'=');
	 dev_id=strtoul(p+1,NULL,16);
	 if(vend_id==VS60_VENDOR_ID && dev_id==VS60_DEVICE_ID)
	    vs60ok=1;
      } else if (strncmp(bp,"Bus ",4)==0) {
	 if(vs60ok) {
	    vs60ok=0;
	    break;
	 }
      } else if(vs60ok && strncmp(bp,"I/O at ",7)==0)
	 return((unsigned short)strtoul(bp+7,NULL,16));
   }
cant_find:
   fprintf(stderr, "Can't find VS60 PCI device in /proc/pci!\n");
   if(default_port!=0) {
      fprintf(stderr, "Use default Port 0x%x instead ...\n",default_port);
      return(default_port);
   }
   exit(1);
}

/* Allocate shared memory segment */
static void get_sh_mem() {
   int i,j, shmid, new_shm=0;
   short *p;
   int shm_size = sizeof(struct TVmem)+sizeof(struct TVchan)*(VSNumCh-1);
   char card_name[32] = "VS60 (IPC Videoscan)";
   int my_pid;

   if(curch != NULL) return;

   /* first try to find existing one */
   shmid = shmget(vs60_shm_key.code, shm_size+sizeof(int), 0777);
   if (shmid<0 && errno==ENOENT) {
      /* if no - try to create a new one */
      shmid = shmget(vs60_shm_key.code, shm_size+sizeof(int), IPC_CREAT|IPC_EXCL|0777);
      new_shm=1;
   }
   if (shmid<0) {
      char msg[80];
      sprintf(msg,"Can't create shared memory segment '%s'",vs60_shm_key.name);
      perror(msg);
      exit(errno);
   }

   /* attach it to our memory space */
   shm_addr = shmat ( shmid, NULL, 0 );
   if ((int)(shm_addr) == -1) {
      char msg[80];
      sprintf(msg,"Can't attach shared memory segment '%s'",vs60_shm_key.name);
      perror(msg);
      exit(errno);
   }
   if((shmctl(shmid, SHM_LOCK, NULL) < 0) < 0) {
      char msg[80];
      sprintf(msg,"Can't prevents swapping of shared memory segment '%s'",vs60_shm_key.name);
      perror(msg);
      exit(errno);
   }

   tvmem = (struct TVmem *)shm_addr;

   if(new_shm) {
      tvmem->size = shm_size;
      tvmem->drv_pid = 0;
      fprintf(stderr,"Create shared memory segment '%s' %dbytes at %x \n",
	   vs60_shm_key.name, shm_size, shm_addr);
   } else {
      fprintf(stderr,"Attach shared memory segment '%s' %dbytes at %x \n",
	   vs60_shm_key.name, tvmem->size, shm_addr);
      if(tvmem->size != shm_size) {
	 fprintf(stderr,"Wrong TV-memory size:%dbytes needs!\n",shm_size);
	 exit(1);
      }
      if(tvmem->numch != VSNumCh) {
	 fprintf(stderr,"Wrong TV-channels number:%d (%d needs)!\n",tvmem->numch,VSNumCh);
	 new_shm=1;
      }
      if(strncmp(tvmem->grab_name,card_name,32) != 0) {
	 fprintf(stderr,"TV-card change?! %s -> %s?!\n",tvmem->grab_name,card_name);
	 new_shm=1;
      }
   }

   my_pid = getpid();

   if( tvmem->drv_pid > 0 && kill(tvmem->drv_pid, 0) >= 0) {
      fprintf(stderr,"vs60 daemon already running! (PID=%d)\n", tvmem->drv_pid);
      sleep(2);
      exit(1);
   }
   strncpy(tvmem->grab_name,card_name,32);
   tvmem->numch = VSNumCh;
   tvmem->drv_pid = my_pid;

   for(i=0,curch=tvmem->chmem; i<NumCh; i++,curch++) {
      tvch[i] = curch;
      if(new_shm || curch->vs_type!=60) {
	 curch->num=i;
	 curch->vs_type=60;
	 curch->tandem=0;
	 curch->activ=0;
	 curch->exposure = 0.04;
	 if(new_shm)
	    curch->seqn=0;
	 curch->tstart = curch->tcurr = time(0);
	 curch->pid=0;
	 curch->command=INULL;
	 curch->nexp=0;
	 curch->nadd=0;
	 curch->nadded=0;
	 curch->addlev=0;
	 curch->next_form=1;
	 curch->form=1;
	 curch->ctrl.vs60.white_lev=255;
	 curch->ctrl.vs60.black_lev=0;
	 curch->img_width  = curch->frm_width  = VS_CAM_BT_CAMERA__XS;
	 curch->img_height = curch->frm_height = VS_CAM_BT_CAMERA__YS;
	 curch->subfrm_width  = curch->subfrm2_width  = VS_CAM_BT_CAMERA__XS;
	 curch->subfrm_height = curch->subfrm2_height = VS_CAM_BT_CAMERA__YS;
	 curch->frm_col0 = curch->frm_lin0 = 0;
	 curch->subfrm_col0 = curch->subfrm_lin0 = 0;
	 curch->subfrm2_col0 = curch->subfrm2_lin0 = 0;
	 curch->pnt_col = curch->pnt_lin = 0;
	 curch->center_x = curch->center_y = 0.0;
	 if(new_shm) {
	    curch->seqmsg=0;
	    curch->msg[0]='\0';
	 }
	 curch->bias = 0;
	 for(j=0; j<MaxWidth; j++)
	     curch->flat[j] = 1;
	 for(j=0; j<2880; j++)
	     curch->fits_header[j] = 0;
	 for(p=curch->image,j=0; j<MaxWidth*MaxHeight; j++)
	     *p++ = (j % curch->img_width) - 100;
      }
   }
   curch = tvmem->chmem;
}

static void chan_msg( char *new_msg) {
    strncpy(curch->msg, new_msg, 99);
    curch->seqmsg++;
}
static void wall_msg( char *new_msg) {
    int nch;
    for(nch=0; nch<NumCh; nch++) {
	strncpy(tvch[nch]->msg, new_msg, 99);
	tvch[nch]->seqmsg++;
    }
}

static int euid,ruid;

static void init_io()
{
   int i;

   ruid = getuid();      /* save real user ID */
   euid = geteuid();     /* save effective user ID (must be "root") */

   if(euid != 0) {
      fprintf(stderr, "Wrong user ID (real or effective)! \n");
      fprintf(stderr, "Should be real 'root' or get superuser rights through programm file owner.\n");
      exit(1);
   }
   setgid(getgid());
   setreuid(euid,euid);  /* set "full root" */

   /* get I/O permissions for I/O registers */
/* if (ioperm(Port, 4, 1)) { */
   i=3;
   if (iopl(i)) {
     if(errno==EPERM)
       fprintf(stderr, "Can't set I/O privilege level (are you a SuperUser?)\n");
     else if(errno==EINVAL)
       fprintf(stderr, "Wrong I/O privilege level(%d)!\n",i);
     else {
       char tmp[40];
       sprintf(tmp,"Error in iopl(%d)! ",i);
       perror(tmp);
     }
     exit (1);
   }

   setreuid(ruid,euid);  /* restore effective and real user IDs */
}

static int vs60_hup=0;

static void mydelay(double dt)
{
  int nfd;
  struct timeval tv;
   tv.tv_sec = (int)dt;
   tv.tv_usec = (int)((dt - tv.tv_sec)*1000000.);
sleeping:
   nfd = select(0, (fd_set *)NULL,(fd_set *)NULL,(fd_set *)NULL, &tv);
   if(nfd < 0) {
      if(errno == EINTR) { /* interrupted?    */
	 if(vs60_hup) {    /* by USR1 signal?  */
	    vs60_hup=0;    /*  ok!              */
	    return;        /*    stop sleeping...*/
	 }
	/*On  Linux,  timeout  is  modified to reflect the amount of
	 time not slept; most other implementations DO NOT do this!*/
	 goto sleeping;
      }
      fprintf(stderr,"Error in mydelay(){ select() }. %s\n",strerror(errno));
   }
}

static double mytime()
{
  struct timeval tv;
  struct timezone tz;
  int ret;
  double t;
  ret = gettimeofday(&tv, &tz);
  t = (double)tv.tv_sec + (double)tv.tv_usec*0.000001;
  t -= timezone;
  return(t);
}

static void myabort(int sig) {
    int ret;
    char ss[10], tmp[80];
    signal(sig,SIG_IGN);
    switch (sig) {
	case SIGHUP : strcpy(ss,"SIGHUP");  break;
	case SIGINT : strcpy(ss,"SIGINT");  break;
	case SIGQUIT: strcpy(ss,"SIGQUIT"); break;
	case SIGILL : strcpy(ss,"SIGILL");  break;
	case SIGABRT: strcpy(ss,"SIGABRT"); break;
	case SIGFPE : strcpy(ss,"SIGFPE");  break;
	case SIGPIPE: strcpy(ss,"SIGPIPE"); break;
	case SIGSEGV: strcpy(ss,"SIGSEGV"); break;
	case SIGTERM: strcpy(ss,"SIGTERM"); break;
	default:  sprintf(ss,"SIG_%d",sig); break;
    }
    switch (sig) {
	case SIGUSR1 :
	     vs60_hup=1;
	     signal(sig, myabort);
	     return;
	default:
	case SIGINT :
	case SIGUSR2 :
	     fprintf(stderr,"%s: %s - Ignore .....\n",myname,ss);
	     fflush(stderr);
	     signal(sig, myabort);
	     return;
	case SIGHUP :
	case SIGQUIT:
	case SIGILL :
	case SIGABRT:
	case SIGFPE :
	case SIGSEGV:
	case SIGPIPE:
	case SIGTERM:
	     signal(SIGALRM, SIG_IGN);
	     fprintf(stderr,"%s: %s - programm stop!\n",myname,ss);
	     vs60_exit(sig);
    }
}

static void vs60_exit(int n) {
   tvmem->drv_pid = 0;
   reset_card();
   exit(n);
}

// ""  
void load_init_file(char *file_name_list[])
{
   unsigned int x, i, c;
   char *fname;
   FILE *fd;

   reset_card();
   fprintf(stderr, "Load init.file: \n");
   for(i=0,fname=file_name_list[0]; fname!=NULL; i++,fname=file_name_list[i]) {
      fprintf(stderr, "\t --- trying  %s ..... ",fname);
      fflush(stderr);
      if((fd=fopen(fname,"r"))==NULL)
	 fprintf(stderr, "Fail!\n");
      else
	 break;
   }
   if(fd==NULL) {
      fprintf(stderr, "Can't find VS60 init.file! Exiting...\n");
      vs60_exit(1);
   }
   while ( (c=fgetc(fd))!=EOF ) {
      //  
      for(i=0;i<8;i++)
      {
	x=(c>>i)&1;
	outb(Port+AMCC_OP_REG_OMB1,x);
	x=x|2;
	outb(Port+AMCC_OP_REG_OMB1,x);
      }
   }
   fclose(fd);
   fprintf(stderr,"Ok!\n");
   fflush(stderr);
}

//C 
void reset_card()
{
   int i;

   BusCS.ul=0;

   BusCS.reg.add_reset=1;

   for(i=0;i<=40;i++)
      outl(Port+AMCC_OP_REG_MCSR, BusCS.ul);

   BusCS.reg.add_reset=0;
   outl(Port+AMCC_OP_REG_MCSR, BusCS.ul);
}


int write_reg(unsigned int reg, unsigned int val)
{
   int timeout;
   //static unsigned long nnn=0;

   BusCS.ul=0;

   timeout=100;
   do{
      BusCS.ul=inl(Port+AMCC_OP_REG_MCSR);
      //printf(".");
      timeout--;
   }while((timeout>0) && (BusCS.reg.pci_add_fifo_Empty!=1));

   if(timeout==0) return(1);

   set_fifo_mode(VS60_MODE_RESET_OUT_FIFO);
   //printf("<%lx (%lu)",Port+AMCC_OP_REG_FIFO,nnn);

   outl(Port+AMCC_OP_REG_FIFO,(unsigned long)((val<<4)|(reg&0xF)));
   //nnn++;
   //printf(">");

   timeout=100;
   do{
      BusCS.ul=inl(Port+AMCC_OP_REG_MCSR);
      //printf(",");
      timeout--;
   }while((timeout>0) && (BusCS.reg.pci_add_fifo_Empty==0));

   if(timeout==0) return(1);

   return(0);
}


void set_fifo_mode(int mode)
{
   switch(mode){
      case VS60_MODE_RESET_IN_FIFO:
	 BusCS.ul=inl(Port+AMCC_OP_REG_MCSR);
	 BusCS.reg.add_pci_fifo_reset=1;
	 outl(Port+AMCC_OP_REG_MCSR, BusCS.ul);
	 break;
      case VS60_MODE_RESET_OUT_FIFO:
	 BusCS.ul=inl(Port+AMCC_OP_REG_MCSR);
	 BusCS.reg.pci_add_fifo_reset=1;
	 outl(Port+AMCC_OP_REG_MCSR, BusCS.ul);
      default : break;
   }
}

//  
void transfer_block(unsigned char *buff, int buff_size)
{

  int nl;

  nl = buff_size / sizeof(long);

  for(; nl>=8; nl -= 8) {
     do {
	BusCS.ul = inl(Port+AMCC_OP_REG_MCSR);
     } while(BusCS.reg.add_pci_fifo_full==0);

     insl_rep(Port+AMCC_OP_REG_FIFO, buff, 8);
     buff += 8 * sizeof(long);
  }
  for(; nl>0; nl--) {
     do {
	BusCS.ul = inl(Port+AMCC_OP_REG_MCSR);
     } while(BusCS.reg.add_pci_fifo_empty==1);

     *((unsigned long *)buff) = inl(Port+AMCC_OP_REG_FIFO);
     buff += sizeof(long);
  }
}

unsigned long read_status()
{
  return(inl(Port+AMCC_OP_REG_IMB4));
}


int set_window(int col0, int lin0, int ncol, int nlin)
{
   int ret;

   if(write_reg(VS60BT_INTREG_WIN_SLINE, (lin0<=0)? 1 : lin0))
      return(1);
   if(write_reg(VS60BT_INTREG_WIN_NLINE, (nlin<1)? 0 : nlin-1))
      return(1);
   if(write_reg(VS60BT_INTREG_WIN_SPIX , (col0<=0)? 1 : col0))
      return(1);
   if(write_reg(VS60BT_INTREG_WIN_NPIX , (ncol<1)? 0 : ncol-1))
      return(1);
   return(0);
}

// 
int wait_tv_sync(int sync_type)
{
   unsigned long t;
   unsigned long npk;

   switch(sync_type) {
      case VS60BT_WAIT_PK0:
	 t = VS60BT_TIMEOUT/10000;
	 while((t>0) && ((read_status()&0x01000000)!=0)) {mydelay(0.011);t--;}
	 if(t==0) {
no_pk0:     fprintf(stderr,"CadrSyncPulse timeout... No video?!\n");
	    wall_msg("Can't read frame, no video!?!");
	    return(1);
	 }
	 t=VS60BT_TIMEOUT;
	 while((t>0) && ((read_status()&0x01000000)==0)) t--;
	 if(t==0)
	    goto no_pk0;
	 break;
      case VS60BT_WAIT_KI:
	 npk = read_status() & 0x01000000;
	 t=VS60BT_TIMEOUT;
	 while((t>0) && ((read_status()&0x01000000)!=npk)) t--;
	 if(t==0) {
	    fprintf(stderr,"HalfCadrPulse timeout... No video?!\n");
	    wall_msg("Can't read frame, no video!?!");
	    return(1);
	 }
	 break;
      default:
	 return(1);
   }
   return(0);
}


//    BT 252
static unsigned set_bt252_byte(unsigned int addr, unsigned char byte)
{
  struct tag_adc_reg {
     unsigned d8_1:8;
     unsigned wadc:1;
     unsigned adr1_0:2;
     unsigned unused:11;
  } ToBt;
  unsigned long *adc,i;
/*int delay = 1000;*/
  int delay = 200;

  adc=(unsigned long *)&ToBt;

  ToBt.d8_1 = byte;
  ToBt.wadc = 0;
  ToBt.adr1_0 = addr;
  for(i=0; i<=delay; i++) {
     if(write_reg(VS60BT_INTREG_BT, *adc))
	return(1);
  }
  ToBt.wadc = 1;
  for(i=0; i<=delay; i++) {
     if(write_reg(VS60BT_INTREG_BT, *adc))
	return(1);
  }
  return(0);
}


//  
int set_channel(unsigned int chan)
{
   unsigned int i;

   chan &= 3;
   i = chan | (chan<<2);
   return(set_bt252_byte(2, VS60BT_LEVEL_SYNC | (i<<4)));
}


//  ITT
int load_itt(unsigned char *itt)  /* load_itt(NULL) - "" ITT */
{
  int i;
  unsigned char ii;

  if(set_bt252_byte(0,0)) return(1);

  for(i=0;i<256;i++){
     ii = (itt==NULL)? i : itt[i];
     if(set_bt252_byte(1, ii)) return(1);  /* ??? */
  }
  return(0);
}

//  
int set_elref(unsigned char black, unsigned char white)
{
  black &= 0x3f;
  white &= 0x3f;
  if(set_bt252_byte(0,1)) return(1);
  if(set_bt252_byte(2,white<<2)) return(1);
  if(set_bt252_byte(0,2)) return(1);
  if(set_bt252_byte(2,black<<2)) return(1);
  if(set_bt252_byte(0,0)) return(1);
  return(0);
}

static int cur_bank=0;

int set_bank_mode(int bank_mode)
{
   switch(bank_mode){
      case VS60BT_SETMODE_BANK0:
	 if(wait_tv_sync(VS60BT_WAIT_PK0)) return(1);

bank0:   if(write_reg(5, 0)) return(1);

	 set_fifo_mode(VS60_MODE_RESET_IN_FIFO);

	 if(write_reg(5, 2)) return(1);
	 cur_bank = 0;
	 break;

      case VS60BT_SETMODE_BANK1:
	 if(wait_tv_sync(VS60BT_WAIT_PK0)) return(1);

bank1:   if(write_reg(5, 0)) return(1);

	 set_fifo_mode(VS60_MODE_RESET_IN_FIFO);

	 if(write_reg(5, 3)) return(1);
	 cur_bank = 1;
	 break;

      case VS60BT_SETMODE_NEXT_BANK:
	 if(wait_tv_sync(VS60BT_WAIT_PK0)) return(1);

	 if(cur_bank==0) goto bank1;
	 else            goto bank0;

      case VS60BT_SETMODE_NEXT_BANK_PK:
	 if(wait_tv_sync(VS60BT_WAIT_KI)) return(1);

	 if(cur_bank==0) goto bank1;
	 else            goto bank0;
      default:return(1);
   }
   return(0);
}

int  single_cadr(unsigned char *image_buffer, int buffer_size, int force_sync)
{
   if(mytime() - cadr_time < 0.12 && force_sync==0)
      return( next_cadr(image_buffer, buffer_size) );

   if(set_bank_mode(VS60BT_SETMODE_BANK1)) return(1);
   mydelay(0.02);
   if(set_bank_mode(VS60BT_SETMODE_BANK0)) return(1);
   cadr_time = mytime();
   transfer_block(image_buffer,  buffer_size);
   return(0);
}

int next_cadr(unsigned char *image_buffer, int buffer_size)
{
   if(mytime() - cadr_time > 0.5)
      return( single_cadr(image_buffer, buffer_size, 1) );

   if(set_bank_mode(VS60BT_SETMODE_NEXT_BANK)) return(1);
   cadr_time = mytime();
   transfer_block(image_buffer,  buffer_size);
   return(0);
}

