#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <signal.h>
#include <time.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/perm.h>
#include <sys/times.h>

#include "vs60_io.h"
#include "pci_amcc.h"
#include "vs60.h"

#include "tv_image_lib.h"  /* use VS56-tools to show images */


static unsigned short find_pci_port(unsigned short default_port);
static void init_io();
static void mydelay(double dt);
static double mytime();
static void myabort(int);
void load_init_file(char *file_name);
void reset_card();
int  write_reg(unsigned int reg, unsigned int val);
void set_fifo_mode(int mode);
void transfer_block(unsigned char *buff, int buff_size);
unsigned long read_status();
int  set_window(int col0, int lin0, int ncol, int nlin);
int  wait_tv_sync(int sync_type);
static unsigned set_bt252_byte(unsigned int addr, unsigned char byte);
int  set_channel(unsigned int chan);
int  load_itt(unsigned char *itt);
int  set_elref(unsigned char black, unsigned char white);
int  set_bank_mode(int bank_mode);
int  single_cadr(unsigned char *image_buffer, int buffer_size);
int  next_cadr(unsigned char *image_buffer, int buffer_size);

char *myname;
char *init_file = "vs60bt2.rbf";

static unsigned short Port = 0x6800;

static union {
   struct BMCSR reg;
   unsigned long ul;
} BusCS;

int main (int argc, char *argv[])
{
   short *img;                     /* image memory */
   int imgw, imgh;                 /* image sizes   */
   int nchan = 0;                  /* TV-chan. number */
   int col0, lin0, width, height;  /* frame: area of interest */
   unsigned char *tv_cadr;         /* one TV-cadr memory   */
   int cadr_w, cadr_h;             /* cadr sizes   */
   int fsize;                      /* frame size = width*height */
   int count=0;
   int i,j,x,y;
   unsigned char *p;
   unsigned char itt[256];
   double t0,t1,q;
   char tmp[80];

   myname = argv[0];
   if(argc>1) nchan=atoi(argv[1]);
   if(argc>2) count=atoi(argv[2]);

  /* use VS56-tools to show images */
   get_vs56_sh_mem();
   img = tv_chan_image (1);
   imgw = tv_chan_img_width();
   imgh = tv_chan_img_height();
   tv_chan_get_frame(&col0, &lin0, &width, &height);
   width = ((width+2)/4)*4;
   fsize = width*height;

  /* allocate cadr buffer */
   tv_cadr = (unsigned char *)malloc(1024*1024);
   cadr_w = VS_CAM_BT_CAMERA__XS;
   cadr_h = VS_CAM_BT_CAMERA__YS;

   Port = find_pci_port(0);
   init_io();

   signal(SIGHUP, myabort);
   signal(SIGINT, myabort);
   signal(SIGQUIT,myabort);
   signal(SIGFPE, myabort);
   signal(SIGPIPE,myabort);
   signal(SIGSEGV,myabort);
   signal(SIGTERM,myabort);

   t0 = mytime();

   load_init_file(init_file);
   t1 = mytime(); printf("load_init_file():%.6f\n",t1-t0); t0=t1;
 /*
  * if(set_channel(nchan)) {
  *    fprintf(stderr,"TV Channel settting FAULT!\n");
  *    goto vs60err;
  * }
  * t1 = mytime(); printf("set_channel(%d):%.6f\n",nchan,t1-t0); t0=t1;
  */
/*while(1){printf("%1c",(read_status()&0x01000000)? '1':'0');fflush(stdout);}*/
 /*
  * while(wait_tv_sync(VS60BT_WAIT_PK0)) {
  *    fprintf(stderr,"Waiting for CADR_SYNC_PULSE... No video?!\n");
  *    mydelay(0.2);
  * }
  */
   t0 = mytime();
   if(set_elref(0, 255)) {
      fprintf(stderr,"El.Ref. settting FAULT!\n");
      goto vs60err;
   }
   t1 = mytime(); printf("set_elref():%.6f\n",t1-t0); t0=t1;

   if(set_channel(nchan)) {
      fprintf(stderr,"TV Channel settting FAULT!\n");
      goto vs60err;
   }
   t1 = mytime(); printf("set_channel(%d):%.6f\n",nchan,t1-t0); t0=t1;

   for (q=0.,i=0; i<256; i++) {
   /* itt[i] = (i<128)? i/2 : (i-128)*3/2+64; */
      itt[i] = (i<128)? i*3/2 : (i-128)/2+192;
   /* itt[i] = (int)(q += (i<64)? 0.5 : (i<192)? 1.5 : 0.5); */
   }
   if(load_itt(NULL)) {
      fprintf(stderr,"Normal ITT loading FAULT!\n");
      goto vs60err;
   }
   t1 = mytime(); printf("load_itt():%.6f\n",t1-t0); t0=t1;
   if(set_window(VS_CAM_BT_CAMERA__XB, VS_CAM_BT_CAMERA__YB, cadr_w, cadr_h)) {
      fprintf(stderr,"Windiow settting FAULT!\n");
      goto vs60err;
   }
   if(single_cadr(tv_cadr, cadr_w*cadr_h)) {
      fprintf(stderr,"Getting single cadr FAULT!\n");
      goto vs60err;
   }
   t1 = mytime(); printf("single_cadr():%.6f\n",t1-t0); t0=t1;
   p=tv_cadr;
   for (j=0; j<imgh; j++) {
      for (i=0; i<imgw; i++) {
	 img[imgw*j+i] = tv_cadr[cadr_w*j+i];
      }
   }
   t1 = mytime(); printf("show_image:%.6f\n",t1-t0); t0=t1;
   tv_chan_set_seq_num(tv_chan_seq_num()+1);

   if(set_window(VS_CAM_BT_CAMERA__XB+col0, VS_CAM_BT_CAMERA__YB+lin0/2, width, height)) {
      fprintf(stderr,"Windiow settting FAULT!\n");
      goto vs60err;
   }

   while(count<0 || fgets(tmp,79,stdin)!=NULL) {
      int cnt,cnt0=(count>0)?count:1;
      t0 = mytime();
      for(cnt=0; cnt<cnt0; cnt++) {
	 int old_col0=col0, old_lin0=lin0, old_width=width, old_height=height;
	 tv_chan_get_frame(&col0, &lin0, &width, &height);
	 lin0 = (lin0/2)*2;
	 if((col0+width)>VS_CAM_BT_CAMERA__XS)
	    width = VS_CAM_BT_CAMERA__XS-col0;
	 if((lin0+height)>VS_CAM_BT_CAMERA__YS)
	    height = VS_CAM_BT_CAMERA__YS-lin0;
	 width = (width<2)? 4 : ((width+2)/4)*4;
	 fsize = width*height;
	 if(old_col0 != col0 || old_lin0 != lin0 ||
	    old_width != width || old_height != height) {
	    if(set_window(VS_CAM_BT_CAMERA__XB+col0, VS_CAM_BT_CAMERA__YB+lin0/2, width, height)) {
	       fprintf(stderr,"Windiow settting FAULT!\n");
	       goto vs60err;
	    }
	    if(single_cadr(tv_cadr, fsize)) {
	       fprintf(stderr,"Getting single cadr FAULT!\n");
	       goto vs60err;
	    }
	 } else {
	    if(next_cadr(tv_cadr, fsize)) {
	       fprintf(stderr,"Getting single cadr FAULT!\n");
	       goto vs60err;
	    }
	 }
      }
      t1 = mytime(); printf("%d Cadr%dx%d:%.6f\n",cnt0,width,height,t1-t0); t0=t1;
      p=tv_cadr;
      for (j=0; j<height; j++) {
	 y=lin0+j;
	 for (i=0; i<width; i++) {
	    x=col0+i;
	    img[imgw*y+x] = tv_cadr[width*j+i];
	 }
      }
      t1 = mytime(); printf("show_image:%.6f\n",t1-t0); t0=t1;
      tv_chan_set_seq_num(tv_chan_seq_num()+1);
   }
   reset_card();
   exit(0);

vs60err:
   reset_card();
   exit(1);
}

#define VS60_VENDOR_ID 0xaa10
#define VS60_DEVICE_ID 0xbb01

static unsigned short find_pci_port(unsigned short default_port)
{
   char tmp[100], *bp, *p;
   static int vs60ok=0;
   unsigned int vend_id, dev_id;
   FILE *fd;
   if((fd=fopen("/proc/pci","r"))==NULL) goto cant_find;
   vs60ok=0;
   while(fgets(tmp,99,fd)!=NULL) {
      for(bp=tmp; *bp==' '; bp++);
      if(strstr(bp,"Vendor id=") && strstr(bp,"Device id=")) {
	 p=strchr(bp,'=');
	 vend_id=strtoul(p+1,NULL,16);
	 p=strrchr(bp,'=');
	 dev_id=strtoul(p+1,NULL,16);
	 if(vend_id==VS60_VENDOR_ID && dev_id==VS60_DEVICE_ID)
	    vs60ok=1;
      } else if (strncmp(bp,"Bus ",4)==0) {
	 if(vs60ok) {
	    vs60ok=0;
	    break;
	 }
      } else if(vs60ok && strncmp(bp,"I/O at ",7)==0)
	 return((unsigned short)strtoul(bp+7,NULL,16));
   }
cant_find:
   fprintf(stderr, "Can't find VS60 PCI device in /proc/pci!\n");
   if(default_port!=0) {
      fprintf(stderr, "Use default Port 0x%x instead ...\n",default_port);
      return(default_port);
   }
   exit(1);
}

static int euid,ruid;

static void init_io()
{
   int i;

   ruid = getuid();      /* save real user ID */
   euid = geteuid();     /* save effective user ID (must be "root") */

   if(euid != 0) {
      fprintf(stderr, "Wrong user ID (real or effective)! \n");
      fprintf(stderr, "Should be real 'root' or get superuser rights through programm file owner.\n");
      exit(1);
   }
   setgid(getgid());
   setreuid(euid,euid);  /* set "full root" */

   /* get I/O permissions for I/O registers */
/* if (ioperm(Port, 4, 1)) { */
   i=3;
   if (iopl(i)) {
     if(errno==EPERM)
       fprintf(stderr, "Can't set I/O privilege level (are you a SuperUser?)\n");
     else if(errno==EINVAL)
       fprintf(stderr, "Wrong I/O privilege level(%d)!\n",i);
     else {
       char tmp[40];
       sprintf(tmp,"Error in iopl(%d)! ",i);
       perror(tmp);
     }
     exit (1);
   }

   setreuid(ruid,euid);  /* restore effective and real user IDs */
}

static void mydelay(double dt)
{
  int nfd;
  struct timeval tv;
   tv.tv_sec = (int)dt;
   tv.tv_usec = (int)((dt - tv.tv_sec)*1000000.);
slipping:
   nfd = select(0, (fd_set *)NULL,(fd_set *)NULL,(fd_set *)NULL, &tv);
   if(nfd < 0) {
      if(errno == EINTR)
	/*On  Linux,  timeout  is  modified to reflect the amount of
	 time not slept; most other implementations DO NOT do this!*/
	 goto slipping;
      fprintf(stderr,"Error in mydelay(){ select() }. %s\n",strerror(errno));
   }
}

static double mytime()
{
  struct timeval tv;
  struct timezone tz;
  int ret;
  double t;
  ret = gettimeofday(&tv, &tz);
  t = (double)tv.tv_sec + (double)tv.tv_usec*0.000001;
  t -= timezone;
  return(t);
}

static void myabort(int sig) {
    int ret;
    char ss[10], tmp[80];
    signal(sig,SIG_IGN);
    switch (sig) {
	case SIGHUP : strcpy(ss,"SIGHUP");  break;
	case SIGINT : strcpy(ss,"SIGINT");  break;
	case SIGQUIT: strcpy(ss,"SIGQUIT"); break;
	case SIGFPE : strcpy(ss,"SIGFPE");  break;
	case SIGPIPE: strcpy(ss,"SIGPIPE"); break;
	case SIGSEGV: strcpy(ss,"SIGSEGV"); break;
	case SIGTERM: strcpy(ss,"SIGTERM"); break;
	default:  sprintf(ss,"SIG_%d",sig); break;
    }
    switch (sig) {
	default:
	case SIGHUP :
	case SIGINT :
	case SIGPIPE:
	     fprintf(stderr,"%s: %s - Ignore .....\n",myname,ss);
	     fflush(stderr);
	     signal(sig, myabort);
	     return;
	case SIGQUIT:
	case SIGFPE :
	case SIGSEGV:
	case SIGTERM:
	     signal(SIGALRM, SIG_IGN);
	     fprintf(stderr,"%s: %s - programm stop!\n",myname,ss);
	     reset_card();
	     exit(sig);
    }
}


// ""  
void load_init_file(char *file_name)
{
   unsigned int x, i, c;
   FILE *fd;

   reset_card();

   if((fd=fopen(file_name,"r"))==NULL) {
      fprintf(stderr, "Can't find VS60 init.file: %s!\n",file_name);
      exit(1);
   }
   while ( (c=fgetc(fd))!=EOF ) {
      //  
      for(i=0;i<8;i++)
      {
	x=(c>>i)&1;
	outb(Port+AMCC_OP_REG_OMB1,x);
	x=x|2;
	outb(Port+AMCC_OP_REG_OMB1,x);
      }
   }
   fclose(fd);
}

//C 
void reset_card()
{
   int i;

   BusCS.ul=0;

   BusCS.reg.add_reset=1;

   for(i=0;i<=40;i++)
      outl(Port+AMCC_OP_REG_MCSR, BusCS.ul);

   BusCS.reg.add_reset=0;
   outl(Port+AMCC_OP_REG_MCSR, BusCS.ul);
}


int write_reg(unsigned int reg, unsigned int val)
{
   int timeout;
   //static unsigned long nnn=0;

   BusCS.ul=0;

   timeout=100;
   do{
      BusCS.ul=inl(Port+AMCC_OP_REG_MCSR);
      //printf(".");
      timeout--;
   }while((timeout>0) && (BusCS.reg.pci_add_fifo_Empty!=1));

   if(timeout==0) return(1);

   set_fifo_mode(VS60_MODE_RESET_OUT_FIFO);
   //printf("<%lx (%lu)",Port+AMCC_OP_REG_FIFO,nnn);

   outl(Port+AMCC_OP_REG_FIFO,(unsigned long)((val<<4)|(reg&0xF)));
   //nnn++;
   //printf(">");

   timeout=100;
   do{
      BusCS.ul=inl(Port+AMCC_OP_REG_MCSR);
      //printf(",");
      timeout--;
   }while((timeout>0) && (BusCS.reg.pci_add_fifo_Empty==0));

   if(timeout==0) return(1);

   return(0);
}


void set_fifo_mode(int mode)
{
   switch(mode){
      case VS60_MODE_RESET_IN_FIFO:
	 BusCS.ul=inl(Port+AMCC_OP_REG_MCSR);
	 BusCS.reg.add_pci_fifo_reset=1;
	 outl(Port+AMCC_OP_REG_MCSR, BusCS.ul);
	 break;
      case VS60_MODE_RESET_OUT_FIFO:
	 BusCS.ul=inl(Port+AMCC_OP_REG_MCSR);
	 BusCS.reg.pci_add_fifo_reset=1;
	 outl(Port+AMCC_OP_REG_MCSR, BusCS.ul);
      default : break;
   }
}

//  
void transfer_block(unsigned char *buff, int buff_size)
{

  int nl;

  nl = buff_size / sizeof(long);

  for(; nl>=8; nl -= 8) {
     do {
	BusCS.ul = inl(Port+AMCC_OP_REG_MCSR);
     } while(BusCS.reg.add_pci_fifo_full==0);

     insl_rep(Port+AMCC_OP_REG_FIFO, buff, 8);
     buff += 8 * sizeof(long);
  }
  for(; nl>0; nl--) {
     do {
	BusCS.ul = inl(Port+AMCC_OP_REG_MCSR);
     } while(BusCS.reg.add_pci_fifo_empty==1);

     *((unsigned long *)buff) = inl(Port+AMCC_OP_REG_FIFO);
     buff += sizeof(long);
  }
}

unsigned long read_status()
{
  return(inl(Port+AMCC_OP_REG_IMB4));
}


int set_window(int col0, int lin0, int ncol, int nlin)
{
   int ret;

   if(write_reg(VS60BT_INTREG_WIN_SLINE, (lin0<=0)? 1 : lin0))
      return(1);
   if(write_reg(VS60BT_INTREG_WIN_NLINE, (nlin<1)? 0 : nlin-1))
      return(1);
   if(write_reg(VS60BT_INTREG_WIN_SPIX , (col0<=0)? 1 : col0))
      return(1);
   if(write_reg(VS60BT_INTREG_WIN_NPIX , (ncol<1)? 0 : ncol-1))
      return(1);
   return(0);
}

// 
int wait_tv_sync(int sync_type)
{
   unsigned long t;
   unsigned long npk;

   switch(sync_type) {
      case VS60BT_WAIT_PK0:
	 t=VS60BT_TIMEOUT;
	 while((t>0) && ((read_status()&0x01000000)!=0)) t--;
	 if(t==0) {
no_pk0:     fprintf(stderr,"CadrSyncPulse timeout... No video?!\n");
	    return(1);
	 }
	 t=VS60BT_TIMEOUT;
	 while((t>0) && ((read_status()&0x01000000)==0)) t--;
	 if(t==0)
	    goto no_pk0;
	 break;
      case VS60BT_WAIT_KI:
	 npk = read_status() & 0x01000000;
	 t=VS60BT_TIMEOUT;
	 while((t>0) && ((read_status()&0x01000000)!=npk)) t--;
	 if(t==0) {
	    fprintf(stderr,"HalfCadrPulse timeout... No video?!\n");
	    return(1);
	 }
	 break;
      default:
	 return(1);
   }
   return(0);
}


//    BT 252
static unsigned set_bt252_byte(unsigned int addr, unsigned char byte)
{
  struct tag_adc_reg {
     unsigned d8_1:8;
     unsigned wadc:1;
     unsigned adr1_0:2;
     unsigned unused:11;
  } ToBt;
  unsigned long *adc,i;
/*int delay = 1000;*/
  int delay = 200;

  adc=(unsigned long *)&ToBt;

  ToBt.d8_1 = byte;
  ToBt.wadc = 0;
  ToBt.adr1_0 = addr;
  for(i=0; i<=delay; i++) {
     if(write_reg(VS60BT_INTREG_BT, *adc))
	return(1);
  }
  ToBt.wadc = 1;
  for(i=0; i<=delay; i++) {
     if(write_reg(VS60BT_INTREG_BT, *adc))
	return(1);
  }
  return(0);
}


//  
int set_channel(unsigned int chan)
{
   unsigned int i;

   chan &= 3;
   i = chan | (chan<<2);
   return(set_bt252_byte(2, VS60BT_LEVEL_SYNC | (i<<4)));
}


//  ITT
int load_itt(unsigned char *itt)  /* load_itt(NULL) - "" ITT */
{
  int i;
  unsigned char ii;

  if(set_bt252_byte(0,0)) return(1);

  for(i=0;i<256;i++){
     ii = (itt==NULL)? i : itt[i];
     if(set_bt252_byte(1, ii)) return(1);  /* ??? */
  }
  return(0);
}

//  
int set_elref(unsigned char black, unsigned char white)
{
  black &= 0x3f;
  white &= 0x3f;
  if(set_bt252_byte(0,1)) return(1);
  if(set_bt252_byte(2,white<<2)) return(1);
  if(set_bt252_byte(0,2)) return(1);
  if(set_bt252_byte(2,black<<2)) return(1);
  if(set_bt252_byte(0,0)) return(1);
  return(0);
}

static int cur_bank=0;

int set_bank_mode(int bank_mode)
{
   switch(bank_mode){
      case VS60BT_SETMODE_BANK0:
	 if(wait_tv_sync(VS60BT_WAIT_PK0)) return(1);
//        mydelay(0.08);

bank0:   if(write_reg(5, 0)) return(1);

	 set_fifo_mode(VS60_MODE_RESET_IN_FIFO);

	 if(write_reg(5, 2)) return(1);
	 cur_bank = 0;
	 break;

      case VS60BT_SETMODE_BANK1:
	 if(wait_tv_sync(VS60BT_WAIT_PK0)) return(1);
//       mydelay(0.08);

bank1:   if(write_reg(5, 0)) return(1);

	 set_fifo_mode(VS60_MODE_RESET_IN_FIFO);

	 if(write_reg(5, 3)) return(1);
	 cur_bank = 1;
	 break;

      case VS60BT_SETMODE_NEXT_BANK:
	 if(wait_tv_sync(VS60BT_WAIT_PK0)) return(1);

	 if(cur_bank==0) goto bank1;
	 else            goto bank0;

      case VS60BT_SETMODE_NEXT_BANK_PK:
	 if(wait_tv_sync(VS60BT_WAIT_KI)) return(1);

	 if(cur_bank==0) goto bank1;
	 else            goto bank0;
      default:return(1);
   }
   return(0);
}

int single_cadr(unsigned char *image_buffer, int buffer_size)
{
   if(set_bank_mode(VS60BT_SETMODE_BANK1)) return(1);
   if(set_bank_mode(VS60BT_SETMODE_BANK0)) return(1);
   transfer_block(image_buffer,  buffer_size);
   return(0);
}

int next_cadr(unsigned char *image_buffer, int buffer_size)
{
   if(set_bank_mode(VS60BT_SETMODE_NEXT_BANK)) return(1);
   transfer_block(image_buffer,  buffer_size);
   return(0);
}

