#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/time.h>
#include <sched.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/mman.h>
#include <sys/ioctl.h>

#include <asm/types.h>
#include "linux/videodev.h"   /* but "videodev2.h" for v4l2 */

#include "vs56_shm.h"         /* to use all VS56 image tools */
#include "bttv_shm_key.h"


#define GRAB_DEVICE       "/dev/video0"
#define GRAB_WIDTH_1      MaxWidth
#define GRAB_HEIGHT_1     MaxHeight
#define GRAB_WIDTH_2      (MaxWidth/3)*2
#define GRAB_HEIGHT_2     (MaxHeight/3)*2

static void get_sh_mem();
static void chan_msg( char *new_msg);
static void wall_msg( char *new_msg);
static void mydelay(double dt);
static double mytime();
static void myabort(int);
static void vid_exit(int n);

char *myname;
static unsigned char *shm_addr;           /* shm-block with buffers   */
static struct TVmem  *tvmem;              /* TV-memory structure      */
static struct TVchan *tvch[MaxNumCh];     /* for TV-channels & images */
static struct TVchan *curch=NULL;         /* current working channel  */

static void swap_rgb24(char *mem, int n)
{
    char  c;
    char *p = mem;
    int   i = n;
    
    while (--i) {
	c = p[0]; p[0] = p[2]; p[2] = c;
	p += 3;
    }
}


/* ---------------------------------------------------------------------- */
/* capture stuff  -  old v4l (bttv)                                       */
/* these work for v4l only, not v4l2 */

#ifdef VIDIOCGCAP

# define GRAB_SOURCE      1
# define GRAB_NORM        VIDEO_MODE_PAL

static struct video_capability   grab_cap;
static struct video_mbuf         grab_buffers;
static struct video_mmap         grab_buff[2];
static struct video_channel	 grab_chan;
static struct video_picture      grab_pict, grab_pict_save;
static int  brightness=32768, contrast=32768;
static int  grab_fd=-1, grab_size, buf_cnt, grab_cnt;
static unsigned char *grab_data, *grab_cadr;

static void grab_open()
{
    static struct sched_param p;
    int pid=0;

    p.sched_priority = 1;
    if (sched_setscheduler(pid, SCHED_FIFO, &p)) {
       perror("Can't set scheduler priority");
    }
    if (mlockall(MCL_CURRENT)) {
       perror("Can't lock process memory");
    }
    /* open video device*/
    if ((grab_fd = open(GRAB_DEVICE,O_RDWR)) < 0 ) {
	perror("open " GRAB_DEVICE);
	exit(1);
    }
    /* obtain video device capability information */
    if (ioctl(grab_fd,VIDIOCGCAP,&grab_cap) < 0 ) {
	fprintf(stderr,"wrong device\n");
	exit(1);
    }
    /* get image source and TV norm */
    grab_chan.channel = GRAB_SOURCE;
    if (ioctl(grab_fd,VIDIOCGCHAN,&grab_chan) < 0 ) {
	perror("ioctl VIDIOCGCHAN");
	exit(1);
    }
    /* get current image properties */
    if (ioctl(grab_fd,VIDIOCGPICT,&grab_pict) < 0 ) {
	perror("ioctl VIDIOCGPICT");
	exit(1);
    }
    grab_pict_save = grab_pict;
}

static void grab_init(int colour)
{
    int i;

    if(grab_fd<0) exit(1); /*against idiots protection :-)*/

    /* set image source and TV norm */
    grab_chan.channel = GRAB_SOURCE;    
    grab_chan.norm    = GRAB_NORM;
    if (ioctl(grab_fd,VIDIOCSCHAN,&grab_chan) < 0 ) {
	perror("ioctl VIDIOCSCHAN");
	exit(1);
    }

    /* set new image properties */
    grab_pict.brightness = brightness;
    grab_pict.contrast = contrast;
    grab_pict.colour = colour;
    if (ioctl(grab_fd,VIDIOCSPICT,&grab_pict) < 0 ) {
	perror("ioctl VIDIOCSPICT");
	exit(1);
    }

    /* map grab buffer */
    if (ioctl(grab_fd,VIDIOCGMBUF,&grab_buffers) < 0) {
	perror("ioctl VIDIOCGMBUF");
	vid_exit(1);
    }
    grab_data = mmap(0,grab_buffers.size,PROT_READ|PROT_WRITE,MAP_SHARED,grab_fd,0);
    if ((int)grab_data == -1) {
	perror("mmap");
	vid_exit(1);
    }
    buf_cnt = (grab_buffers.frames <= 2)? grab_buffers.frames : 2;
    grab_cnt = 0;
    for(i=0; i<buf_cnt; i++) {
	if(colour)
	   grab_buff[i].format = VIDEO_PALETTE_RGB24;
	else
	   grab_buff[i].format = VIDEO_PALETTE_GREY;
	grab_buff[i].frame  = i;
	grab_buff[i].width  = (GRAB_WIDTH_1<grab_cap.maxwidth)? GRAB_WIDTH_1 : grab_cap.maxwidth;
	grab_buff[i].height = (GRAB_HEIGHT_1<grab_cap.maxheight)? GRAB_HEIGHT_1 : grab_cap.maxheight;
    }
    grab_size = grab_buff[0].width * grab_buff[0].height;
    if(grab_buff[0].format == VIDEO_PALETTE_RGB24) grab_size *= 3;
}

static unsigned char* grab_next(int first, int last, int *width, int *height)
{
    int old_frame = grab_cnt % buf_cnt;
    int new_frame = (grab_cnt+1) % buf_cnt;

    if(grab_fd<0) return(NULL); /*against idiots protection :-)*/

    if( first ) {
	if (ioctl(grab_fd,VIDIOCMCAPTURE,&grab_buff[old_frame]) < 0) {
	    perror("ioctl VIDIOCMCAPTURE");
	    return(NULL);
	}
    }
    if( !last ) {
	if (ioctl(grab_fd,VIDIOCMCAPTURE,&grab_buff[new_frame]) < 0) {
	    perror("ioctl VIDIOCMCAPTURE");
	    return(NULL);
	} else
	    grab_cnt++;
    }
    if (ioctl(grab_fd,VIDIOCSYNC,&grab_buff[old_frame]) < 0) {
	perror("ioctl VIDIOCSYNC");
	return(NULL);
    } else {
	if( last ) grab_cnt++;
	if(grab_buff[old_frame].format == VIDEO_PALETTE_RGB24)
	   swap_rgb24(grab_data,grab_buff[old_frame].width*grab_buff[old_frame].height);
	*width  = grab_buff[old_frame].width;
	*height = grab_buff[old_frame].height;
	return(grab_data + grab_buffers.offsets[old_frame]);
    }
}

#endif

int main (int argc, char *argv[])
{
   short *img;                     /* image memory */
   int imgw, imgh;                 /* image sizes   */
   int col0, lin0, width, height;  /* frame: area of interest */
   int nch, i,j,  x,y, fsize,  nadd, seqn;
   int ii, grabw, grabh, colour=0, toadd=1;
   unsigned char *p;
   short *q;
   char tmp[100];

   myname = argv[0];
   if(argc>1) toadd=atoi(argv[1]);
   if(argc>2) colour=atoi(argv[2]);
   if(argc>3) brightness=atoi(argv[3]);
   if(brightness<0) brightness=0;
   if(brightness>65535) brightness=65535;
   if(argc>4) contrast=atoi(argv[4]);
   if(contrast<0) contrast=0;
   if(contrast>65535) contrast=65535;

   grab_open();

   get_sh_mem();

   signal(SIGHUP, myabort);
   signal(SIGINT, myabort);
   signal(SIGQUIT,myabort);
   signal(SIGILL, myabort);
   signal(SIGABRT,myabort);
   signal(SIGFPE, myabort);
   signal(SIGSEGV,myabort);
   signal(SIGPIPE,myabort);
   signal(SIGTERM,myabort);
   signal(SIGUSR1,myabort);
   signal(SIGUSR2,myabort);

   grab_init(colour);

   printf("VidIO Capabilities for %s:\n",GRAB_DEVICE);
   printf("\tname=%s\n",grab_cap.name);
   printf("\ttype=0x%x:",grab_cap.type);
   if(grab_cap.type&VID_TYPE_CAPTURE)    printf("\tCan capture to memory\n");
   if(grab_cap.type&VID_TYPE_TUNER)      printf("\t\t\tHas a tuner of some form\n");
   if(grab_cap.type&VID_TYPE_TELETEXT)   printf("\t\t\tHas teletext capability\n");
   if(grab_cap.type&VID_TYPE_OVERLAY)    printf("\t\t\tCan overlay its image onto the frame buffer\n");
   if(grab_cap.type&VID_TYPE_CHROMAKEY)  printf("\t\t\tOverlay is Chromakeyed\n");
   if(grab_cap.type&VID_TYPE_CLIPPING)   printf("\t\t\tOverlay clipping is supported\n");
   if(grab_cap.type&VID_TYPE_FRAMERAM)   printf("\t\t\tOverlay overwrites frame buffer memory\n");
   if(grab_cap.type&VID_TYPE_SCALES)     printf("\t\t\tThe hardware supports image scaling\n");
   if(grab_cap.type&VID_TYPE_MONOCHROME) printf("\t\t\tImage capture is grey scale only\n");
   if(grab_cap.type&VID_TYPE_SUBCAPTURE) printf("\t\t\tCapture can be of only part of the image\n");
   printf("\tTV channels=%d  ",grab_cap.channels);
   printf("\tAudio devices=%d\n",grab_cap.audios);
   printf("\tMaxWidth=%d MaxHeight=%d ",grab_cap.maxwidth,grab_cap.maxheight);
   printf("\t(MinWidth=%d MinHeight=%d)\n",grab_cap.minwidth,grab_cap.minheight);
   printf("Image Properties:\n");
   printf("\tbrightness=%d->%d\n", grab_pict_save.brightness,grab_pict.brightness);
   printf("\thue       =%d\n", grab_pict.hue      );
   printf("\tcolour    =%d->%d\n", grab_pict_save.colour, grab_pict.colour);
   printf("\tcontrast  =%d->%d\n", grab_pict_save.contrast, grab_pict.contrast );
   printf("\twhiteness =%d\n", grab_pict.whiteness);
   printf("\tdepth     =%d\n", grab_pict.depth    );
   printf("\tpalette   =%d\n", grab_pict.palette  );


   nadd = curch->nadd;
   seqn = curch->seqn;
   img  = curch->image;
   imgw = curch->img_width;
   imgh = curch->img_height;
   col0 = curch->frm_col0;
   lin0 = curch->frm_lin0;
   width= curch->frm_width;
   height=curch->frm_height;

   for(nadd=1; nadd<=toadd; nadd++) {
      if((grab_cadr = grab_next(nadd==1, nadd==toadd, &grabw, &grabh)) == NULL)
	 exit(1);
      printf("Grab%d: width=%d height=%d\n",grab_cnt,grabw,grabh);

      i=grabh*grabw/2+grabw/3;
      ii=i+20;
      if(grab_buff[0].format == VIDEO_PALETTE_RGB24) {
	 p += i*3;
	 for(;i<ii;i++) {
	    printf("<%02x",  *p++);
	    printf(" %02x ", *p++);
	    printf("%02x> ", *p++);
	    if(i%7==6) printf("\n");
	 }
      } else {
	 p += i;
	 for(;i<ii;i++) {
	    printf(" %02x", *p++);
	    if(i%20==19) printf("\n");
	 }
      }
      printf("\n");


      if(grab_buff[0].format == VIDEO_PALETTE_RGB24)
	 for (j=0; j<height; j++) {
	    y=lin0+j;
	    p= &grab_cadr[(grabw*y+col0)*3];
	    q = &img[imgw*y+col0];
	    if(nadd==1)
	       for (i=0; i<width; i++) *q++ = *p++ + *p++ + *p++;
	    else
	       for (i=0; i<width; i++) *q++ += *p++ + *p++ + *p++;

	 }
      else
	 for (j=0; j<height; j++) {
	    y=lin0+j;
	    p= &grab_cadr[grabw*y+col0];
	    q = &img[imgw*y+col0];
	    if(nadd==1)
	       for (i=0; i<width; i++) *q++ = *p++;
	    else
	       for (i=0; i<width; i++) *q++ += *p++;

	 }
      curch->nadded = nadd;
      curch->seqn++;
      sprintf(tmp,"Frame%d (Add%d)",curch->seqn,curch->nadded);
      chan_msg(tmp);
   }
   vid_exit(0);
}


/* Allocate shared memory segment */
static void get_sh_mem() {
   int i,j, shmid, new_shm=0;
   short *p;
   int shm_size = sizeof(struct TVmem)+sizeof(struct TVchan)*(BTNumCh-1);
   int chip_type=848;
   char card_name[32] = "BT848";
   int my_pid;


   if(curch != NULL) return;

   /* first try to find existing one */
   shmid = shmget(bttv_shm_key.code, shm_size+sizeof(int), 0777);
   if (shmid<0 && errno==ENOENT) {
      /* if no - try to create a new one */
      shmid = shmget(bttv_shm_key.code, shm_size+sizeof(int), IPC_CREAT|IPC_EXCL|0777);
      new_shm=1;
   }
   if (shmid<0) {
      char msg[80];
      sprintf(msg,"Can't create shared memory segment '%s'",bttv_shm_key.name);
      perror(msg);
      exit(errno);
   }

   /* attach it to our memory space */
   shm_addr = shmat ( shmid, NULL, 0 );
   if ((int)(shm_addr) == -1) {
      char msg[80];
      sprintf(msg,"Can't attach shared memory segment '%s'",bttv_shm_key.name);
      perror(msg);
      exit(errno);
   }
   if((shmctl(shmid, SHM_LOCK, NULL) < 0) < 0) {
      char msg[80];
      sprintf(msg,"Can't prevents swapping of shared memory segment '%s'",bttv_shm_key.name);
      perror(msg);
      exit(errno);
   }

   strncpy(card_name,(grab_fd>0)?grab_cap.name:"BT848",32);

   tvmem = (struct TVmem *)shm_addr;
   if(new_shm) {
      tvmem->size = shm_size;
      tvmem->drv_pid = 0;
      fprintf(stderr,"Create shared memory segment '%s' %dbytes at %x \n",
	   bttv_shm_key.name, shm_size, shm_addr);
   } else {
      fprintf(stderr,"Attach shared memory segment '%s' %dbytes at %x \n",
	   bttv_shm_key.name, tvmem->size, shm_addr);
      if(tvmem->size != shm_size) {
	 fprintf(stderr,"Wrong TV-memory size:%dbytes needs!\n",shm_size);
	 exit(1);
      }
      if(tvmem->numch != BTNumCh) {
	 fprintf(stderr,"Wrong TV-channels number:%d (%d needs)!\n",tvmem->numch,BTNumCh);
	 new_shm=1;
      }
      if(strncmp(tvmem->grab_name,card_name,32) != 0) {
	 fprintf(stderr,"TV-card change?! %s -> %s?!\n",tvmem->grab_name,card_name);
	 new_shm=1;
      }
   }

   my_pid = getpid();
   if( tvmem->drv_pid > 0 && kill(tvmem->drv_pid, 0) >= 0) {
      fprintf(stderr,"vs_bttv daemon already running! (PID=%d)\n", tvmem->drv_pid);
      sleep(2);
      exit(1);
   }
   strncpy(tvmem->grab_name,card_name,32);
   tvmem->numch = BTNumCh;
   tvmem->drv_pid = my_pid;

   if(strstr(tvmem->grab_name,"849")) chip_type=849;
   if(strstr(tvmem->grab_name,"878")) chip_type=878;
   if(strstr(tvmem->grab_name,"879")) chip_type=879;

   for(i=0,curch=tvmem->chmem; i<NumCh; i++,curch++) {
      tvch[i] = curch;
      if(new_shm || curch->vs_type!=chip_type) {
	 curch->num=i;
	 curch->vs_type=chip_type;
	 curch->tandem=0;
	 curch->activ=0;
	 curch->exposure = 0.04;
	 if(new_shm)
	    curch->seqn=0;
	 curch->tstart = curch->tcurr = time(0);
	 curch->pid=0;
	 curch->command=INULL;
	 curch->nexp=0;
	 curch->nadd=0;
	 curch->nadded=0;
	 curch->addlev=0;
	 curch->next_form=1;
	 curch->form=1;
	 curch->ctrl.bttv.brightness=grab_pict.brightness>>8;
	 curch->ctrl.bttv.contrast=grab_pict.contrast>>8;
	 curch->img_width  = curch->frm_width  = curch->subfrm_width  = GRAB_WIDTH_1;
	 curch->img_height = curch->frm_height = curch->subfrm_height = GRAB_HEIGHT_1;
	 curch->frm_col0 = curch->frm_lin0 = curch->subfrm_col0 = curch->subfrm_lin0 = 0;
	 curch->pnt_col = curch->pnt_lin = 0;
	 curch->center_x = curch->center_y = 0.0;
	 if(new_shm) {
	    curch->seqmsg=0;
	    curch->msg[0]='\0';
	 }
	 curch->bias = 0;
	 for(j=0; j<MaxWidth; j++)
	     curch->flat[j] = 1;
	 for(p=curch->image,j=0; j<MaxWidth*MaxHeight; j++)
	     *p++ = (j % curch->img_width) - 100;
      }
   }
   curch = tvmem->chmem;
}


static void chan_msg( char *new_msg) {
    strncpy(curch->msg, new_msg, 99);
    curch->seqmsg++;
}
static void wall_msg( char *new_msg) {
    int nch;
    for(nch=0; nch<NumCh; nch++) {
	strncpy(tvch[nch]->msg, new_msg, 99);
	tvch[nch]->seqmsg++;
    }
}

static int vd_hup=0;

static void mydelay(double dt)
{
  int nfd;
  struct timeval tv;
   tv.tv_sec = (int)dt;
   tv.tv_usec = (int)((dt - tv.tv_sec)*1000000.);
sleeping:
   nfd = select(0, (fd_set *)NULL,(fd_set *)NULL,(fd_set *)NULL, &tv);
   if(nfd < 0) {
      if(errno == EINTR) { /* interrupted?    */
	 if(vd_hup) {      /* by USR1 signal?  */
	    vd_hup=0;      /*  ok!              */
	    return;        /*    stop sleeping...*/
	 }
	/*On  Linux,  timeout  is  modified to reflect the amount of
	 time not slept; most other implementations DO NOT do this!*/
	 goto sleeping;
      }
      fprintf(stderr,"Error in mydelay(){ select() }. %s\n",strerror(errno));
   }
}

static double mytime()
{
  struct timeval tv;
  struct timezone tz;
  int ret;
  double t;
  ret = gettimeofday(&tv, &tz);
  t = (double)tv.tv_sec + (double)tv.tv_usec*0.000001;
  t -= timezone;
  return(t);
}

static void myabort(int sig) {
    int ret;
    char ss[10], tmp[80];
    signal(sig,SIG_IGN);
    switch (sig) {
	case SIGHUP : strcpy(ss,"SIGHUP");  break;
	case SIGINT : strcpy(ss,"SIGINT");  break;
	case SIGQUIT: strcpy(ss,"SIGQUIT"); break;
	case SIGILL : strcpy(ss,"SIGILL");  break;
	case SIGABRT: strcpy(ss,"SIGABRT"); break;
	case SIGFPE : strcpy(ss,"SIGFPE");  break;
	case SIGPIPE: strcpy(ss,"SIGPIPE"); break;
	case SIGSEGV: strcpy(ss,"SIGSEGV"); break;
	case SIGTERM: strcpy(ss,"SIGTERM"); break;
	default:  sprintf(ss,"SIG_%d",sig); break;
    }
    switch (sig) {
	case SIGUSR1 :
	     vd_hup=1;
	     signal(sig, myabort);
	     return;
	default:
	case SIGINT :
	case SIGUSR2 :
	     fprintf(stderr,"%s: %s - Ignore .....\n",myname,ss);
	     fflush(stderr);
	     signal(sig, myabort);
	     return;
	case SIGHUP :
	case SIGQUIT:
	case SIGILL :
	case SIGABRT:
	case SIGFPE :
	case SIGSEGV:
	case SIGPIPE:
	case SIGTERM:
	     signal(SIGALRM, SIG_IGN);
	     fprintf(stderr,"%s: %s - programm stop!\n",myname,ss);
	     vid_exit(sig);
    }
}

static void vid_exit(int n) {
   tvmem->drv_pid = 0;
   /* restore image properties */
   if (ioctl(grab_fd,VIDIOCSPICT,&grab_pict_save) < 0 ) {
       perror("ioctl VIDIOCSPICT");
   }
   close(grab_fd);
   exit(n);
}

