
/*      BT848/BT878
 *     VS60/VS56
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/time.h>
#include <sched.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/mman.h>
#include <sys/ioctl.h>

#include <asm/types.h>
#include "linux/videodev.h"   /* but "videodev2.h" for v4l2 */

#include "vs56_shm.h"         /* to use all VS56 image tools */
#include "bttv_shm_key.h"


#define GRAB_DEVICE       "/dev/video0"
#define GRAB_WIDTH_1      MaxWidth
#define GRAB_HEIGHT_1     MaxHeight
#define GRAB_WIDTH_2      (MaxWidth/3)*2
#define GRAB_HEIGHT_2     (MaxHeight/3)*2

static void get_sh_mem();
static void chan_msg( char *new_msg);
static void wall_msg( char *new_msg);
static void mydelay(double dt);
static double mytime();
static void myabort(int);
static void vid_exit(int n);

char *myname;
static unsigned char *shm_addr;           /* shm-block with buffers   */
static struct TVmem  *tvmem;              /* TV-memory structure      */
static struct TVchan *tvch[MaxNumCh];     /* for TV-channels & images */
static struct TVchan *curch=NULL;         /* current working channel  */

static int vd_hup=0;                      /* clients USR1-signal flag */

/* ---------------------------------------------------------------------- */
/* capture stuff  -  old v4l (bttv)                                       */
/* these work for v4l only, not v4l2 */

#ifdef VIDIOCGCAP

# define GRAB_SOURCE      0
# define GRAB_NORM        VIDEO_MODE_PAL

static struct video_capability   grab_cap;
static struct video_mbuf         grab_buffers;
static struct video_mmap         grab_buff[2];
static struct video_channel      grab_chan;
static struct video_picture      grab_pict, grab_pict_save;
static int  brightness=120, contrast=80;
static int  grab_fd=-1, grab_size, buf_cnt, grab_need_sync=0;
static unsigned int grab_cnt;
static unsigned char *grab_data, *grab_cadr;

static void grab_open(char *grab_dev_name)
{
    static struct sched_param p;
    int pid=0;
    char tmp[100];

    p.sched_priority = 1;
    if (sched_setscheduler(pid, SCHED_FIFO, &p)) {
       perror("Can't set REAL TIME mode (SuperUser?)");
    } else {
       if (mlockall(MCL_CURRENT)) {
	  perror("Can't lock process memory");
       }
    }
    /* open video device*/
    if ((grab_fd = open(grab_dev_name,O_RDWR)) < 0 ) {
	sprintf(tmp,"Can't open Video Device %s",grab_dev_name);
	perror(tmp);
	exit(1);
    }
    /* obtain video device capability information */
    if (ioctl(grab_fd,VIDIOCGCAP,&grab_cap) < 0 ) {
	fprintf(stderr,"Wrong video device %s?!\n",grab_dev_name);
	exit(1);
    }
    /* get image source and TV norm */
    grab_chan.channel = GRAB_SOURCE;
    if (ioctl(grab_fd,VIDIOCGCHAN,&grab_chan) < 0 ) {
	perror("ioctl VIDIOCGCHAN");
	exit(1);
    }
    /* get current image properties */
    if (ioctl(grab_fd,VIDIOCGPICT,&grab_pict) < 0 ) {
	perror("ioctl VIDIOCGPICT");
	exit(1);
    }
    grab_pict_save = grab_pict;
}

static void grab_init(int colour)
{
    int i;

    if(grab_fd<0) exit(1); /*against idiots protection :-)*/

    /* set image source and TV norm */
    grab_chan.channel = GRAB_SOURCE;    
    grab_chan.norm    = GRAB_NORM;
    if (ioctl(grab_fd,VIDIOCSCHAN,&grab_chan) < 0 ) {
	perror("ioctl VIDIOCSCHAN");
	exit(1);
    }

    /* set new image properties */
    grab_pict.brightness = brightness<<8;
    grab_pict.contrast = contrast<<8;
    grab_pict.colour = colour<<8;
    if(colour)
       grab_pict.palette = VIDEO_PALETTE_RGB24;
    else
       grab_pict.palette = VIDEO_PALETTE_GREY;
    if (ioctl(grab_fd,VIDIOCSPICT,&grab_pict) < 0 ) {
	perror("ioctl VIDIOCSPICT");
	exit(1);
    }

    /* map grab buffer */
    if (ioctl(grab_fd,VIDIOCGMBUF,&grab_buffers) < 0) {
	perror("ioctl VIDIOCGMBUF");
	vid_exit(1);
    }
    grab_data = mmap(0,grab_buffers.size,PROT_READ|PROT_WRITE,MAP_SHARED,grab_fd,0);
    if ((int)grab_data == -1) {
	perror("mmap");
	vid_exit(1);
    }
    buf_cnt = (grab_buffers.frames <= 2)? grab_buffers.frames : 2;
    grab_cnt = 0;
    for(i=0; i<buf_cnt; i++) {
	if(colour)
	   grab_buff[i].format = VIDEO_PALETTE_RGB24;
	else
	   grab_buff[i].format = VIDEO_PALETTE_GREY;
	grab_buff[i].frame  = i;
	grab_buff[i].width  = (GRAB_WIDTH_1<grab_cap.maxwidth)? GRAB_WIDTH_1 : grab_cap.maxwidth;
	grab_buff[i].height = (GRAB_HEIGHT_1<grab_cap.maxheight)? GRAB_HEIGHT_1 : grab_cap.maxheight;
    }
    grab_size = grab_buff[0].width * grab_buff[0].height;
    if(colour) grab_size *= 3;
}

static void swap_rgb24(char *mem, int n)
{
    char  c;
    char *p = mem;
    int   i = n;
    
    while (--i) {
	c = p[0]; p[0] = p[2]; p[2] = c;
	p += 3;
    }
}

static unsigned char* grab_next(int first, int last, int *width, int *height)
{
    int old_frame = grab_cnt % buf_cnt;
    int new_frame = (grab_cnt+1) % buf_cnt;

    if(grab_fd<0) return(NULL); /*against idiots protection :-)*/

    if( first && grab_need_sync==0) {
	if (ioctl(grab_fd,VIDIOCMCAPTURE,&grab_buff[old_frame]) < 0) {
	    perror("ioctl VIDIOCMCAPTURE");
	    return(NULL);
	}
	grab_need_sync++;
/*printf("VIDIOCMCAPTURE for grab_buf%d (grab_offset=%d)\n",grab_buff[old_frame].frame, grab_buffers.offsets[old_frame]);*/
    }
    if( !last && grab_need_sync<buf_cnt) {
	if (ioctl(grab_fd,VIDIOCMCAPTURE,&grab_buff[new_frame]) < 0) {
	    perror("ioctl VIDIOCMCAPTURE");
	    return(NULL);
	} else
	    grab_cnt++;
	grab_need_sync++;
/*printf("VIDIOCMCAPTURE for grab_buf%d (grab_offset=%d)\n",grab_buff[new_frame].frame, grab_buffers.offsets[new_frame]);*/
    }
    if (grab_need_sync>0) {
vidiosync:
	if (ioctl(grab_fd,VIDIOCSYNC,&grab_buff[old_frame]) < 0) {
	   if(errno == EINTR) { /* interrupted?    */
	      if(vd_hup)        /* by USR1 signal?  */
		 vd_hup=0;      /*  ok!              */
	      else {
		 perror("ioctl VIDIOCSYNC");
		 mydelay(0.1);
	      }
	      goto vidiosync;   /* continue with VIDIOCSYNC */
	   }
	   perror("ioctl VIDIOCSYNC");
	   grab_need_sync=0;
	   return(NULL);
       } else {
	   grab_need_sync--;
	 /*if(grab_buff[old_frame].format == VIDEO_PALETTE_RGB24)
	  *   swap_rgb24(grab_data,grab_buff[old_frame].width*grab_buff[old_frame].height);
	  */
	   *width  = grab_buff[old_frame].width;
	   *height = grab_buff[old_frame].height;
/*printf("grab_cnt=%d grab_offset=%d\n",grab_cnt, grab_buffers.offsets[old_frame]);*/
	   return(grab_data + grab_buffers.offsets[old_frame]);
       }
    } else {
       fprintf(stderr,"Internal error! No frame for VIDIOCSYNC! Use old one...\n");
       return(grab_data + grab_buffers.offsets[old_frame]);
    }
}

static void grab_set_chan(int ch_num, int colour, int ch_bright, int ch_contr)
{
    int i,width,height;

    if(grab_fd<0) return; /*against idiots protection :-)*/

    if(grab_need_sync)
       grab_cadr = grab_next(0, 1, &width, &height);

    /* get image source and TV norm */
    grab_chan.channel = ch_num;
    if (ioctl(grab_fd,VIDIOCGCHAN,&grab_chan) < 0 ) {
	perror("ioctl VIDIOCGCHAN");
	vid_exit(1);
    }
    /* set image source and TV norm */
    grab_chan.channel = ch_num;
    grab_chan.norm    = GRAB_NORM;
    if (ioctl(grab_fd,VIDIOCSCHAN,&grab_chan) < 0 ) {
	perror("ioctl VIDIOCSCHAN");
	vid_exit(1);
    }

    /* set new image properties */
    grab_pict.brightness = ch_bright<<8;
    grab_pict.contrast = ch_contr<<8;
    grab_pict.colour = colour<<8;
    if(colour)
       grab_pict.palette = VIDEO_PALETTE_RGB24;
    else
       grab_pict.palette = VIDEO_PALETTE_GREY;
    if (ioctl(grab_fd,VIDIOCSPICT,&grab_pict) < 0 ) {
	perror("ioctl VIDIOCSPICT");
	vid_exit(1);
    }
}

#endif

int main (int argc, char *argv[])
{
   short *img;                     /* image memory */
   int imgw, imgh;                 /* image sizes   */
   int col0, lin0, width, height;  /* frame: area of interest */
   int nch, i,j,  x,y, fsize,  nadd, seqn;
   unsigned int img_cnt;
   int ii, grabw, grabh, colour=0;
   int ccd_mode, ccd_favr, *ccd_f;
   int quick_mode, quick_add_mode, slow_add_mode;
   unsigned char *p;
   unsigned short *q;
   int first[BTNumCh], lev[BTNumCh], brt_val[BTNumCh], cnt_val[BTNumCh];
   int hist[1024];
   char tmp[100];
   char *grab_dev = GRAB_DEVICE;
   double tlast, tdelay;

   myname = argv[0];
   if(argc>1 && (argv[1][0]=='-' && (argv[1][1]=='h' ||( argv[1][1]=='-' && argv[1][2]=='h')))) {
	fprintf(stderr,"Usage of '%s' - daemon for BT848/B878-based grabbers:\n",myname);
	fprintf(stderr,"\t%s\n",myname);
	fprintf(stderr,"or:\n");
	fprintf(stderr,"\t%s [dev_name [colour [brightness [contrast]]]]\n",myname);
	fprintf(stderr,"Where:\n");
	fprintf(stderr,"\t\tdev_name - Video device (default: %s)\n",GRAB_DEVICE);
	fprintf(stderr,"\t\tcolour - 0-255 video input colour level\n");
	fprintf(stderr,"\t\t\t(default: 0 - greyscale video input)\n");
	fprintf(stderr,"\t\t\t(else: 1-255 - software conversion to greyscale)\n");
	fprintf(stderr,"\t\tbrightness - 0-255 initial level (default: %d)\n",brightness);
	fprintf(stderr,"\t\tcontrast - 0-255 initial level (default: %d)\n",contrast);
	exit(0);
   }
   if(argc>1) grab_dev=argv[1];
   if(argc>2)     colour=atoi(argv[2]);
   if(colour<0)   colour=0;
   if(colour>255) colour=255;
   if(argc>3)         brightness=atoi(argv[3]);
   if(brightness<0)   brightness=0;
   if(brightness>255) brightness=255;
   if(argc>4)       contrast=atoi(argv[4]);
   if(contrast<0)   contrast=0;
   if(contrast>255) contrast=255;

   grab_open(grab_dev);

   get_sh_mem();

   signal(SIGHUP, myabort);
   signal(SIGINT, myabort);
   signal(SIGQUIT,myabort);
   signal(SIGILL, myabort);
   signal(SIGABRT,myabort);
   signal(SIGFPE, myabort);
   signal(SIGSEGV,myabort);
   signal(SIGPIPE,myabort);
   signal(SIGTERM,myabort);
   signal(SIGUSR1,myabort);
   signal(SIGUSR2,myabort);

   grab_init(colour);

   printf("VidIO Capabilities for %s:\n",grab_dev);
   printf("\tname=%s\n",grab_cap.name);
   printf("\ttype=0x%x:",grab_cap.type);
   if(grab_cap.type&VID_TYPE_CAPTURE)    printf("\tCan capture to memory\n");
   if(grab_cap.type&VID_TYPE_TUNER)      printf("\t\t\tHas a tuner of some form\n");
   if(grab_cap.type&VID_TYPE_TELETEXT)   printf("\t\t\tHas teletext capability\n");
   if(grab_cap.type&VID_TYPE_OVERLAY)    printf("\t\t\tCan overlay its image onto the frame buffer\n");
   if(grab_cap.type&VID_TYPE_CHROMAKEY)  printf("\t\t\tOverlay is Chromakeyed\n");
   if(grab_cap.type&VID_TYPE_CLIPPING)   printf("\t\t\tOverlay clipping is supported\n");
   if(grab_cap.type&VID_TYPE_FRAMERAM)   printf("\t\t\tOverlay overwrites frame buffer memory\n");
   if(grab_cap.type&VID_TYPE_SCALES)     printf("\t\t\tThe hardware supports image scaling\n");
   if(grab_cap.type&VID_TYPE_MONOCHROME) printf("\t\t\tImage capture is grey scale only\n");
   if(grab_cap.type&VID_TYPE_SUBCAPTURE) printf("\t\t\tCapture can be of only part of the image\n");
   printf("\tTV channels=%d  ",grab_cap.channels);
   printf("\tAudio devices=%d\n",grab_cap.audios);
   printf("\tMaxWidth=%d MaxHeight=%d ",grab_cap.maxwidth,grab_cap.maxheight);
   printf("\t(MinWidth=%d MinHeight=%d)\n",grab_cap.minwidth,grab_cap.minheight);
   printf("Image Properties:\n");
   printf("\tbrightness=%d->%d\n", grab_pict_save.brightness>>8,grab_pict.brightness>>8);
   printf("\thue       =%d\n", grab_pict.hue>>8      );
   printf("\tcolour    =%d->%d\n", grab_pict_save.colour>>8, grab_pict.colour>>8);
   printf("\tcontrast  =%d->%d\n", grab_pict_save.contrast>>8, grab_pict.contrast>>8 );
   printf("\twhiteness =%d\n", grab_pict.whiteness>>8);
   printf("\tdepth     =%d\n", grab_pict.depth    );
   printf("\tpalette   =%d", grab_pict_save.palette);
   switch(grab_pict_save.palette) {
      case VIDEO_PALETTE_GREY:   printf("(GREY)");  break;
      case VIDEO_PALETTE_RGB565: printf("(RGB565)"); break;
      case VIDEO_PALETTE_RGB24:  printf("(RGB24)"); break;
   }
   if(grab_pict_save.palette!=grab_pict.palette) {
      printf("->%d", grab_pict.palette);
      switch(grab_pict.palette) {
	 case VIDEO_PALETTE_GREY:   printf("(GREY)");  break;
	 case VIDEO_PALETTE_RGB565: printf("(RGB565)"); break;
	 case VIDEO_PALETTE_RGB24:  printf("(RGB24)"); break;
      }
   }
   printf("\n");

   for(nch=0; nch<NumCh; nch++) {
      first[nch] = 1;
      lev[nch] = 0;
      brt_val[nch] = tvch[nch]->ctrl.bttv.brightness;
      cnt_val[nch] = tvch[nch]->ctrl.bttv.contrast;
   }
   tlast = mytime();
   tdelay = 0.1;
   img_cnt = grab_cnt;

   while(1) {
     for(nch=0; nch<NumCh; nch++) {
       if(tvch[nch]->command == INULL) {
	  continue;
       }
       if(tvch[nch]->next_form != tvch[nch]->form) {
	  if(grab_need_sync)
	     grab_cadr = grab_next(0, 1, &grabw, &grabh);
	  if(tvch[nch]->next_form == 1) {
	     imgw = GRAB_WIDTH_1; imgh = GRAB_HEIGHT_1;
	  } else {
	     imgw = GRAB_WIDTH_2; imgh = GRAB_HEIGHT_2;
	  }
	  if(imgw>grab_cap.maxwidth)  imgw = grab_cap.maxwidth;
	  if(imgh>grab_cap.maxheight) imgh = grab_cap.maxheight;
	  for(i=0; i<buf_cnt; i++) {
	      grab_buff[i].width  = imgw;
	      grab_buff[i].height = imgh;
	  }
	  grab_size = imgw * imgh;
	  if(colour) grab_size *= 3;
	  tvch[nch]->img_width  = imgw;
	  tvch[nch]->img_height = imgh;
	  tvch[nch]->frm_width     = imgw;
	  tvch[nch]->frm_height    = imgh;
	  tvch[nch]->subfrm_width  = tvch[nch]->subfrm2_width  = tvch[nch]->frm_width;
	  tvch[nch]->subfrm_height = tvch[nch]->subfrm2_height = tvch[nch]->frm_height;

	  tvch[nch]->form = tvch[nch]->next_form;
	  fprintf(stderr,"Change Chan%d cadr format: %dx%d\n",nch,imgw,imgh);
       }
       if(tvch[nch]->command == IINIT) {
printf("BTTV Init\n",nch);
	  tvch[nch]->command = INULL;

	  wall_msg("BTTV Init");
	  continue;
       }
       tdelay = 0.021;
       if(tvch[nch]->command&ICLEAR) {
	  first[nch] = 1;
	  lev[nch] = 0;
	  tvch[nch]->nadded = 0;
	  tvch[nch]->addlev = 0;
	  tvch[nch]->tstart = tvch[nch]->tcurr = mytime();
	  if((tvch[nch]->command&(IADD|IQUICK)) == 0) {
	     for(q=tvch[nch]->image,j=0; j<MaxWidth*MaxHeight; j++) *q++ = 0;
	     tvch[nch]->seqn++;
	     tvch[nch]->command &= ~ICLEAR;
	  }
       }
       if(tvch[nch]->command&(IADD|IQUICK)) {
	  if(tvch[nch]->nadd==0 || tvch[nch]->nexp==0) {
	     tvch[nch]->command = INULL;
	     continue;
	  }
	  if(curch!=tvch[nch] ||
	     brt_val[nch] != tvch[nch]->ctrl.bttv.brightness ||
	     cnt_val[nch] != tvch[nch]->ctrl.bttv.contrast) {

	     brt_val[nch] = tvch[nch]->ctrl.bttv.brightness;
	     cnt_val[nch] = tvch[nch]->ctrl.bttv.contrast;
	     grab_set_chan(nch, colour, brt_val[nch], cnt_val[nch]);
	     if(curch!=tvch[nch]) {
		grab_cadr = grab_next(1, 0, &grabw, &grabh);
		grab_cadr = grab_next(0, 1, &grabw, &grabh);
	     }
	     curch=tvch[nch];
	     fprintf(stderr,"Set Chan%d Brightness=%d Contrast=%d\n",nch,brt_val[nch], cnt_val[nch]);
	  }
	  quick_mode = (curch->command&IQUICK)? 1 : 0;
	  quick_add_mode = ((curch->command&IQUICK) && curch->nadd>1)? 1 : 0;
	  slow_add_mode = (((curch->command&IQUICK)==0 && curch->nadd>1)||(curch->nadd<0))? 1 : 0;
	  ccd_mode = (curch->command&ICCD)? 1 : 0;


	  ccd_f = curch->flat;
	  nadd = curch->nadd;
	  seqn = curch->seqn;
	  img  = curch->image;
	  imgw = curch->img_width;
	  imgh = curch->img_height;
	  col0 = curch->frm_col0;
	  lin0 = curch->frm_lin0;
	  width= curch->frm_width;
	  height=curch->frm_height;
	  fsize = width*height;

	  if(ccd_mode) {
	     ccd_favr = 1;
	     for(i=0; i<width; i++) {
		x=col0+i;
		if(ccd_f[x]<1) ccd_f[x]=1;
		ccd_favr += ccd_f[x];
	     }
	     ccd_favr = (ccd_favr + width/2) / width;
	  }

	  for(; nadd>0||curch->nadd<0; nadd--) {
	     int lev0 = (first[nch] || quick_mode)? 0 : lev[nch];

	     if(grab_need_sync) {      /* is grabbing already in progress? */
		if(img_cnt == grab_cnt) {
		   fprintf(stderr,"Internal error! Wrong grabbing frame sequence! \n");
		   if((grab_cadr = grab_next(0, 1, &grabw, &grabh)) == NULL)
		      vid_exit(1);
		} else if(img_cnt == grab_cnt-1) { /* next image needs ? */
		   img_cnt = grab_cnt;
		   if((grab_cadr = grab_next(0, 0, &grabw, &grabh)) == NULL)
		      vid_exit(1);
		} else { /* if (img_cnt<grab_cnt-1)  image already ok! Use it...*/
		   img_cnt = grab_cnt-1;
		}
	     } else {                  /* no, start grabbing process*/
		img_cnt = grab_cnt;
		if((grab_cadr = grab_next(1, 0, &grabw, &grabh)) == NULL)
		   vid_exit(1);
/*printf("Start grab; img_cnt=%d\n",img_cnt);*/
	     }
	   /*if((grab_cadr = grab_next(nadd==curch->nadd, nadd==1||curch->nadd==0, &grabw, &grabh)) == NULL)
	    *    vid_exit(1);
	    */
	   /*printf("Grab%d: width=%d height=%d\n",grab_cnt,grabw,grabh);
	    */
	     tvch[nch]->tcurr = tlast = mytime();

	     if(colour)
		for (j=0; j<height; j++) {
		   y=lin0+j;
		   p= &grab_cadr[(grabw*y+col0)*3];
		   q = &img[imgw*y+col0];
		   if(first[nch])
		      for (i=0; i<width; i++) *q++ = *p++ + *p++ + *p++;
		   else if( !quick_add_mode )
		      for (i=0; i<width; i++) *q++ += *p++ + *p++ + *p++ - lev0;
		   else
		      for (i=0; i<width; i++) *q++ += *p++ + *p++ + *p++;

		}
	     else
		for (j=0; j<height; j++) {
		   y=lin0+j;
		   p= &grab_cadr[grabw*y+col0];
		   q = &img[imgw*y+col0];
		   if(first[nch])
		      for (i=0; i<width; i++) *q++ = *p++;
		   else if( !quick_add_mode )
		      for (i=0; i<width; i++) *q++ += *p++ - lev0;
		   else
		      for (i=0; i<width; i++) *q++ += *p++;

		}
	     curch->nadded++;
	     if(!first[nch] && !quick_add_mode)
	       curch->addlev += lev0;

	     if(!quick_add_mode) {
		curch->seqn++;
		sprintf(tmp,"Frame%d (Add%d SubLev=%d)",
		   curch->seqn,curch->nadded,curch->addlev);
		chan_msg(tmp);
	     } else {                  /* "quick" mode */
		seqn++;
	     }

	     if(first[nch] && !quick_mode) {
		int i05,i95,n05,n95,n100,si,sj;
		int nhist = (colour)? 256*3 : 256;
		double havr;

		for(i=0; i<nhist; i++) hist[i]=0;
		si = width/100+1;
		sj = height/100+1;
		n100=0;
		for (j=0; j<height; j+=sj) {
		   y=lin0+j;
		   q = &img[imgw*y+col0];
		   for (i=0; i<width; i+=si) {hist[*q++]++;n100++;}
		}
		n05 = (int)(n100*0.05 + 0.5);
		n95 = (int)(n100*0.95 + 0.5);
		i05=0;
		i95=255;
		for(i=j=0; i<nhist; i++) {
		   int j0 = j;
		   j += hist[i];
		   if(j0<n05 && n05<=j) i05=i;
		   if(j0<n95 && n95<=j) i95=i;
		}
		for(havr=0.,j=0,i=i05; i<=i95; i++) {
		   havr += hist[i]*i;
		   j += hist[i];
		}
		if(j>1) havr /= j;
		lev[nch] = lev0 = (int)(havr+0.5);
	     }
	     if(ccd_mode && !quick_add_mode) {
		for (j=0; j<height; j++) {
		   int *cfp = ccd_f+col0;
		   y=lin0+j;
		   q = &img[imgw*y+col0];
		   for (i=0; i<width; i++) {
		      *q++ -= ((*cfp++ - ccd_favr)>>8);
		   }
		}
	     }
	     first[nch] = 0;
	  }
	  if(ccd_mode && quick_add_mode) {
	     for (j=0; j<height; j++) {
		int *cfp = ccd_f+col0;
		y=lin0+j;
		q = &img[imgw*y+col0];
		for (i=0; i<width; i++,q++,cfp++) {
		   *q -= ((*cfp-ccd_favr)*curch->nadd + 128) >> 8;
		}
	     }
	  }
	  if(quick_add_mode) {
	     curch->seqn = seqn;
	     sprintf(tmp,"Frame%d (Add%d SubLev=%d)",
		  curch->seqn,curch->nadded,curch->addlev);
	     chan_msg(tmp);
	  }
	  if(curch->nexp>0)
	     curch->nexp--;
	  if(curch->nexp==0) {
/*printf("Chan%d: Nadd=%d dTime=%.3f\n",curch->num, curch->nadd, curch->tcurr-curch->tstart);*/
	     curch->command &= ~(IEXP|IQUICK);
	     curch->nadd = 0;
	     if(curch->pid != 0) {  /* client needs ACK? */
		if(kill(curch->pid, SIGUSR1)<0)
		   perror("Sending ACK-signal to client");
		curch->pid = 0;
	     }
	  }
       }
     }
     vd_hup=0;

     if(grab_need_sync) {      /* grabbing in progress now? */
	double tcurr = mytime();
	if(tcurr-tlast < 1.0) { /* continue with grabbing up to 1 sec */
	   if((grab_cadr = grab_next(0, 0, &grabw, &grabh)) == NULL)
	      vid_exit(1);
	} else {                 /* stop grabbing */
	   if((grab_cadr = grab_next(0, 1, &grabw, &grabh)) == NULL)
	      vid_exit(1);
/* printf("Stop grab; grab_cnt=%dt\n",grab_cnt); */
	}
     } else {
	mydelay(tdelay);
	if(mytime()-tlast>1.2 && tdelay < 0.1)
	   tdelay += 0.01;
     }
   }
   vid_exit(0);
}


/* Allocate shared memory segment */
static void get_sh_mem() {
   int i,j, shmid, new_shm=0;
   short *p;
   int shm_size = sizeof(struct TVmem)+sizeof(struct TVchan)*(BTNumCh-1);
   int chip_type=848;
   char card_name[32] = "BT848";
   int my_pid;


   if(curch != NULL) return;

   /* first try to find existing one */
   shmid = shmget(bttv_shm_key.code, shm_size+sizeof(int), 0777);
   if (shmid<0 && errno==ENOENT) {
      /* if no - try to create a new one */
      shmid = shmget(bttv_shm_key.code, shm_size+sizeof(int), IPC_CREAT|IPC_EXCL|0777);
      new_shm=1;
   }
   if (shmid<0) {
      char msg[80];
      sprintf(msg,"Can't create shared memory segment '%s'",bttv_shm_key.name);
      perror(msg);
      exit(errno);
   }

   /* attach it to our memory space */
   shm_addr = shmat ( shmid, NULL, 0 );
   if ((int)(shm_addr) == -1) {
      char msg[80];
      sprintf(msg,"Can't attach shared memory segment '%s'",bttv_shm_key.name);
      perror(msg);
      exit(errno);
   }
   if((shmctl(shmid, SHM_LOCK, NULL) < 0) < 0) {
      char msg[80];
      sprintf(msg,"Can't prevents swapping of shared memory segment '%s'",bttv_shm_key.name);
      perror(msg);
      exit(errno);
   }

   strncpy(card_name,(grab_fd>0)?grab_cap.name:"BT848",32);

   tvmem = (struct TVmem *)shm_addr;
   if(new_shm) {
      tvmem->size = shm_size;
      tvmem->drv_pid = 0;
      fprintf(stderr,"Create shared memory segment '%s' %dbytes at %x \n",
	   bttv_shm_key.name, shm_size, shm_addr);
   } else {
      fprintf(stderr,"Attach shared memory segment '%s' %dbytes at %x \n",
	   bttv_shm_key.name, tvmem->size, shm_addr);
      if(tvmem->size != shm_size) {
	 fprintf(stderr,"Wrong TV-memory size:%dbytes needs!\n",shm_size);
	 exit(1);
      }
      if(tvmem->numch != BTNumCh) {
	 fprintf(stderr,"Wrong TV-channels number:%d (%d needs)!\n",tvmem->numch,BTNumCh);
	 new_shm=1;
      }
      if(strncmp(tvmem->grab_name,card_name,32) != 0) {
	 fprintf(stderr,"TV-card change?! %s -> %s?!\n",tvmem->grab_name,card_name);
	 new_shm=1;
      }
   }

   my_pid = getpid();
   if( tvmem->drv_pid > 0 && kill(tvmem->drv_pid, 0) >= 0) {
      fprintf(stderr,"vs_bttv daemon already running! (PID=%d)\n", tvmem->drv_pid);
      sleep(2);
      exit(1);
   }
   strncpy(tvmem->grab_name,card_name,32);
   tvmem->numch = BTNumCh;
   tvmem->drv_pid = my_pid;

   if(strstr(tvmem->grab_name,"849")) chip_type=849;
   if(strstr(tvmem->grab_name,"878")) chip_type=878;
   if(strstr(tvmem->grab_name,"879")) chip_type=879;

   for(i=0,curch=tvmem->chmem; i<NumCh; i++,curch++) {
      tvch[i] = curch;
      if(new_shm || curch->vs_type!=chip_type) {
	 curch->num=i;
	 curch->vs_type=chip_type;
	 curch->tandem=0;
	 curch->activ=0;
	 curch->exposure = 0.04;
	 if(new_shm)
	    curch->seqn=0;
	 curch->tstart = curch->tcurr = time(0);
	 curch->pid=0;
	 curch->command=INULL;
	 curch->nexp=0;
	 curch->nadd=0;
	 curch->nadded=0;
	 curch->addlev=0;
	 curch->next_form=1;
	 curch->form=1;
	 curch->ctrl.bttv.brightness=brightness;
	 curch->ctrl.bttv.contrast=contrast;
	 curch->img_width  = (GRAB_WIDTH_1<grab_cap.maxwidth)? GRAB_WIDTH_1 : grab_cap.maxwidth;
	 curch->img_height = (GRAB_HEIGHT_1<grab_cap.maxheight)? GRAB_HEIGHT_1 : grab_cap.maxheight;
	 curch->frm_width  = curch->img_width;
	 curch->frm_height = curch->img_height;
	 curch->subfrm_width  = curch->subfrm2_width  = curch->frm_width;
	 curch->subfrm_height = curch->subfrm2_height = curch->frm_height;
	 curch->frm_col0 = curch->frm_lin0 = 0;
	 curch->subfrm_col0 = curch->subfrm_lin0 = 0;
	 curch->subfrm2_col0 = curch->subfrm2_lin0 = 0;
	 curch->pnt_col = curch->pnt_lin = 0;
	 curch->center_x = curch->center_y = 0.0;
	 if(new_shm) {
	    curch->seqmsg=0;
	    curch->msg[0]='\0';
	 }
	 curch->bias = 0;
	 for(j=0; j<MaxWidth; j++)
	     curch->flat[j] = 1;
	 for(j=0; j<2880; j++)
	     curch->fits_header[j] = 0;
	 for(p=curch->image,j=0; j<MaxWidth*MaxHeight; j++)
	     *p++ = (j % curch->img_width) - 100;
      }
   }
   curch = tvmem->chmem;
}


static void chan_msg( char *new_msg) {
    strncpy(curch->msg, new_msg, 99);
    curch->seqmsg++;
}
static void wall_msg( char *new_msg) {
    int nch;
    for(nch=0; nch<NumCh; nch++) {
	strncpy(tvch[nch]->msg, new_msg, 99);
	tvch[nch]->seqmsg++;
    }
}

static void mydelay(double dt)
{
  int nfd;
  struct timeval tv;
   tv.tv_sec = (int)dt;
   tv.tv_usec = (int)((dt - tv.tv_sec)*1000000.);
sleeping:
   nfd = select(0, (fd_set *)NULL,(fd_set *)NULL,(fd_set *)NULL, &tv);
   if(nfd < 0) {
      if(errno == EINTR) { /* interrupted?    */
	 if(vd_hup) {      /* by USR1 signal?  */
	    vd_hup=0;      /*  ok!              */
	    return;        /*    stop sleeping...*/
	 }
	/*On  Linux,  timeout  is  modified to reflect the amount of
	 time not slept; most other implementations DO NOT do this!*/
	 goto sleeping;
      }
      fprintf(stderr,"Error in mydelay(){ select() }. %s\n",strerror(errno));
   }
}

static double mytime()
{
  struct timeval tv;
  struct timezone tz;
  int ret;
  double t;
  ret = gettimeofday(&tv, &tz);
  t = (double)tv.tv_sec + (double)tv.tv_usec*0.000001;
  t -= timezone;
  return(t);
}

static void myabort(int sig) {
    int ret;
    char ss[10], tmp[80];
    signal(sig,SIG_IGN);
    switch (sig) {
	case SIGHUP : strcpy(ss,"SIGHUP");  break;
	case SIGINT : strcpy(ss,"SIGINT");  break;
	case SIGQUIT: strcpy(ss,"SIGQUIT"); break;
	case SIGILL : strcpy(ss,"SIGILL");  break;
	case SIGABRT: strcpy(ss,"SIGABRT"); break;
	case SIGFPE : strcpy(ss,"SIGFPE");  break;
	case SIGPIPE: strcpy(ss,"SIGPIPE"); break;
	case SIGSEGV: strcpy(ss,"SIGSEGV"); break;
	case SIGTERM: strcpy(ss,"SIGTERM"); break;
	default:  sprintf(ss,"SIG_%d",sig); break;
    }
    switch (sig) {
	case SIGUSR1 :
	     vd_hup=1;
	     signal(sig, myabort);
	     return;
	default:
	case SIGINT :
	case SIGUSR2 :
	     fprintf(stderr,"%s: %s - Ignore .....\n",myname,ss);
	     fflush(stderr);
	     signal(sig, myabort);
	     return;
	case SIGHUP :
	case SIGQUIT:
	case SIGILL :
	case SIGABRT:
	case SIGFPE :
	case SIGSEGV:
	case SIGPIPE:
	case SIGTERM:
	     signal(SIGALRM, SIG_IGN);
	     fprintf(stderr,"%s: %s - programm stop!\n",myname,ss);
	     vid_exit(sig);
    }
}

static void vid_exit(int n) {
   tvmem->drv_pid = 0;
   /* restore image properties */
   if (ioctl(grab_fd,VIDIOCSPICT,&grab_pict_save) < 0 ) {
       perror("ioctl VIDIOCSPICT");
   }
   close(grab_fd);
   exit(n);
}

